//
//   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
//   you may not use TUMME – version 3.1 except in compliance with the License.
//   You may obtain a copy of the License at
//       http://www.apache.org/licenses/LICENSE-2.0
//   The license is also given in the LICENSE file.
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and limitations under the License.
//
// -------------------------------------------------------------------------------------------
// 
//   TUMME 3.1
//           
//   Rui Ming Zhang
//     
//   Center for Combustion Energy, Department of Energy and Power
//   Engineering, and Key Laboratory for Thermal Science and Power 
//   Engineering of Ministry of Education, Tsinghua University, Beijing
//   
//   Department of Chemistry, Chemical Theory Center, and Minnesota 
//   Supercomputing Institute, University of Minnesota, Minneapolis, 
//   Minnesota
//                  
//   Xuefei Xu
//
//   Center for Combustion Energy, Department of Energy and Power
//   Engineering, and Key Laboratory for Thermal Science and Power 
//   Engineering of Ministry of Education, Tsinghua University, Beijing
//
//   Donald G. Truhlar
// 
//   Department of Chemistry, Chemical Theory Center, and Minnesota 
//   Supercomputing Institute, University of Minnesota, Minneapolis, 
//   Minnesota
//  
//   The master equation code includes multi-structural variational 
//   transition state theory with small-curvature tunneling (MS-VTST/SCT)
//   for calculating the microcanonical flux coefficients needed as input
//   to the master equation. The TUMME program can do double, quadruple, 
//   and octuple precision, it can execute with MP or MPI parallel schemes, 
//   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
//   output files. Report bugs to: zrm17@tsinghua.org.cn
// 
//   Copyright 2020~2023
//   Donald G. Truhalar and Regents of the University of Minnesota
//   Xuefei Xu and the Tsinghua University 
//


#include "Interface_qd.h"
#include "ME_solver_qd.h"

//#define DEBUG
//#define PSEU_DEBUG
#define print(i) std::cout<<i<<std::endl



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                        Interface to ME      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void ME_qd_Main(
    int32_t ompnumthread, 
    WellStack well_stack,   int32_t Nwell, 
    kEStack kE_stack,       int32_t kE_stack_sz,
    Param param,            int32_t Nph, 
    double *evals,          double *evecs)
{
    #ifdef DEBUG
    print("ME solver in quadruple begins:");
    #endif
    
    unsigned int oldcw;
    fpu_fix_start(&oldcw);
    omp_set_dynamic(0);
    omp_set_num_threads(ompnumthread);

    #ifdef DEBUG
    #pragma omp parallel 
    {
    std::cout<<"Using number of threads in openMP: "<<ompnumthread<<std::endl;
    }
    #endif

    qd_real** kernel_stack_qd   = new qd_real*[Nwell];
    double ** kE_map            = new double *[Nwell* Nph];
    int32_t*  well_offset       = new int32_t[Nwell];
    int32_t   mat_size          = 0;
    
    //redirect the pointer to allocated memory 
    for ( int32_t i = 0; i < Nwell; i++ ){
        double * _dos = new double[ (well_stack+i)->dos_sz ];
        memmove( _dos, (well_stack+i)->dos, (well_stack+i)->dos_sz*sizeof(double) );
        (well_stack+i)->dos = _dos;
    }
    
    for ( int32_t i = 0; i < kE_stack_sz; i++ ){
        double * _ke = new double[(kE_stack+i)->kE_sz];
        memmove( _ke, (kE_stack+i)->kE, (kE_stack+i)->kE_sz*sizeof(double) );
        (kE_stack+i)->kE = _ke;
    }

    //create a index map of kE
    for( int32_t i = 0; i < Nwell; i++ )
        for( int32_t j = 0; j < Nph; j++ )
            kE_map[ j + i * Nph ] = NULL;
    for(int32_t i = 0; i < kE_stack_sz; i++ ) {
        int32_t I = kE_stack[i].ph1_index;
        int32_t J = kE_stack[i].ph2_index;
        kE_map[ J + I * Nph ] = kE_stack[i].kE;
    }

    //create wells' kernel and truncate dos
    Well well = well_stack;
    for( int32_t i = 0; i < Nwell; i++ ){
        Creat_Kernel( well, param, kernel_stack_qd[i] );
        Trunc_W_DoS( well, param );
        well++;
    }

    //deal with variable about size
    mat_size = well_stack[0].size;
    well_offset[0]=0;
    for(int32_t i=1; i<Nwell; i++) {
        well_offset[i] = well_offset[i-1] + well_stack[i-1].size;
        mat_size   += well_stack[i].size;
    }

    #ifdef DEBUG
    std::cout<<"Truncate global relaxation matrix size = "<<mat_size<<std::endl;
    #endif
    
    //solve ME using qd
    ME_qd_Solver(
    well_stack, 
    param, 
    well_offset, 
    Nwell, Nph, 
    mat_size, 
    kernel_stack_qd, 
    kE_map, 
    evals, 
    evecs);

    //release
    for(int32_t i = 1; i < Nwell; i++){
        delete[]kernel_stack_qd[i];
    }
    delete[]well_offset;
    delete[]kE_map;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                    Interface to pseudo ME      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void pseuME_qd_Main(
    int32_t ompnumthread, 
    WellStack well_stack,   int32_t Nwell, 
    kEStack kE_stack,       int32_t kE_stack_sz,
    dkStack dk_stack,       int32_t dk_stack_sz,
    Param param,            int32_t Nph,  
    Pseubim pseubim_stack,  int32_t Npseubim,
    double *evals,          double *evecs )
{
    #ifdef PSEU_DEBUG
    print("ME solver in quadruple begins:");
    #endif
    
    unsigned int oldcw;
    fpu_fix_start(&oldcw);
    omp_set_dynamic(0);
    omp_set_num_threads(ompnumthread);

    #ifdef PSEU_DEBUG
    std::cout<<"Using number of threads in openMP: "<<ompnumthread<<std::endl;
    #endif

    qd_real** kernel_stack_qd   = new qd_real*[Nwell];
    double ** kE_map            = new double *[Nwell* Nph];
    double ** dk_map            = new double *[(Nph-Nwell)*Nph];
    int32_t*  well_offset       = new int32_t[Nwell];
    int32_t   mat_size          = 0;
    
    //redirect the pointer to allocated memory 
    for ( int32_t i = 0; i < Nwell; i++ ){
        double * _dos = new double[ (well_stack+i)->dos_sz ];
        memmove( _dos, (well_stack+i)->dos, (well_stack+i)->dos_sz*sizeof(double) );
        (well_stack+i)->dos = _dos;
    }
    
    for ( int32_t i = 0; i < kE_stack_sz; i++ ){
        double * _ke = new double[(kE_stack+i)->kE_sz];
        memmove( _ke, (kE_stack+i)->kE, (kE_stack+i)->kE_sz*sizeof(double) );
        (kE_stack+i)->kE = _ke;
    }

    for ( int32_t i = 0; i < dk_stack_sz; i++ ){
        double * _dk = new double[(dk_stack+i)->dk_sz];
        memmove( _dk, (dk_stack+i)->dk, (dk_stack+i)->dk_sz*sizeof(double) );
        (dk_stack+i)->dk = _dk;
    }

    //create a index map of kE
    for( int32_t i = 0; i < Nwell; i++ )
        for( int32_t j = 0; j < Nph; j++ )
            kE_map[ j + i * Nph ] = NULL;
    for(int32_t i = 0; i < kE_stack_sz; i++ ) {
        int32_t I = kE_stack[i].ph1_index;
        int32_t J = kE_stack[i].ph2_index;
        kE_map[ J + I * Nph ] = kE_stack[i].kE;
    }

    //create a index map of dk
    for( int32_t i = 0; i < Nph-Nwell; i++ )
        for( int32_t j = 0; j < Nph; j++ )
            dk_map[ j + i * Nph ] = NULL;
    for(int32_t i = 0; i < dk_stack_sz; i++ ) {
        int32_t I = dk_stack[i].ph1_index;
        int32_t J = dk_stack[i].ph2_index;
        dk_map[ J + I * Nph ] = dk_stack[i].dk;
    }

    //create wells' kernel and truncate dos
    Well well = well_stack;
    for( int32_t i = 0; i < Nwell; i++ ){
        Creat_Kernel( well, param, kernel_stack_qd[i] );
        Trunc_W_DoS( well, param );
        well++;
    }

    //deal with variable about size
    mat_size = well_stack[0].size;
    well_offset[0]=0;
    for(int32_t i=1; i<Nwell; i++) {
        well_offset[i] = well_offset[i-1] + well_stack[i-1].size;
        mat_size      += well_stack[i].size;
    }
    mat_size  += Npseubim;

    #ifdef PSEU_DEBUG
    std::cout<<"Transition matrix size = "<<mat_size<<std::endl;
    #endif

    //solve pseudo-first order ME using qd
    pseuME_qd_Solver(
        well_stack, 
        param, 
        pseubim_stack,
        well_offset, 
        Nwell, 
        Nph, 
        Npseubim,
        mat_size, 
        kernel_stack_qd, 
        kE_map,
        dk_map, 
        evals, 
        evecs);


    //release
    for(int32_t i = 1; i < Nwell; i++){
        delete[]kernel_stack_qd[i];
    }
    delete[]well_offset;
    delete[]kE_map;
}
