//
//   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
//   you may not use TUMME – version 3.1 except in compliance with the License.
//   You may obtain a copy of the License at
//       http://www.apache.org/licenses/LICENSE-2.0
//   The license is also given in the LICENSE file.
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and limitations under the License.
//
// -------------------------------------------------------------------------------------------
// 
//   TUMME 3.1
//           
//   Rui Ming Zhang
//     
//   Center for Combustion Energy, Department of Energy and Power
//   Engineering, and Key Laboratory for Thermal Science and Power 
//   Engineering of Ministry of Education, Tsinghua University, Beijing
//   
//   Department of Chemistry, Chemical Theory Center, and Minnesota 
//   Supercomputing Institute, University of Minnesota, Minneapolis, 
//   Minnesota
//                  
//   Xuefei Xu
//
//   Center for Combustion Energy, Department of Energy and Power
//   Engineering, and Key Laboratory for Thermal Science and Power 
//   Engineering of Ministry of Education, Tsinghua University, Beijing
//
//   Donald G. Truhlar
// 
//   Department of Chemistry, Chemical Theory Center, and Minnesota 
//   Supercomputing Institute, University of Minnesota, Minneapolis, 
//   Minnesota
//  
//   The master equation code includes multi-structural variational 
//   transition state theory with small-curvature tunneling (MS-VTST/SCT)
//   for calculating the microcanonical flux coefficients needed as input
//   to the master equation. The TUMME program can do double, quadruple, 
//   and octuple precision, it can execute with MP or MPI parallel schemes, 
//   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
//   output files. Report bugs to: zrm17@tsinghua.org.cn
// 
//   Copyright 2020~2023
//   Donald G. Truhalar and Regents of the University of Minnesota
//   Xuefei Xu and the Tsinghua University 
//

#include "ME_solver_qd.h"

//#define DEBUG
//#define PSEU_DEBUG
#define print(i) std::cout<<i<<std::endl


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Construct symmetrized transition matrix and calc eigenpairs      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void  ME_qd_Solver(
                    WellStack &well_stack,      Param &param, 
                    int32_t* &well_offset,      const int32_t &Nwell,
                    const int32_t &Nph,         const int32_t &mat_size,
                    qd_real** &kernel_stack_qd, double** &kE_map, 
                    double* &evals,             double* &evecs)
{
    //create kE sum
    double** kE_sum = new double*[Nwell];
    for( int32_t i  =0; i < Nwell; i++ ){
        kE_sum[i] = new double[ well_stack[i].size ];
        memset( kE_sum[i], 0, well_stack[i].size*sizeof(double) );
        for( int32_t j = 0; j < Nph; j++ ){ 
            if ( kE_map[ j + i * Nph ] != NULL )
                array_add( kE_sum[i], kE_map[j+i*Nph], well_stack[i].size );
        }
    }

    //create transition matrix
    int32_t lda = mat_size;
    Well wellx;
    Well welly;
    qd_real* kin_mat_qd = new qd_real[ mat_size*mat_size ];
    for( int32_t well_i = 0; well_i < Nwell; well_i++ ){
        wellx = well_stack + well_i;
        int32_t i0 = well_offset[ well_i ]; 
        for( int32_t well_j = 0; well_j < Nwell; well_j++){
            welly = well_stack + well_j;
            int32_t j0 = well_offset[ well_j ];
            if ( well_i == well_j){ //diag-block
                qd_real* kernel_qd = kernel_stack_qd[ well_i ];
                #pragma omp parallel for 
                for ( int32_t i = 0; i < wellx->size; i++ ){
                    for ( int32_t j = 0; j < welly->size; j++ ){
                        if ( i != j )  //non-diag element
                            kin_mat_qd(i0 + i, j0 + j)= - wellx->Z*kernel_qd[j + wellx->size*i]*\
                                                        sqrt(exp( (j-i)*(qd_real)param->ener_step*param->beta )*wellx->dos_t[j]/wellx->dos_t[i]);
                        else  //diag element
                            kin_mat_qd(i0 + i, j0 + j) = wellx->Z*( 1E0 - kernel_qd[ j + wellx->size*i ] ) + kE_sum[well_i][i];
                    }
                }
            }
            else{//non-diag block
                double* kE = kE_map[ well_i + well_j * Nph ];
                if (kE != NULL){
                    #pragma omp parallel for
                    for(int32_t i = 0; i < min( wellx->size, welly->size ); i++){
                        kin_mat_qd( i0 + i, j0 + i ) = - kE[i] * sqrt( (qd_real)welly->dos_t[i] /(qd_real)wellx->dos_t[i] );
                    }
                }
            }
        }
    }

    //eigen calc
    qd_real * w = new qd_real[mat_size];
    Sym_Eig(kin_mat_qd,w,lda);

    //assign eigen 
    #pragma omp parallel for    
    for ( size_t i = 0; i < mat_size; i++ ){
	    evals[i] = to_double( w[i] );
	}
    #pragma omp parallel for
    for ( size_t i = 0; i < mat_size*mat_size; i++ ){
    	evecs[i] = to_double( kin_mat_qd[i] );
    } 
    
    //release 
    delete[]kin_mat_qd;
    delete[]w;
    for(int32_t i=1;i<Nwell;i++){
        delete[]kE_sum[i];
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    Construct pseudo-first-order symmetrized transition matrix and calc eigenpairs      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void  pseuME_qd_Solver(
                    WellStack &well_stack,      Param &param, 
                    Pseubim &Pseubim_stack,     int32_t* &well_offset,     
                    const int32_t &Nwell,       const int32_t &Nph,
                    const int32_t &Npseubim,    const int32_t &mat_size,  
                    qd_real** &kernel_stack_qd,
                    double** &kE_map,           double** &dk_map,           
                    double* &evals,             double* &evecs)
{

    //create kE sum
    double** kE_sum = new double*[Nwell];
    for( int32_t i  =0; i < Nwell; i++ ){
        kE_sum[i] = new double[ well_stack[i].size ];
        memset( kE_sum[i], 0, well_stack[i].size*sizeof(double) );
        for( int32_t j = 0; j < Nph; j++ ){ 
            if ( kE_map[ j + i * Nph ] != NULL )
                array_add( kE_sum[i], kE_map[j+i*Nph], well_stack[i].size );
        }
    }

    //create dk sum
    int32_t  Nbim   = Nph-Nwell;
    double* dk_sum  = new double[Nbim];
    memset( dk_sum, 0, Nbim*sizeof(double) );
    for( int32_t i  = 0; i < Nbim; i++ ){
        for( int32_t j = 0; j < Nph; j++ ){
            if ( dk_map[ j + i * Nph ] != NULL )
                for ( int32_t ii = 0; ii < well_stack[j].size; ii++ )
                    dk_sum[i] += (dk_map[j+i*Nph])[ii] ;
        }
    }

    //create transition matrix
    int32_t i0, j0;
    int32_t lda = mat_size;
    int32_t Ny  = mat_size - Npseubim;
    Well    wellx;
    Well    welly;
    Pseubim pseubim;
    qd_real* kin_mat_qd = new qd_real[ mat_size*mat_size ];
    memset( kin_mat_qd, 0, mat_size*mat_size*sizeof(qd_real) );

    //well -->
    for( int32_t well_i = 0; well_i < Nwell; well_i++ ){
        wellx = well_stack + well_i;
        i0 = well_offset[ well_i ]; 
        //well->well
        for( int32_t well_j = 0; well_j < Nwell; well_j++){
            welly       = well_stack + well_j;
            j0  = well_offset[ well_j ];
            if ( well_i == well_j){ //diag-block
                qd_real* kernel_qd = kernel_stack_qd[ well_i ];
                #pragma omp parallel for 
                for ( int32_t i = 0; i < wellx->size; i++ ){
                    for ( int32_t j = 0; j < welly->size; j++ ){
                        if ( i != j )  //non-diag element
                            kin_mat_qd(i0 + i, j0 + j)= - wellx->Z*kernel_qd[j + wellx->size*i]*\
                                                        sqrt(exp( (j-i)*(qd_real)param->ener_step*param->beta )*wellx->dos_t[j]/wellx->dos_t[i]);
                        else  //diag element
                            kin_mat_qd(i0 + i, j0 + j) = wellx->Z*( 1E0 - kernel_qd[ j + wellx->size*i ] ) + kE_sum[well_i][i];
                    }
                }
            }
            else{//non-diag block
                double* kE = kE_map[ well_i + well_j * Nph ];
                if (kE != NULL){
                    #pragma omp parallel for
                    for(int32_t i = 0; i < min( wellx->size, welly->size ); i++){
                        kin_mat_qd( i0 + i, j0 + i ) = - kE[i] * sqrt( (qd_real)welly->dos_t[i] /(qd_real)wellx->dos_t[i] );
                    }
                }
            }
        }

        //bimolecular->well
        j0      =   Ny;
        for( int32_t i = 0; i < wellx->size; i++){
            for( int32_t pseubim_j = 0; pseubim_j < Npseubim; pseubim_j++ ){
                pseubim =   Pseubim_stack + pseubim_j;
                int32_t bim_j   =   pseubim->index ;
                if ( kE_map[ Nwell + bim_j + well_i*Nph ] != NULL )
                    kin_mat_qd( i0 + i, j0 + pseubim_j ) = - sqrt( (dk_map[ well_i + bim_j*Nph ])[i] * pseubim->excessconc * (kE_map[ Nwell + bim_j + well_i*Nph ])[i] );
            }            
        }
    }

    //well->bimolecular
    i0 = Ny;
    for( int32_t pseubim_i = 0; pseubim_i < Npseubim; pseubim_i++){
        pseubim = Pseubim_stack + pseubim_i;
        int32_t bim_i   = pseubim->index;
        for( int32_t well_j = 0; well_j < Nwell; well_j++){
            if ( kE_map[ Nwell + bim_i + well_j*Nph ] != NULL ){
                welly = well_stack + well_j;
                j0 = well_offset[well_j];
                for ( int32_t j = 0; j < welly->size; j++)
                    kin_mat_qd( i0 + pseubim_i, j0 + j ) = - sqrt( (dk_map[ well_j + bim_i*Nph ])[j] * pseubim->excessconc * (kE_map[ Nwell + bim_i + well_j*Nph ])[j] );  
            }
        }
    }    

    //bimolecular -->
    i0 = Ny;
    for( int32_t bim_i = 0; bim_i < Npseubim; bim_i++){
        pseubim = Pseubim_stack + bim_i;
        kin_mat_qd( i0 + bim_i, i0 + bim_i ) = dk_sum[ pseubim->index ] * pseubim->excessconc;  
        }  

    //eigen calc
    qd_real * w = new qd_real[mat_size];
    Sym_Eig(kin_mat_qd,w,lda);

    //assign eigen 
    #pragma omp parallel for    
    for ( size_t i = 0; i < mat_size; i++ ){
	    evals[i] = to_double( w[i] );
	}
    #pragma omp parallel for
    for ( size_t i = 0; i < mat_size*mat_size; i++ ){
    	evecs[i] = to_double( kin_mat_qd[i] );
    } 
    
    //release 
    delete[]kin_mat_qd;
    delete[]w;
    for(int32_t i=1;i<Nwell;i++){
        delete[]kE_sum[i]; 
    }

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                Calculate eigenpairs of a matrix      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void Sym_Eig(qd_real* &A, qd_real* &w, const int32_t  &ndim)
{
    mpackint n = ndim;
    mpackint lwork, info;

    //work space query
    lwork = -1;
    qd_real *work = new qd_real[1];

    Rsyev("V", "U", n, A, n, w, work, &lwork, &info);
    lwork = (int)work[0].x[0];
    delete[]work;
    work = new qd_real[std::max((mpackint)1, lwork)];
    //inverse matrix
    Rsyev("V", "U", n, A, n, w, work, &lwork, &info);
    delete[]work;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
             Normalize the relaxational possibility kernel      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void Creat_Kernel( Well &well, Param &param, qd_real* &kernel_qd ){
    int32_t dos_i,dos_j ;
    int32_t lda  = well->size ;
    qd_real E0   = well->E0;
    qd_real E0_t = well->E0_t ;
    qd_real * C  = new qd_real[lda] ;
    qd_real * kernel_T_qd = new qd_real[lda*lda] ;
    qd_real ener_i = 0 ;
    qd_real ener_j = 0 ;
    qd_real T_factor = 0 ;

    T_factor = pow( (qd_real) 1/(qd_real)param->beta/(qd_real)param->Tc, (qd_real)param->exponent );

    #pragma omp parallel for private(ener_i,ener_j,dos_i,dos_j)
    for ( int32_t j = 0; j < lda; j++ ){
        for( int32_t i = 0; i < j; i ++ ){
            ener_j = (qd_real)( (j+1) * param->ener_step + E0_t );
            ener_i = (qd_real)( (i+1) * param->ener_step + E0_t );
            kernel_T_qd(j,i) = exp( -(ener_j-ener_i) / T_factor / ( (qd_real)param->slope * (ener_j - E0) +(qd_real)param->EdConst ) );              
        }
        for( int32_t i = j; i < lda; i++ ){
            ener_j = (qd_real)( (j+1) * param->ener_step + E0_t );
            ener_i = (qd_real)( (i+1) * param->ener_step + E0_t );
            dos_j  = int32_t( round( to_double( ( ener_j - well->E0 ) / param->dE) ) );
            dos_i  = int32_t( round( to_double( ( ener_i - well->E0 ) / param->dE) ) );
            kernel_T_qd(j,i) = exp( -(ener_i-ener_j) / T_factor / ( (qd_real)param->slope * (ener_i-E0) + (qd_real)param->EdConst ) - (ener_i-ener_j)*param->beta ) *\
                               (qd_real)well->dos[dos_i]/(qd_real)well->dos[dos_j];
        }
    }

    int32_t cycle_index = 0, index_trunc = 0;
    qd_real unit = 1E0;
    while(true){
        index_trunc = cycle_index;
        for ( int32_t i = lda-1; i >= index_trunc; i-- ){
            qd_real sum1_qd = 0;
            qd_real sum2_qd = 0;
            
            #pragma omp parallel
            { 
            qd_real sum1_qd_priv = 0;
            qd_real sum2_qd_priv = 0;
            #pragma omp for
            for ( int32_t j = index_trunc; j <= i; j++ ) sum1_qd_priv += kernel_T_qd(i,j);
            #pragma omp for
            for ( int32_t j = i+1; j < lda; j++ ) sum2_qd_priv += kernel_T_qd(i,j) * C[j];
            #pragma omp critical
            {
            sum1_qd += sum1_qd_priv;
            sum2_qd += sum2_qd_priv;
            }
            }
            if ( sum2_qd > unit ){
                cycle_index = i + 1;
                break;
            }
            else 
                C[i] = ( unit - sum2_qd ) / sum1_qd;
        }
        if ( index_trunc == cycle_index ) break;
    }

    #pragma omp parallel for
    for ( int32_t i = index_trunc; i < lda; i++ ){
        for ( int32_t j = index_trunc; j <= i; j++ ) kernel_T_qd(i,j) *= C[i];
        for ( int32_t j = i+1; j < lda; j++ ) kernel_T_qd(i,j) *= C[j];
    }
    Trunc_Transpose(kernel_T_qd, index_trunc, lda, kernel_qd);

    well->size  -=  index_trunc;
    well->E0_t  +=  ( index_trunc+1 ) * param->ener_step;

    delete[]kernel_T_qd;
    delete[]C;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
            Truncate and transpose relaxational kernel metrix    
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void Trunc_Transpose( 
    qd_real* &kernel_T_qd, const int32_t &index_trunc, 
    const int32_t &well_size, qd_real* &kernel_qd ){

    int32_t  lda            = well_size - index_trunc;
    qd_real* kernel_qd_rev  = new qd_real[ lda * lda ];
    kernel_qd               = new qd_real[ lda * lda ];

    #pragma omp parallel for
    for ( int32_t j = 0; j < lda; j++ )
        for ( int32_t i = 0; i < lda; i++ )
            kernel_qd_rev( i, j ) = kernel_T_qd[ (i+index_trunc) + well_size * (j+index_trunc) ];

    #pragma omp parallel for
    for ( int32_t i = 0; i < lda*lda; i++ )
        kernel_qd[ i ] = kernel_qd_rev [ lda*lda - 1 - i ]; 
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                      Create truncated DoS array      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void Trunc_W_DoS(Well &well,Param &param){
    well->dos_t = new double[well->size];

    #pragma omp parallel for
    for (int32_t i = 0; i < well->size; i++ )
        well->dos_t[i] = well->dos[ int32_t(round( ( param->ener_ref - well->E0 - param->ener_step*i ) / param->dE ) ) ];
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                        Add two array      
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void array_add(double* &cum,double* &a,int32_t &size){
    #pragma omp parallel for
    for( int32_t i = 0; i < size; i++ ) cum[i] += a[i];
}
