#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import numpy         as np
import global_lib    as gl
import const_lib     as cl



class LJ_collision_class(gl.CycleParam_class):

    def __init__(
        self,
        _epsilonM,
        _epsilonA,
        _massM,
        _massA,
        _diameterM,
        _diameterA
):
        self.__epsilonM =   _epsilonM 
        self.__epsilonA =   _epsilonA  
        self.__diameterM=   _diameterM      
        self.__diameterA=   _diameterA      
        self.__massM    =   _massM        
        self.__massA    =   _massA           
        self.__kc       =   1
        self.__Z        =   1
        self.__omega    =   1
        self.__kHS      =   1


    @property
    def epsilonM (self): return self.__epsilonM
    @property
    def epsilonA (self): return self.__epsilonA
    @property
    def diameterM(self): return self.__diameterM  
    @property
    def diameterA(self): return self.__diameterA  
    @property
    def massM    (self): return self.__massM   
    @property
    def massA    (self): return self.__massA    
    @property
    def kc       (self): return self.__kc   
    @property
    def Z        (self): return self.__Z   
    @property
    def omega    (self): return self.__omega
    @property
    def kHS      (self): return self.__kHS
    @property
    def name     (self): return 'LJ'



    def setCollisionFrequency(self):
        T   =   self.T
        p   =   self.p
        mu  =   (self.massA * self.massM) / (self.massM + self.massA)
        # collison rate constant for hard-sphere
        self.__kHS = np.pi * ( self.diameterA + self.diameterM )**2 / 4 * np.sqrt(8 * cl.cst.kb * T / np.pi / mu )
        #omega
        tmp = cl.cst.kb * T / np.sqrt( self.epsilonM * self.epsilonA )
        if tmp >= 0.3 and tmp <= 3:
            self.__omega = 1 / ( 0.697 + 0.5185*np.log10(tmp))
        elif tmp >= 3 and tmp <= 300:
            self.__omega = 1 / ( 0.636 + 0.567*np.log10(tmp))
        elif tmp < 0.3:
            self.__omega = 1 / ( 0.697 + 0.5185*np.log10(tmp))
            print(" WARNING : kBT/sqrt(epsilonA*epsilonM) < 0.3. It should be located in [0.3, 300].\nThe Omega*2,2 will use the expression as [0.3,3]")
        elif tmp > 300:
            self.__omega = 1 / ( 0.636 + 0.567*np.log10(tmp))
            print(" WARNING : kBT/sqrt(epsilonA*epsilonM) > 300. It should be located in [0.3, 300].\nThe Omega*2,2 will use the expression as [3,300]")

        #Concentration
        n_M =   p / cl.cst.kb / T
        #collision rate constant
        self.__kc  =   self.omega * self.kHS
        #collision frequency
        self.__Z   =   self.kc * n_M
        


class HS_collision_class(gl.CycleParam_class):

    def __init__(
        self,
        _bmax, 
        _massM,
        _massA
):  
        self.__bmax     =   _bmax         
        self.__massM    =   _massM        
        self.__massA    =   _massA           
        self.__kc       =   1
        self.__Z        =   1
        self.__omega    =   1
        self.__kHS      =   1

    @property
    def bmax     (self): return self.__bmax  
    @property
    def massM    (self): return self.__massM   
    @property
    def massA    (self): return self.__massA    
    @property
    def kc       (self): return self.__kc   
    @property
    def Z        (self): return self.__Z   
    @property
    def omega    (self): return self.__omega
    @property
    def kHS      (self): return self.__kHS
    @property
    def name     (self): return 'HS'


    def setCollisionFrequency(self):
        T   =   self.T
        p   =   self.p
        mu  =   (self.massA * self.massM) / (self.massM + self.massA)
        # collison rate constant for hard-sphere
        self.__kHS = np.pi * self.bmax**2 * np.sqrt(8 * cl.cst.kb * T / np.pi / mu )

        #Concentration
        n_M =   p / cl.cst.kb / T
        #collision rate constant
        self.__kc  =   self.kHS
        #collision frequency
        self.__Z   =   self.kc * n_M


    