#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



from math import pi

class Const_class(object):
    
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #             PHYSIC CONSTANT in atomic unit
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    __h   =   2 * pi             
    __kb  =   1.0               
    __NA  =   6.022140857E+23
    __hba =   1.0


    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                   UNIT EXCHANGE
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    #
    #mass
    #
    __amu2kg          =   1E-3    /  __NA
    __kg2amu          =   1       /  __amu2kg
    __gmol2kg         =   1       /  ( 1000 * __NA )
    __amu2au          =   1822.888
    __au2amu          =   1       /  __amu2au
    __au2kg           =   __au2amu*  __amu2kg
    __kg2au           =   1       /  __au2kg
    #
    #energy
    #
    __cm2J            =   1.986445824E-23
    __kcalmol2kJmol   =   4.184
    __cm2Jmol         =   __cm2J            *   __NA
    __cm2kJ           =   __cm2J            /   1000
    __kcalmol2cm      =   349.7551
    __kcal2kJ         =   4.184
    __cm2kcalmol      =   1               /   __kcalmol2cm
    __kcal2J          =   __kcal2kJ       *   1000
    __kJ2kcal         =   1               /   __kcal2kJ
    __J2kcal          =   1               /   __kcal2J
    __J2cm            =   1               /   __cm2J
    __Jmol2cm         =   __J2cm          /   __NA
    __kJmol2cm        =   1000            *   __Jmol2cm
    __kcalmol2J       =   __kcalmol2kJmol *   1000 / __NA
    __kcalmol2Jmol    =   __kcalmol2kJmol *   1000
    __eV2au           =   0.03674931711
    # au -> 
    __au2J            =   4.359744650E-18
    __au2kcal         =   1.042003979E-21
    __au2kcalmol      =   627.5095
    __au2cal          =   __au2kcal         *   1000
    __au2kJ           =   __au2J            /   1000
    __au2cm           =   __au2J            /   __cm2J
    __au2kJmol        =   __au2kcalmol      *   __kcalmol2kJmol
    __au2eV           =   27.21139
    # -> au
    __cm2au           =   1               /   __au2cm
    __J2au            =   1               /   __au2J
    __kcal2au         =   1               /   __au2kcal
    __kcalmol2au      =   1               /   __au2kcalmol
    __cal2au          =   1               /   __au2cal
    __kJ2au           =   1               /   __au2kJ
    __kJmol2au        =   1               /   __au2kJmol

    #
    #length
    #
    __au2A            =   0.52917721067
    __A2au            =   1               /   __au2A
    __A2m             =   1.0E-10
    __m2cm            =   100
    __cm2m            =   0.01
    __dm2au           =   1E9             *  __A2au
    __au2m            =   __au2A          *  __A2m
    __m2au            =   1 / __au2m

    #
    #pressure
    #
    __torr2Pa         =   133.322
    __Pa2torr         =   1               /   __torr2Pa
    __atm2Pa          =   101325
    __bar2Pa          =   1E5
    __au2Pa           =   2.9421015697E13
    __Pa2au           =   1               /     __au2Pa
    __torr2au         =   __torr2Pa         *   __Pa2au
    __au2torr         =   1               /     __torr2au
    __atm2au          =   __atm2Pa          *   __Pa2au
    __bar2au          =   __bar2Pa          *   __Pa2au             

    #
    # temperature
    #
    __au2K            =   3.1577464E5
    __K2au            =   1               /   __au2K

    #
    # time
    #
    __au2s            =   2.4188843265857E-17
    __s2au            =   1               /   __au2s

    #
    # concentration
    #
    __molL2au        =    __NA / (__dm2au)**3

    #
    #type
    #
    __str             =   "str"
    __float           =   "float"
    __arr             =   "array"
    __elev            =   "elev"
    __geom            =   "geom"
    __bool            =   "bool"

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                   Dictionaries
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    #
    # dictionary of mass
    #
    __mass_dict   =   {                                                  
           'H':   1.007825*__amu2au , 'He':  4.002600*__amu2au , 'Li':  7.016000*__amu2au ,  'Be':  9.012180*__amu2au ,  'B':  11.009310*__amu2au ,
           'C':  12.000000*__amu2au , 'N':  14.003070*__amu2au , 'O' : 15.994910*__amu2au ,  'F':  18.998400*__amu2au ,  'Ne': 19.992440*__amu2au ,  
           'Na': 22.989800*__amu2au , 'Mg': 23.985040*__amu2au , 'Al': 26.981530*__amu2au ,  'Si': 27.976930*__amu2au ,  'P':  30.973760*__amu2au , 
           'S':  31.972070*__amu2au , 'Cl': 34.968850*__amu2au , 'Ar': 39.948000*__amu2au ,  'K':  38.963710*__amu2au ,  'Ca': 39.962590*__amu2au ,
           'Sc': 44.955920*__amu2au , 'Ti': 47.900000*__amu2au , 'V' : 50.944000*__amu2au ,  'Cr': 51.940500*__amu2au ,  'Mn': 54.938100*__amu2au , 
           'Fe': 55.934900*__amu2au , 'Co': 58.933200*__amu2au , 'Ni': 57.935300*__amu2au ,  'Cu': 62.929800*__amu2au ,  'Zn': 63.929100*__amu2au , 
           'Ga': 68.925700*__amu2au , 'Ge': 73.921900*__amu2au , 'As': 74.921600*__amu2au ,  'Se': 79.916500*__amu2au ,  'Br': 78.918300*__amu2au , 
           'Kr': 83.911500*__amu2au , 'Rb': 84.911700*__amu2au , 'Sr': 87.905600*__amu2au ,  'Y':  89.905400*__amu2au ,  'Zr': 89.904300*__amu2au ,
           'Nb': 92.906000*__amu2au , 'Mo': 97.905500*__amu2au , 'Tc': 97.000000*__amu2au ,  'Ru':101.903700*__amu2au ,  'Rh':102.904800*__amu2au ,  
           'Pd':105.903200*__amu2au , 'Ag':106.904100*__amu2au , 'Cd':113.903600*__amu2au ,  'In':114.904100*__amu2au ,  'Sn':119.902200*__amu2au ,  
           'Sb':120.903800*__amu2au , 'Te':129.906700*__amu2au , 'I' :126.904400*__amu2au ,  'Xe':131.904200*__amu2au ,  'Cs':132.905400*__amu2au ,  
           'Ba':137.905200*__amu2au , 'La':138.906300*__amu2au , 'Ce':139.905400*__amu2au ,  'Pr':140.907600*__amu2au ,  'Nd':141.907700*__amu2au ,
           'Pm':144.912700*__amu2au , 'Sm':151.919700*__amu2au , 'Eu':152.921200*__amu2au ,  'Gd':157.924100*__amu2au ,  'Tb':158.925300*__amu2au , 
           'Dy':163.929200*__amu2au , 'Ho':164.930300*__amu2au , 'Er':165.930300*__amu2au ,  'Tm':168.934200*__amu2au ,  'Yb':173.938900*__amu2au ,  
           'Lu':174.940800*__amu2au , 'Hf':179.946500*__amu2au , 'Ta':180.948000*__amu2au ,  'W': 183.950900*__amu2au ,  'Re':186.955700*__amu2au ,  
           'Os':191.961500*__amu2au , 'Ir':192.962900*__amu2au , 'Pt':194.964800*__amu2au ,  'Au':196.966500*__amu2au ,  'Hg':201.970600*__amu2au ,
           'Tl':204.974400*__amu2au , 'Pb':207.976600*__amu2au , 'Bi':208.980400*__amu2au ,  'Po':208.982400*__amu2au ,  'At':209.987100*__amu2au ,  
           'Rn':222.017600*__amu2au , 'Fr':223.019700*__amu2au , 'Ra':226.025400*__amu2au ,  'Ac':227.027800*__amu2au ,  'Th':232.038100*__amu2au ,  
           'Pa':231.035900*__amu2au , 'U': 238.050800*__amu2au , 'Np':237.048200*__amu2au ,  'Pu':244.064200*__amu2au ,  'Am':243.061400*__amu2au ,  
           'Cm':247.070300*__amu2au , 'Bk':247.070300*__amu2au , 'Cf':251.079600*__amu2au ,  'Es':252.082900*__amu2au ,  'Fm':257.075100*__amu2au ,
           'Md':258.098600*__amu2au , 'No':259.100900*__amu2au , 'Lr':260.105300*__amu2au 
           }
    #
    # dictionary of mass
    #
    __symbol_dict   =   {                                             
             1:'H' ,    2:'He' ,   3:'Li' ,   4:'Be' ,   5:'B' , 
             6:'C' ,    7:'N' ,    8:'O' ,    9:'F' ,   10:'Ne' , 
            11:'Na' ,  12:'Mg' ,  13:'Al' ,  14:'Si' ,  15:'P' , 
            16:'S' ,   17:'Cl' ,  18:'Ar' ,  19:'K' ,   20:'Ca' , 
            21:'Sc' ,  22:'Ti' ,  23:'V' ,   24:'Cr' ,  25:'Mn' , 
            26:'Fe' ,  27:'Co' ,  28:'Ni' ,  29:'Cu' ,  30:'Zn' , 
            31:'Ga' ,  32:'Ge' ,  33:'As' ,  34:'Se' ,  35:'Br' , 
            36:'Kr' ,  37:'Rb' ,  38:'Sr' ,  39:'Y' ,   40:'Zr' , 
            41:'Nb' ,  42:'Mo' ,  43:'Tc' ,  44:'Ru' ,  45:'Rh' , 
            46:'Pd' ,  47:'Ag' ,  48:'Cd' ,  49:'In' ,  50:'Sn' , 
            51:'Sb' ,  52:'Te' ,  53:'I' ,   54:'Xe' ,  55:'Cs' , 
            56:'Ba' ,  57:'La' ,  58:'Ce' ,  59:'Pr' ,  60:'Nd' , 
            61:'Pm' ,  62:'Sm' ,  63:'Eu' ,  64:'Gd' ,  65:'Tb' , 
            66:'Dy' ,  67:'Ho' ,  68:'Er' ,  69:'Tm' ,  70:'Yb' , 
            71:'Lu' ,  72:'Hf' ,  73:'Ta' ,  74:'W' ,   75:'Re' , 
            76:'Os' ,  77:'Ir' ,  78:'Pt' ,  79:'Au' ,  80:'Hg' , 
            81:'Tl' ,  82:'Pb' ,  83:'Bi' ,  84:'Po' ,  85:'At' , 
            86:'Rn' ,  87:'Fr' ,  88:'Ra' ,  89:'Ac' ,  90:'Th' , 
            91:'Pa' ,  92:'U' ,   93:'Np' ,  94:'Pu' ,  95:'Am' , 
            96:'Cm' ,  97:'Bk' ,  98:'Cf' ,  99:'Es' , 100:'Fm' , 
           101:'Md' , 102:'No' , 103:'Lr'                         
          }

    #
    # dictionary of unit
    #
    __unit_dict   =   {
        "eV":               __eV2au,
        "kcal/mol":         __kcalmol2au,
        "a.u.":             1,
        "kJ/mol":           __kJmol2au,
        "cm-1":             __cm2au,
        "torr":             __torr2au,
        "atm":              __atm2au,
        "bar":              __bar2au,
        "bohr":             1,
        "A":                __A2au,
        "K":                __K2au,
        "amu":              __amu2au,
        "s"  :              __s2au,
        "mol/L":            __molL2au,
        "cm3/molecule/s":   (__cm2m*__m2au)**3 / __s2au,  
        "1/s":              1 / __s2au
    }

    @property
    def  h            (self):    return Const_class.__h            
    @property
    def  kb           (self):    return Const_class.__kb              
    @property
    def  NA           (self):    return Const_class.__NA            
    @property
    def  hba          (self):    return Const_class.__hba          
    @property
    def  amu2kg       (self):    return Const_class.__amu2kg   
    @property
    def  kg2amu       (self):    return Const_class.__kg2amu
    @property
    def  gmol2kg      (self):    return Const_class.__gmol2kg        
    @property
    def  amu2au       (self):    return Const_class.__amu2au    
    @property
    def  au2kg        (self):    return Const_class.__au2kg
    @property
    def  kg2au        (self):    return Const_class.__kg2au
    @property
    def  au2amu       (self):    return Const_class.__au2amu                 
    @property
    def  cm2J         (self):    return Const_class.__cm2J              
    @property
    def  kcalmol2kJmol(self):    return Const_class.__kcalmol2kJmol    
    @property
    def  cm2Jmol      (self):    return Const_class.__cm2Jmol                 
    @property
    def  cm2kJ        (self):    return Const_class.__cm2kJ                 
    @property
    def  kcalmol2cm   (self):    return Const_class.__kcalmol2cm         
    @property
    def  kcal2kJ      (self):    return Const_class.__kcal2kJ               
    @property
    def  cm2kcalmol   (self):    return Const_class.__cm2kcalmol         
    @property
    def  kcal2J       (self):    return Const_class.__kcal2J               
    @property
    def  kJ2kcal      (self):    return Const_class.__kJ2kcal              
    @property
    def  J2kcal       (self):    return Const_class.__J2kcal               
    @property
    def  J2cm         (self):    return Const_class.__J2cm                   
    @property
    def  Jmol2cm      (self):    return Const_class.__Jmol2cm                
    @property
    def  kJmol2cm     (self):    return Const_class.__kJmol2cm              
    @property
    def  kcalmol2J    (self):    return Const_class.__kcalmol2J              
    @property
    def  kcalmol2Jmol (self):    return Const_class.__kcalmol2Jmol            
    @property
    def  eV2au        (self):    return Const_class.__eV2au                    
    @property
    def  au2J         (self):    return Const_class.__au2J                   
    @property
    def  au2kcal      (self):    return Const_class.__au2kcal                 
    @property
    def  au2kcalmol   (self):    return Const_class.__au2kcalmol           
    @property
    def  au2cal       (self):    return Const_class.__au2cal               
    @property
    def  au2kJ        (self):    return Const_class.__au2kJ                
    @property
    def  au2cm        (self):    return Const_class.__au2cm                
    @property
    def  au2kJmol     (self):    return Const_class.__au2kJmol            
    @property
    def  au2eV        (self):    return Const_class.__au2eV               
    @property
    def  cm2au        (self):    return Const_class.__cm2au                
    @property
    def  J2au         (self):    return Const_class.__J2au                 
    @property
    def  kcal2au      (self):    return Const_class.__kcal2au               
    @property
    def  kcalmol2au   (self):    return Const_class.__kcalmol2au           
    @property
    def  cal2au       (self):    return Const_class.__cal2au              
    @property
    def  kJ2au        (self):    return Const_class.__kJ2au                
    @property
    def  kJmol2au     (self):    return Const_class.__kJmol2au             
    @property
    def  au2A         (self):    return Const_class.__au2A                
    @property
    def  A2au         (self):    return Const_class.__A2au                         
    @property
    def  A2m          (self):    return Const_class.__A2m                       
    @property
    def  m2cm         (self):    return Const_class.__m2cm  
    @property
    def  m2au         (self):    return Const_class.__m2au 
    @property
    def  cm2m         (self):    return Const_class.__cm2m      
    @property
    def  au2m         (self):    return Const_class.__au2m                 
    @property
    def  torr2Pa      (self):    return Const_class.__torr2Pa                      
    @property
    def  Pa2torr      (self):    return Const_class.__Pa2torr                       
    @property
    def  atm2Pa       (self):    return Const_class.__atm2Pa                            
    @property
    def  bar2Pa       (self):    return Const_class.__bar2Pa                      
    @property
    def  Pa2au        (self):    return Const_class.__Pa2au                             
    @property
    def  torr2au      (self):    return Const_class.__torr2au  
    @property
    def  au2torr      (self):    return Const_class.__au2torr                   
    @property
    def  atm2au       (self):    return Const_class.__atm2au                    
    @property
    def  bar2au       (self):    return Const_class.__bar2au                     
    @property
    def  au2K         (self):    return Const_class.__au2K                      
    @property
    def  K2au         (self):    return Const_class.__K2au                  
    @property
    def  au2s         (self):    return Const_class.__au2s              
    @property
    def  s2au         (self):    return Const_class.__s2au                      
    @property
    def  mass_dict    (self):    return Const_class.__mass_dict   
    @property
    def  symbol_dict  (self):    return Const_class.__symbol_dict                
    @property
    def  unit_dict    (self):    return Const_class.__unit_dict                 
    @property
    def  STR          (self):    return Const_class.__str  
    @property
    def  FLOAT        (self):    return Const_class.__float
    @property
    def  ARR          (self):    return Const_class.__arr
    @property
    def  ELEV         (self):    return Const_class.__elev
    @property
    def  GEOM         (self):    return Const_class.__geom 
    @property
    def  BOOL         (self):    return Const_class.__bool 



cst             =   Const_class()
