#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#


import  numpy   as      np
from    scipy   import  interpolate
import  numba   as      nb

@nb.jit(nopython=True)
def Repatch(data, index, errorf = print ):
    'repatch the data[index] by linear function'
    L=len(data)
    for i,x in enumerate(index):
        x_i=x_j=x
        I=i
        if x==0:
            while x_i == index[I]:
                x_i += 1
                I += 1
                if x_i >=L:
                    errorf(" Error when repatching dos ... ")
                    return False
            x_j=x_i+1
            while x_j == index[I]:
                x_j += 1
                I += 1
                if x_j >=L:
                    errorf(" Error when repatching dos ... ")
                    return False
            data[x]=data[x_i]-x_i*(data[x_j]-data[x_i])/(x_j-x_i)
        elif x==L-1:
            data[x]=2*data[x-1]-data[x-2]
        else:
            while x_j == index[I]:
                x_j += 1
                I += 1
                if x_j >=L:
                    errorf(" Error when repatching dos ... ")
                    return False
            data[x]=data[x-1]+(data[x_j]-data[x-1])/(x_j-x+1)
    return True


def CheckNaN( Arr, errorf = print ):
    index = np.argwhere( np.isnan(Arr)==True )
    index = index.reshape( index.size )
    if index.size > 0 :
        if not Repatch( Arr, index ):
            exit(1)

class SplineFit :
    'Fit the functopn using the spline method'
    def __init__(self,x,y,LOG):
        'x,y is list'
        if LOG == 'no' or LOG == 'NO':
            self.__fpp    =   interpolate.splrep(x,y) 
            self.__base   =   'no'
        elif LOG == 'lg' or LOG== 'LG' or LOG == 'lg10' or LOG == 'LOG10':
            self.__fpp    =   interpolate.splrep( x, np.log10(np.array(y)) )
            self.__base   =   'lg'
        elif LOG == 'ln' or LOG=='LN':
            self.__fpp    =    interpolate.splrep( x, np.log(np.array(y)) )
            self.__base   =    'ln'
        else:
            print("Fitting class init wrong!")

    def val(self,x):
        'return the true value of  f(x0) for a given x0' 
        if   self.__base == 'no':
            return interpolate.splev( x, self.__fpp )
        elif self.__base == 'lg':
            return np.power( 10, interpolate.splev(x,self.__fpp) )
        elif self.__base == 'ln':
            return np.exp( interpolate.splev(x,self.__fpp) )