#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import numpy as np
import numpy.ctypeslib as npct
import ctypes


def ME_dd( gvr ):
    class __well(ctypes.Structure):
        _fields_=[
            ('dos',         ctypes.POINTER(ctypes.c_double)),
            ('dos_sz',      ctypes.c_int32),
            ('dos_t',       ctypes.POINTER(ctypes.c_double)),
            ('E0',          ctypes.c_double),
            ('E0_t',        ctypes.c_double),
            ('size',        ctypes.c_int32),
            ('Z',           ctypes.c_double) ]
    class __kE(ctypes.Structure):
        _fields_=[
            ('kE',          ctypes.POINTER(ctypes.c_double)),
            ('kE_sz',       ctypes.c_int32),
            ('ph1_index',   ctypes.c_int32),
            ('ph2_index',   ctypes.c_int32)]
    class __param(ctypes.Structure):
        _fields_=[
            ('slope',       ctypes.c_double),
            ('EdConst',     ctypes.c_double),
            ('Tc',          ctypes.c_double),
            ('exponent',    ctypes.c_double),
            ('dE',          ctypes.c_double),
            ('ener_step',   ctypes.c_double),
            ('ener_ref',    ctypes.c_double),
            ('beta',        ctypes.c_double)]

    #set parameters
    ener_ref      = gvr.cycle_param.ener_ref
    ener_step     = gvr.cycle_param.ener_step
    Nwell         = np.int32(len( gvr.well_stack  ))
    Nph           = np.int32(len( gvr.phase_stack ))
    ompnumthread  = np.int32(gvr.ompnumthread) 

    #set well_stack_dd
    well_stack_dd = ( __well * Nwell )()
    dos_stack     = [] 
    for i,well in enumerate(gvr.well_stack):
        dos_stack.append(   np.array( well.dos,  dtype=ctypes.c_double )   )
        well_stack_dd[ i ].dos       =   dos_stack[ i ].ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        well_stack_dd[ i ].dos_sz    =   len(dos_stack[ i ])
        well_stack_dd[ i ].E0        =   well.E0
        well_stack_dd[ i ].E0_t      =   well.E0 + ( ener_ref - well.E0 ) % ener_step
        well_stack_dd[ i ].size      =   int( (ener_ref-well.E0)//ener_step )
        well_stack_dd[ i ].Z         =   well.collid.Z

    #set param_dd
    param_dd           =   __param()
    param_dd.slope     =   gvr.slope
    param_dd.EdConst   =   gvr.EdConst
    param_dd.Tc        =   gvr.Tc
    param_dd.exponent  =   gvr.exponent
    param_dd.dE        =   gvr.dE
    param_dd.ener_step =   ener_step
    param_dd.ener_ref  =   ener_ref
    param_dd.beta      =   1 / gvr.cycle_param.T

    #set size of kE_stack_dd
    kE_stack_sz   = 0
    for rxn in gvr.rxn_stack:
        if rxn.rxn_type == 'w2w' or rxn.rxn_type == 'w2b':
            kE_stack_sz += 1
    #init kE_stack
    kE_stack_dd   = ( __kE   * kE_stack_sz )()
    #set kE_stack_dd
    i = 0
    kE_stack           =   []
    for well_i in range( Nwell ):
        for ph_j in range( Nph ):
            if gvr.rxn_map[well_i][ph_j]:
                kE_stack.append(   np.array( gvr.rxn_map[well_i][ph_j].kE, dtype=ctypes.c_double )    )
                kE_stack_dd[ i ].kE           =   kE_stack[i].ctypes.data_as(ctypes.POINTER(ctypes.c_double))
                kE_stack_dd[ i ].kE_sz        =   len(kE_stack[i])
                kE_stack_dd[ i ].ph1_index    =   well_i
                kE_stack_dd[ i ].ph2_index    =   ph_j
                i                            +=   1

    #set mat_size
    mat_size = np.sum( [ x.size for x in gvr.well_stack  ] )

    #set eigen pair
    evals    = np.zeros( ( mat_size ),           dtype=ctypes.c_double )
    evecs    = np.zeros( ( mat_size, mat_size ), dtype=ctypes.c_double )

    #call ME_dd lib
    libcd                    =      npct.load_library("libMEdd",gvr.root_path+"/lib/")
    libcd.ME_dd_Main.argtypes=      [ ctypes.c_int32,
                                      ctypes.POINTER(__well),  ctypes.c_int32,
                                      ctypes.POINTER(__kE),    ctypes.c_int32,
                                      ctypes.POINTER(__param), ctypes.c_int32,
                                      npct.ndpointer(dtype=ctypes.c_double,shape=(mat_size,),        flags='C_CONTIGUOUS'),
                                      npct.ndpointer(dtype=ctypes.c_double,shape=(mat_size,mat_size),flags='C_CONTIGUOUS')]
    libcd.ME_dd_Main( 
        ompnumthread,
        well_stack_dd, 
        Nwell,
        kE_stack_dd,
        kE_stack_sz,
        param_dd,
        Nph, 
        evals,
        evecs
        )
    
    return evals, np.transpose(evecs)



def pseudo_ME_dd( gvr ):
    class __well(ctypes.Structure):
        _fields_=[
            ('dos',         ctypes.POINTER(ctypes.c_double)),
            ('dos_sz',      ctypes.c_int32),
            ('dos_t',       ctypes.POINTER(ctypes.c_double)),
            ('E0',          ctypes.c_double),
            ('E0_t',        ctypes.c_double),
            ('size',        ctypes.c_int32),
            ('Z',           ctypes.c_double) ]
    class __kE(ctypes.Structure):
        _fields_=[
            ('kE',          ctypes.POINTER(ctypes.c_double)),
            ('kE_sz',       ctypes.c_int32),
            ('ph1_index',   ctypes.c_int32),
            ('ph2_index',   ctypes.c_int32)]
    class __dk(ctypes.Structure):
        _fields_=[
            ('dk',          ctypes.POINTER(ctypes.c_double)),
            ('dk_sz',       ctypes.c_int32),
            ('ph1_index',   ctypes.c_int32),
            ('ph2_index',   ctypes.c_int32)]
    class __param(ctypes.Structure):
        _fields_=[
            ('slope',       ctypes.c_double),
            ('EdConst',     ctypes.c_double),
            ('Tc',          ctypes.c_double),
            ('exponent',    ctypes.c_double),
            ('dE',          ctypes.c_double),
            ('ener_step',   ctypes.c_double),
            ('ener_ref',    ctypes.c_double),
            ('beta',        ctypes.c_double)]
    class __pseubim(ctypes.Structure):
        _fields_=[
            ('index',       ctypes.c_int32),
            ('excessconc',  ctypes.c_double)
        ]


    #set parameters
    ener_ref      = gvr.cycle_param.ener_ref  
    ener_step     = gvr.cycle_param.ener_step
    Nwell         = np.int32(len( gvr.well_stack  ))
    Nph           = np.int32(len( gvr.phase_stack ))
    ompnumthread  = np.int32(gvr.ompnumthread) 

    #set well_stack_dd
    well_stack_dd = ( __well * Nwell )()
    dos_stack     = [] 
    for i,well in enumerate(gvr.well_stack):
        dos_stack.append(   np.array( well.dos,  dtype=ctypes.c_double )   )
        well_stack_dd[ i ].dos       =   dos_stack[ i ].ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        well_stack_dd[ i ].dos_sz    =   len(dos_stack[ i ])
        well_stack_dd[ i ].E0        =   well.E0
        well_stack_dd[ i ].E0_t      =   well.E0 + ( ener_ref - well.E0 ) % ener_step
        well_stack_dd[ i ].size      =   int( (ener_ref-well.E0)//ener_step )
        well_stack_dd[ i ].Z         =   well.collid.Z

    #set param_dd
    param_dd           =   __param()
    param_dd.slope     =   gvr.slope
    param_dd.EdConst   =   gvr.EdConst
    param_dd.Tc        =   gvr.Tc
    param_dd.exponent  =   gvr.exponent
    param_dd.dE        =   gvr.dE
    param_dd.ener_step =   ener_step
    param_dd.ener_ref  =   ener_ref
    param_dd.beta      =   1 / gvr.cycle_param.T

    #set size of kE_stack_dd
    kE_stack_sz   = 0
    for rxn in gvr.rxn_stack:
        if rxn.rxn_type == 'w2w' or rxn.rxn_type == 'w2b':
            kE_stack_sz += 1
    #init kE_stack
    kE_stack_dd   = ( __kE   * kE_stack_sz )()
    #set kE_stack_dd
    i = 0
    kE_stack           =   []
    for ph_i in range( Nwell ):
        for ph_j in range( Nph ):
            if gvr.rxn_map[ph_i][ph_j]:
                kE_stack.append(   np.array( gvr.rxn_map[ph_i][ph_j].kE, dtype=ctypes.c_double )    )
                kE_stack_dd[ i ].kE           =   kE_stack[i].ctypes.data_as(ctypes.POINTER(ctypes.c_double))
                kE_stack_dd[ i ].kE_sz        =   len(kE_stack[i])
                kE_stack_dd[ i ].ph1_index    =   ph_i
                kE_stack_dd[ i ].ph2_index    =   ph_j
                i                            +=   1
    
    #set size of dk_stack_dd
    dk_stack_sz   = 0
    for rxn in gvr.rxn_stack:
        if rxn.rxn_type == 'b2w':
            dk_stack_sz += 1
    #init dk_stack
    dk_stack_dd   = ( __dk   * dk_stack_sz )()
    #set dk_stack_dd
    i             = 0
    dk_stack      = []
    for ph_i in range( Nwell, Nph ): 
        for ph_j in range( Nph ):
            if gvr.rxn_map[ph_i][ph_j]:
                dk_stack.append(   np.array( gvr.rxn_map[ph_i][ph_j].dk, dtype=ctypes.c_double )    )
                dk_stack_dd[ i ].dk           =   dk_stack[i].ctypes.data_as(ctypes.POINTER(ctypes.c_double))
                dk_stack_dd[ i ].dk_sz        =   len(dk_stack[i])
                dk_stack_dd[ i ].ph1_index    =   ph_i - Nwell
                dk_stack_dd[ i ].ph2_index    =   ph_j
                i                            +=   1


    #set pseudobim_stack
    pseubimname        =   gvr.pseubimname
    Npseubim           =   len(pseubimname)
    bim_stack          =   gvr.bim_stack
    bim_name           =   [ x.name for x in bim_stack ]
    pseubim_stack_dd   =   ( __pseubim  * Npseubim )()
    for i in range(Npseubim):
        pseubim_stack_dd[i].index       =   bim_name.index( pseubimname[i] ) 
        pseubim_stack_dd[i].excessconc  =   gvr.excessconc[i]


    #set mat_size
    mat_size  =  np.sum( [ x.size for x in gvr.well_stack  ] )
    mat_size += len(pseubimname)

    #set eigen pair
    evals     =  np.zeros( ( mat_size ),           dtype=ctypes.c_double )
    evecs     =  np.zeros( ( mat_size, mat_size ), dtype=ctypes.c_double )

    #call ME_dd lib
    libcd                           =       npct.load_library("libMEdd",gvr.root_path+"/lib/")  
    libcd.pseuME_dd_Main.argtypes   =     [ ctypes.c_int32,
                                            ctypes.POINTER(__well),                 ctypes.c_int32,
                                            ctypes.POINTER(__kE),                   ctypes.c_int32,
                                            ctypes.POINTER(__dk),                   ctypes.c_int32,
                                            ctypes.POINTER(__param),                ctypes.c_int32,
                                            ctypes.POINTER(__pseubim),              ctypes.c_int32,
                                            npct.ndpointer(dtype=ctypes.c_double,   shape=(mat_size,),        flags='C_CONTIGUOUS'),
                                            npct.ndpointer(dtype=ctypes.c_double,   shape=(mat_size,mat_size),flags='C_CONTIGUOUS')]
    libcd.pseuME_dd_Main( 
        ompnumthread,
        well_stack_dd,      Nwell,
        kE_stack_dd,        kE_stack_sz,
        dk_stack_dd,        dk_stack_sz,
        param_dd,           Nph,
        pseubim_stack_dd,   Npseubim, 
        evals,              evecs
        )
    
    return evals, np.transpose(evecs)
