#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



class CycleParam_class:
    'class for temperature and pressure specific parameters'
    __T         =      0.0
    __p         =      0.0
    __ener_step =      0.0
    __ener_ref  =      0.0
    
    def set_T(self,x):                      CycleParam_class.__T          = x        
    def set_p(self,x):                      CycleParam_class.__p          = x            
    def set_ener_step(self,x):              CycleParam_class.__ener_step  = x 
    def set_ener_ref(self,x):               CycleParam_class.__ener_ref   = x 

    @property 
    def T               (self):             return CycleParam_class.__T
    @property
    def p               (self):             return CycleParam_class.__p
    @property
    def ener_step       (self):             return CycleParam_class.__ener_step
    @property
    def ener_ref        (self):             return CycleParam_class.__ener_ref   


class GlobalVar_class():
    def __init__(self):
        self.__temperature             =   None
        self.__pressure                =   None
        self.__ESOT                    =   0.1                 #energy step over temperature
        self.__EEOT                    =   30                  #max energy over temperature 
        self.__merge_threshold         =   0.2                 #threhold to determine whether merging species
        self.__precision               =   'double'            #precision, available for double, quadruple and octuple 
        self.__ompnumthread            =   None                #OMP_NUM_THREAD by default, it will be os.cpu_count()/nproc  
        self.__dE                      =   4.55634E-06         #energy step for DoS calculation
        self.__Emax                    =   0.637440575         #max energy for DoS calculation
        self.__slope                   =   0                   #Energy dependent <DeltaE_d> = (T/Tc)^(exponent)*[slope * (E-E0,gamma) + EdConst]
        self.__EdConst                 =   None                
        self.__Tc                      =   9.5004462E-4
        self.__exponent                =   0                             
        self.__gndSpecies              =   None                #ground species name 
        self.__collid                  =   None                #collision paramater       
        self.__cycle_param             =   CycleParam_class()  #paramers for each cycle consisting of T,P, ener_step, ener_ref
        self.__root_path               =   None                #root path
        self.__printevalnum            =   None                #number of eigenvalues to be print out
        self.__printevecnum            =   None                #number of eigenvectors to be print out
        self.__printkernelname         =   None                #the name list of the wells of which the relaxation kernels will be printed
        self.__timeevolution_sw        =   False               #switch to determine whether to calculate time-evolution of populations      
        self.__initspecies             =   None                #initial species for time-evolution
        self.__pseubimname             =   None                #names for bimolecular pairs which will be treated under pseudo-first-order condition   
        self.__excessconc              =   None                #concentration of the excessive fragment of pseudo-first-order bimolecular pairs.

        self.__well_stack              =   []
        self.__ts_stack                =   []
        self.__bim_stack               =   []
        self.__phase_stack             =   []
        self.__rxn_stack               =   []
        self.__rxn_map                 =   []
        self.__ME_stack                =   []

        self.__well_dict               =   {}
        self.__ts_dict                 =   {}
        self.__bim_dict                =   {}
        self.__phase_dict              =   {}

        self.__time_step               =   None
        self.__time_num                =   None


    @property   
    def  temperature             (self):   return self.__temperature             
    @property   
    def  pressure                (self):   return self.__pressure                
    @property   
    def  ESOT                    (self):   return self.__ESOT                    
    @property   
    def  EEOT                    (self):   return self.__EEOT                    
    @property   
    def  merge_threshold         (self):   return self.__merge_threshold 
    @property
    def  precision               (self):   return self.__precision   
    @property
    def  ompnumthread            (self):   return self.__ompnumthread
    @property   
    def  dE                      (self):   return self.__dE                      
    @property   
    def  Emax                    (self):   return self.__Emax   
    @property
    def  slope                   (self):   return self.__slope
    @property
    def  EdConst                 (self):   return self.__EdConst    
    @property
    def  Tc                      (self):   return self.__Tc 
    @property
    def  exponent                (self):   return self.__exponent
    @property   
    def  gndSpecies              (self):   return self.__gndSpecies 
    @property   
    def  collid                  (self):   return self.__collid 
    @property
    def  cycle_param             (self):   return self.__cycle_param
    @property
    def  root_path               (self):   return self.__root_path
    @property
    def  printevalnum            (self):   return self.__printevalnum
    @property
    def  printevecnum            (self):   return self.__printevecnum
    @property
    def  printkernelname         (self):   return self.__printkernelname
    @property
    def  timeevolution_sw        (self):   return self.__timeevolution_sw
    @property
    def  initspecies             (self):   return self.__initspecies
    @property
    def  pseubimname             (self):   return self.__pseubimname
    @property
    def  excessconc              (self):   return self.__excessconc

    @property   
    def  well_stack              (self):   return self.__well_stack              
    @property   
    def  ts_stack                (self):   return self.__ts_stack                       
    @property   
    def  bim_stack               (self):   return self.__bim_stack                    
    @property   
    def  phase_stack             (self):   return self.__phase_stack                              
    @property   
    def  rxn_stack               (self):   return self.__rxn_stack  
    @property   
    def  rxn_map                 (self):   return self.__rxn_map           
    @property
    def  ME_stack                (self):   return self.__ME_stack
    @property   
    def  well_dict               (self):   return self.__well_dict  
    @property   
    def  ts_dict                 (self):   return self.__ts_dict 
    @property   
    def  bim_dict                (self):   return self.__bim_dict      
    @property   
    def  phase_dict              (self):   return self.__phase_dict 

    @property
    def  init_well               (self):   return self.__init_well
    @property
    def  time_step               (self):   return self.__time_step
    @property
    def  time_num                (self):   return self.__time_num


    def set_temperature              (self,x):   self.__temperature                 =   x 
    def set_pressure                 (self,x):   self.__pressure                    =   x 
    def set_ESOT                     (self,x):   self.__ESOT                        =   x 
    def set_EEOT                     (self,x):   self.__EEOT                        =   x 
    def set_dE                       (self,x):   self.__dE                          =   x 
    def set_Emax                     (self,x):   self.__Emax                        =   x 
    def set_slope                    (self,x):   self.__slope                       =   x
    def set_EdConst                  (self,x):   self.__EdConst                     =   x
    def set_Tc                       (self,x):   self.__Tc                          =   x
    def set_exponent                 (self,x):   self.__exponent                    =   x
    def set_gndSpecies               (self,x):   self.__gndSpecies                  =   x 
    def set_rxn_map                  (self,x):   self.__rxn_map                     =   x 
    def set_collid                   (self,x):   self.__collid                      =   x
    def set_merge_threshold          (self,x):   self.__merge_threshold             =   x  
    def set_precision                (self,x):   self.__precision                   =   x
    def set_ME_stack                 (self,x):   self.__ME_stack                    =   x
    def set_phase_stack              (self,x):   self.__phase_stack                 =   x
    def set_phase_dict               (self,x):   self.__phase_dict                  =   x

    def set_time_step                (self,x):   self.__time_step                   =   x
    def set_time_num                 (self,x):   self.__time_num                    =   x
    def set_root_path                (self,x):   self.__root_path                   =   x
    def set_evalnum                  (self,x):   self.__printevalnum                =   x
    def set_evecnum                  (self,x):   self.__printevecnum                =   x
    def set_printkernelname          (self,x):   self.__printkernelname             =   x
    def set_ompnumthread             (self,x):   self.__ompnumthread                =   x
    def set_initspecies              (self,x):   self.__initspecies                 =   x
    def set_pseubimname              (self,x):   self.__pseubimname                 =   x
    def set_excessconc               (self,x):   self.__excessconc                  =   x

    def push_well_stack              (self,x):   self.__well_stack.append     (x) 
    def push_ts_stack                (self,x):   self.__ts_stack.append       (x) 
    def push_bim_stack               (self,x):   self.__bim_stack.append      (x) 
    def push_rxn_stack               (self,x):   self.__rxn_stack.append      (x)
    def push_ME_stack                (self,x):   self.__ME_stack.append       (x)
    def push_well_dict               (self,key,x):   self.__well_dict [key]         =   x 
    def push_ts_dict                 (self,key,x):   self.__ts_dict   [key]         =   x 
    def push_bim_dict                (self,key,x):   self.__bim_dict  [key]         =   x 

    def turn_on_time_evolution       (self):      self.__timeevolution_sw           =   True


    def check(self, errorf):
        if len(self.__well_stack)==0:
            errorf ( ' Error for Well. No well defined.')
        if len(self.__rxn_stack)==0:
            errorf ( ' Error for Reaction. No reaction defined.')
        if self.__EdConst  is None:
            errorf ( ' Error for energy-dependent <DeltaE_d>, EdConst should be specifid, where <DeltaE_d>=(T/Tc)[slope*(E-E0,gamma)+EdConst].')
        if self.__EdConst  <= 0:
            errorf ( ' Error for energy-dependent <DeltaE_d>, EdConst should be positive, where <DeltaE_d>=(T/Tc)[slope*(E-E0,gamma)+EdConst].')    
        if self.__Tc <= 0:
            errorf ( ' Error for energy-dependent <DeltaE_d>, Tc should be positive, where <DeltaE_d>=(T/Tc)[slope*(E-E0,gamma)+EdConst].')
        if self.__temperature           is None:
            errorf ( ' Error for temperatures. No value.' )
        if self.__pressure              is None:
            errorf ( ' Error for pressures. No value.' )
        if self.__gndSpecies            is None:
            errorf ( ' Error for ground species. No value.' )
        if self.__collid                is None:
            errorf ( ' Error for Collision. No value.' )
        if self.__collid.name == 'LJ':
            if self.__collid.epsilonM       is None:
                errorf ( ' Error for LJ_Collision.epsilonM. No value' )
            if self.__collid.diameterM      is None:
                errorf ( ' Error for LJ_Collision.diameterM. No value' )
            if self.__collid.massM          is None:
                errorf ( ' Error for LJ_Collision.massM. No value' )
        if self.__collid.name == 'HS':
            if self.__collid.massM          is None:
                errorf ( ' Error for HS_Collision.massM. No value' )
        for x in self.__temperature:
            if x <= 0:
                errorf ( ' Error in temperatures. Values should not be negtive.' )
        for x in self.__pressure:
            if x <= 0:
                errorf ( ' Error in pressures. Values should not be negtive.'    )    
        if self.__EEOT <= 0:
            errorf ( ' Error in EEOT. Values should not be negtive.' )
        if self.__ESOT <= 0:
            errorf ( ' Error in ESOT. Values should not be negtive.' )
        if self.__merge_threshold <= 0:
            errorf ( ' Error in MERGETHRESHOLD. Values should not be negtive.' )
        if self.__dE <= 0:
            errorf ( ' Error in dE. Values should not be negtive.' )
        if self.__Emax <= 0:
            errorf ( ' Error in Emax. Values should not be negtive.' )
        if self.__collid.name == 'LJ':
            if self.__collid.epsilonM <=0:
                errorf ( ' Error in LJ_Collision epsilonM. Values should not be negtive.' )
            if self.__collid.diameterM <=0:
                errorf ( ' Error in LJ_Collision diameterM. Values should not be negtive.' )
            if self.__collid.massM <=0:
                errorf ( ' Error in LJ_Collision massM. Values should not be negtive.' )
        if self.__collid.name == 'HS':
            if self.__collid.bmax <=0:
                errorf ( ' Error in HS_Collision bmax. Values should not be negtive.' )
            if self.__collid.massM <=0:
                errorf ( ' Error in HS_Collision massM. Values should not be negtive.' )
        if self.__gndSpecies not in self.__phase_dict:
            errorf ( ' Error in gound species. It has not been defined in Well and Bimolecular.' )
        if self.__precision != 'double' and self.__precision != 'quadruple' and self.__precision != 'octuple':
            errorf ( ' Error of precision. Should  be \"double\", \"quadruple\", \"octuple\" ' )
        if self.__ompnumthread < 1:
            errorf ( ' Error in ompnumthread. Should be positive integer.' )
        if self.__pseubimname is not None:
            bim_name = [ x.name for x in self.__bim_stack ]
            for x in self.__pseubimname:
                if x not in bim_name:
                    errorf(' Error of bimolecular name '+ x +' in Bimmolname subkeyword in PBimol subsection.')
            if self.__excessconc is not None:
                if len(self.__excessconc) != len(self.__pseubimname):
                    errorf(' Error of PseuBimName and ExcessConoc. They should have the same length.')
            else:
                errorf(' Error, ExcessConc should be defined.')
        if self.__excessconc is not None:
            if self.__pseubimname is None:
                errorf(' Error, PseuBimName should be defined.')
        if self.__printevalnum is not None:
            if self.__printevalnum < 1:
                errorf( ' Error of number of eigenvalues is negative. should be positive interger')
            self.__printevalnum = int(self.__printevalnum)
        if self.__printevecnum is not None:
            if self.__printevecnum < 1:
                errorf( ' Error of number of eigenvectors is negative. should be positive interger')
            self.__printevecnum = int(self.__printevecnum)
        if self.__printkernelname is not None:
            well_name = [ x.name for x in self.__well_stack ]
            for x in self.__printkernelname:
                if x not in well_name:
                    errorf(' Error of isomer name '+ x +' in printkernelname subkeyword in RelaxKernel subsection.')
        if self.__timeevolution_sw:
            if self.__initspecies is not None:
                if self.__pseubimname is not None:
                    config_name = [ x.name for x in self.__well_stack ] + self.__pseubimname
                else:
                    config_name = [ x.name for x in self.__well_stack ]
                for x in self.__initspecies:
                    if x not in config_name:
                        errorf(' Error of species name '+ x +' in InitSpecies subkeyword. Should be Well or PseuBimolecular.')
            else:
                errorf(' Error, InitSpecies should be specified when TimeEvolution is defined.')
            if self.__time_step is not None:
                if self.__time_step < 0:
                    errorf(' Error, the time step should be positive.')
            if self.__time_num is not None:
                if self.__time_num < 1:
                    errorf(' Error, the number of time steps shoul be greater than 1.')
        


class Default_class():
    __mst_file      =   None
    __pyr_file      =   None
    __variation     =   'tst'
    __tunnel        =   None
    __g09_file      =   None
    __rotsigma      =   1
    __opticalnum    =   1
    __freqscale     =   1
    __ele_level     =   [[1,0]]
    __name          =   None
    __symbol        =   None        
    __geom          =   None    
    __freq          =   None    
    __e_type        =   None    
    __e             =   None  
    __diameter      =   None
    __epsilon       =   None
    __bmax          =   None
    __s_lower       =   None
    __s_upper       =   None
    __source        =   "SPECIES block"
    __invlaplace    =   None
    __expression    =   None
    __A1            =   None
    __T1            =   None
    __n1            =   None
    __Ea1           =   None
    __A2            =   None
    __T2            =   None
    __n2            =   None
    __Ea2           =   None

    @property
    def mst_file (self):    return Default_class.__mst_file  
    @property
    def pyr_file (self):    return Default_class.__pyr_file  
    @property
    def variation (self):   return Default_class.__variation
    @property
    def tunnel (self):      return Default_class.__tunnel
    @property
    def rotsigma (self):    return Default_class.__rotsigma
    @property
    def opticalnum(self):   return Default_class.__opticalnum
    @property
    def freqscale(self):    return Default_class.__freqscale
    @property
    def ele_level (self):   return Default_class.__ele_level
    @property
    def name     (self):    return Default_class.__name
    @property
    def g09_file (self):    return Default_class.__g09_file
    @property
    def symbol  (self):     return Default_class.__symbol
    @property
    def geom    (self):     return Default_class.__geom  
    @property
    def freq    (self):     return Default_class.__freq  
    @property
    def e_type  (self):     return Default_class.__e_type
    @property
    def e       (self):     return Default_class.__e     
    @property
    def diameter(self):     return Default_class.__diameter
    @property
    def epsilon (self):     return Default_class.__epsilon
    @property
    def bmax    (self):     return Default_class.__bmax
    @property
    def s_lower (self):     return Default_class.__s_lower
    @property
    def s_upper (self):     return Default_class.__s_upper
    @property
    def source  (self):     return Default_class.__source
    @property
    def invlaplace(self):   return Default_class.__invlaplace
    @property
    def expression (self):  return Default_class.__expression
    @property
    def A1         (self):  return Default_class.__A1        
    @property
    def T1         (self):  return Default_class.__T1     
    @property
    def n1         (self):  return Default_class.__n1        
    @property
    def Ea1        (self):  return Default_class.__Ea1       
    @property
    def A2         (self):  return Default_class.__A2        
    @property
    def T2         (self):  return Default_class.__T2     
    @property
    def n2         (self):  return Default_class.__n2        
    @property
    def Ea2        (self):  return Default_class.__Ea2       


class Key_class():
    #global parameter key
    __param          =   "parameter"
    __LJcollid       =   "ljcollision"
    __HScollid       =   "hscollision"
    __epsilonM       =   "epsilonm"
    __epsilonA       =   "epsilona"
    __diameterM      =   "diameterm"
    __diameterA      =   "diametera"
    __massM          =   "massm"
    __massA          =   "massa"
    __bmax           =   "bmax"
    __Edown          =   "edown"
    __slope          =   "slope"
    __EdConst        =   "edconst"
    __Tc             =   "tc"
    __exponent       =   "exponent"
    __temperature    =   "temperature"
    __pressure       =   "pressure"
    __energy         =   "energy"
    __ESOT           =   "esot"
    __EEOT           =   "eeot"
    __dE             =   "de"
    __Emax           =   "emax"
    __gndSpecies     =   "groundspecies"
    __pseubimolecular=   "pbimol"
    __pseubimname    =   "bimolname"
    __excessconc     =   "excessconc"
    __prints         =   "print"
    __evals          =   "eigenvalue"     
    __evalfile       =   "evalfile"
    __evalnum        =   "evalnum"
    __evecs          =   "eigenvector"
    __evecfile       =   "evecfile"
    __evecnum        =   "evecnum"
    __partifunc      =   "partitionfunction"
    __kE             =   "microfluxcoefficient"
    __relaxkernel    =   "relaxkernel"
    __relaxfile      =   "relaxfile"
    __printkernelname=   "kernelname"
    __timeEvolution  =   "timeevolution"
    __time_file      =   "timefile"
    __time_step      =   "timestep"
    __time_num       =   "timenum"
    __initspecies    =   "initspecies"
    __merge_threshold=   "mergethreshold"
    __precision      =   "precision"
    __ompnumthread   =   "ompnumthread"

    #species key
    __species        =   "species"
    __well           =   "well"
    __tranState      =   "transtate"
    __bim            =   "bim"
    __sp1_mol        =   "sp1mol"
    __sp2_mol        =   "sp2mol"
    __sp1_atom       =   "sp1atom"
    __sp2_atom       =   "sp2atom"
    __symbol         =   "symbol"
    __name           =   "name"
    __geom           =   "geometry"
    __freq           =   "frequency"
    __E0K            =   "e0k"
    __Eele           =   "eele"
    __ele_level      =   "elelevel"
    __rotsigma       =   "rotsigma"
    __mst_file       =   "mstfile"
    __g09_file       =   "g09file"
    __image_freq     =   "imagfreq"
    __diameter       =   "diameter"
    __epsilon        =   "epsilon"
    __bmax           =   "bmax"
    __opticalnum     =   "opticalnum"
    __freqscale      =   "freqscale"

    #reaction parameter
    __reaction       =   "reaction"
    __elem_rxn       =   "barrierrxn"
    __info           =   "info"
    __pyr_file       =   "pyrfile"
    __variation      =   "variation"
    __tunnel         =   "tunneling"
    __barrierlessrxn =   "barrierlessrxn"
    __rxn_ener       =   "rxnenergy"
    __avgdiameter    =   "avgdiameter" 
    __s_lower        =   "slower"
    __s_upper        =   "supper"
    __invlaplace     =   "invlaplace"
    __expression     =   "expression"
    __A1             =   "prefactor1"
    __T1             =   "t1"
    __n1             =   "n1"
    __Ea1            =   "ea1"
    __A2             =   "prefactor2"
    __T2             =   "t2"
    __n2             =   "n2"
    __Ea2            =   "ea2"

 
    @property
    def param       (self): return Key_class.__param    
    @property
    def LJcollid    (self): return Key_class.__LJcollid
    @property
    def HScollid    (self): return Key_class.__HScollid       
    @property
    def epsilonM    (self): return Key_class.__epsilonM     
    @property
    def epsilonA    (self): return Key_class.__epsilonA     
    @property
    def diameterM   (self): return Key_class.__diameterM       
    @property
    def diameterA   (self): return Key_class.__diameterA       
    @property
    def massM       (self): return Key_class.__massM        
    @property
    def massA       (self): return Key_class.__massA      
    @property
    def bmax        (self): return Key_class.__bmax  
    @property
    def Edown       (self): return Key_class.__Edown 
    @property
    def slope       (self): return Key_class.__slope
    @property
    def EdConst     (self): return Key_class.__EdConst
    @property
    def Tc          (self): return Key_class.__Tc
    @property
    def exponent    (self): return Key_class.__exponent
    @property
    def temperature (self): return Key_class.__temperature  
    @property
    def pressure    (self): return Key_class.__pressure     
    @property
    def energy      (self): return Key_class.__energy       
    @property
    def ESOT        (self): return Key_class.__ESOT         
    @property
    def EEOT        (self): return Key_class.__EEOT         
    @property
    def dE          (self): return Key_class.__dE           
    @property
    def Emax        (self): return Key_class.__Emax         
    @property
    def gndSpecies  (self): return Key_class.__gndSpecies 
    @property
    def merge_threshold(self): return  Key_class.__merge_threshold
    @property
    def precision   (self): return Key_class.__precision
    @property
    def ompnumthread(self): return Key_class.__ompnumthread
    @property 
    def pseubimolecular(self): return Key_class.__pseubimolecular
    @property
    def pseubimname  (self): return Key_class.__pseubimname    
    @property 
    def excessconc  (self): return Key_class.__excessconc  
    @property
    def prints      (self): return Key_class.__prints   
    @property
    def evals       (self): return Key_class.__evals
    @property
    def evalfile    (self): return Key_class.__evalfile
    @property
    def evalnum     (self): return Key_class.__evalnum
    @property
    def evecs       (self): return Key_class.__evecs
    @property
    def evecfile    (self): return Key_class.__evecfile
    @property
    def evecnum     (self):  return Key_class.__evecnum
    @property
    def partifunc   (self): return Key_class.__partifunc
    @property
    def relaxkernel (self): return Key_class.__relaxkernel
    @property
    def relaxfile   (self): return Key_class.__relaxfile
    @property
    def printkernelname(self): return Key_class.__printkernelname
    @property
    def kE          (self): return Key_class.__kE
    @property
    def timeEvolution(self):return Key_class.__timeEvolution
    @property
    def time_file   (self): return Key_class.__time_file
    @property
    def time_step   (self): return Key_class.__time_step     
    @property
    def time_num    (self): return Key_class.__time_num     
    @property 
    def initspecies (self): return Key_class.__initspecies   
    @property
    def species     (self): return Key_class.__species      
    @property
    def well        (self): return Key_class.__well         
    @property
    def tranState   (self): return Key_class.__tranState    
    @property
    def bim         (self): return Key_class.__bim          
    @property
    def sp1_mol     (self): return Key_class.__sp1_mol          
    @property
    def sp2_mol     (self): return Key_class.__sp2_mol          
    @property
    def sp1_atom    (self): return Key_class.__sp1_atom     
    @property
    def sp2_atom    (self): return Key_class.__sp2_atom     
    @property
    def symbol      (self): return Key_class.__symbol             
    @property
    def name        (self): return Key_class.__name         
    @property
    def geom        (self): return Key_class.__geom         
    @property
    def freq        (self): return Key_class.__freq         
    @property
    def E0K         (self): return Key_class.__E0K          
    @property
    def Eele        (self): return Key_class.__Eele         
    @property
    def ele_level   (self): return Key_class.__ele_level     
    @property
    def rotsigma    (self): return Key_class.__rotsigma         
    @property
    def mst_file    (self): return Key_class.__mst_file     
    @property
    def g09_file    (self): return Key_class.__g09_file   
    @property
    def image_freq  (self): return Key_class.__image_freq 
    @property
    def diameter    (self): return Key_class.__diameter
    @property
    def epsilon     (self): return Key_class.__epsilon
    @property
    def opticalnum  (self): return Key_class.__opticalnum
    @property
    def freqscale   (self): return Key_class.__freqscale

    @property
    def reaction    (self): return Key_class.__reaction     
    @property
    def elem_rxn    (self): return Key_class.__elem_rxn     
    @property
    def info        (self): return Key_class.__info         
    @property
    def pyr_file    (self): return Key_class.__pyr_file     
    @property
    def variation   (self): return Key_class.__variation    
    @property
    def tunnel      (self): return Key_class.__tunnel 
    @property
    def barrierlessrxn (self): return Key_class.__barrierlessrxn
    @property
    def rxn_ener       (self): return Key_class.__rxn_ener      
    @property
    def avgdiameter (self): return Key_class.__avgdiameter
    @property
    def s_lower     (self): return Key_class.__s_lower
    @property
    def s_upper     (self): return Key_class.__s_upper
    @property
    def invlaplace(self): return Key_class.__invlaplace 
    @property
    def expression(self): return Key_class.__expression           
    @property
    def A1        (self): return Key_class.__A1             
    @property
    def T1        (self): return Key_class.__T1                   
    @property
    def n1        (self): return Key_class.__n1       
    @property
    def Ea1       (self): return Key_class.__Ea1
    @property
    def A2        (self): return Key_class.__A2             
    @property
    def T2        (self): return Key_class.__T2                   
    @property
    def n2        (self): return Key_class.__n2       
    @property
    def Ea2       (self): return Key_class.__Ea2             

    def printInputKeywords():
        string  = '\nThe available keywords for Input file are : \n'
        string += Key_class.__param + '\n'
        string += Key_class.__species + '\n'
        string += Key_class.__reaction + '\n'
        return string

    def printParameterKeywords():
        string  =   '\nThe available keywords for Paramter block are : \n'
        string +=   Key_class.__LJcollid + '\n'
        string +=   Key_class.__HScollid + '\n'
        string +=   Key_class.__temperature + '\n'
        string +=   Key_class.__pressure + '\n'
        string +=   Key_class.__energy + '\n'
        string +=   Key_class.__gndSpecies + '\n'
        string +=   Key_class.__merge_threshold + '\n'
        string +=   Key_class.__precision + '\n'
        string +=   Key_class.__ompnumthread + '\n'
        string +=   Key_class.__pseubimolecular + '\n'
        string +=   Key_class.__prints + '\n'
        return string

    def printLJCollisionKeywords():
        string  =   '\nThe available keywords for LJCollision section are : \n'
        string +=   Key_class.__epsilonM + '\n'
        string +=   Key_class.__epsilonA + '\n'
        string +=   Key_class.__diameterM + '\n'
        string +=   Key_class.__diameterA + '\n'
        string +=   Key_class.__massM + '\n'
        string +=   Key_class.__massA + '\n'
        return string
    
    def printHSCollisionKeywords():
        string  =   '\nThe available keywords for HSCollision section are : \n'
        string +=   Key_class.__bmax + '\n'
        string +=   Key_class.__massM + '\n'
        string +=   Key_class.__massA + '\n'
        return string

    def printEdownKeywords():
        string  =   '\nThe available keywords for Edown section are : \n'
        string +=   Key_class.__slope + '\n'
        string +=   Key_class.__EdConst + '\n'
        string +=   Key_class.__Tc + '\n'
        string +=   Key_class.__exponent + '\n'
        return string 

    def printEnergyKeywords():
        string  =   '\nThe available keywords for Energy section are : \n'
        string +=   Key_class.__ESOT + '\n'
        string +=   Key_class.__EEOT + '\n'
        string +=   Key_class.__dE + '\n'
        string +=   Key_class.__Emax + '\n'
        return string
    
    def printPseuBimolecularKeywords():
        string  =   '\nThe available keywords for PseuBimolecular section are : \n'
        string +=   Key_class.__pseubimname + '\n'
        string +=   Key_class.__excessconc + '\n'
        return string

    def printPrintsKeywords():
        string  =   '\nThe available keywords for Print section are : \n'
        string +=   Key_class.__evals + '\n'
        string +=   Key_class.__evecs + '\n'
        string +=   Key_class.__partifunc + '\n'
        string +=   Key_class.__kE + '\n'
        string +=   Key_class.__Edown + '\n'
        string +=   Key_class.__relaxkernel + '\n'
        return string
    
    def printTimeEvolutionKeywords():
        string  =   '\nThe available keywords for TimeEvolution section are : \n'
        string +=   Key_class.__time_file + '\n'
        string +=   Key_class.__time_step + '\n'
        string +=   Key_class.__time_num  + '\n'
        string +=   Key_class.__initspecies + '\n'
        return string

    def printEigenvalueKeywords():
        string  =   '\nThe available keywords for Eigenvalue subsection are : \n'
        string +=   Key_class.__evalfile + '\n'
        string +=   Key_class.__evalnum + '\n'
        return string

    def printEigenvectorKeywords():
        string  =   '\nThe available keywords for Eigenvector subsection are : \n'
        string +=   Key_class.__evecfile + '\n'
        string +=   Key_class.__evecnum + '\n'
        return string

    def printRelaxKernelKeywords():
        string  =   '\nThe available keywords for RelaxKernel subsection are :\n'
        string +=   Key_class.__relaxfile       + '\n'
        string +=   Key_class.__printkernelname + '\n'
        return string

    def printReactionKeywords():
        string  =   '\nThe available keywords for Reaction block are : \n'
        string +=   Key_class.__elem_rxn + '\n'
        string +=   Key_class.__barrierlessrxn + '\n'
        return string

    def printBarrierRxnKeywords():
        string  = '\nThe available keywords for BarrierRxn section are : \n'
        string += Key_class.__info + '\n'
        string += Key_class.__pyr_file + '\n'
        string += Key_class.__variation + '\n'
        string += Key_class.__tunnel + '\n'
        string += Key_class.__variation + '\n'
        string += Key_class.__s_lower + '\n'
        string += Key_class.__s_upper + '\n'
        string += Key_class.__invlaplace + '\n'            
        return string

    def printBarrierlessRxnKeywords():
        string  = '\nThe available keywords for BarrierRxn section are : \n'
        string += Key_class.__info + '\n'
        string += Key_class.__rxn_ener + '\n'
        string += Key_class.__avgdiameter + '\n'
        string += Key_class.__invlaplace + '\n'       
        return string

    def printSpeciesKeywords():
        string  = '\nThe available keywords for species block are : \n'
        string += Key_class.__well + '\n'
        string += Key_class.__tranState + '\n'
        string += Key_class.__bim + '\n'
        return string

    def printUnimolecularKeywords():
        string  = '\nThe available keywords for unimolecule section are : \n'
        string += Key_class.__name + '\n'
        string += Key_class.__geom + '\n'
        string += Key_class.__freq + '\n'
        string += Key_class.__E0K  + '\n'
        string += Key_class.__Eele + '\n'
        string += Key_class.__ele_level  + '\n'
        string += Key_class.__rotsigma   + '\n'
        string += Key_class.__mst_file   + '\n'
        string += Key_class.__g09_file   + '\n'
        string += Key_class.__image_freq + '\n'
        string += Key_class.__epsilon    + '\n'
        string += Key_class.__diameter   + '\n'
        string += Key_class.__bmax       + '\n'
        string += Key_class.__opticalnum + '\n'
        string += Key_class.__freqscale  + '\n'
        return string
    
    def printBimolecularKeywords():
        string  = '\nThe available keywords for bimolecule section are : \n'
        string += Key_class.__sp1_atom + '\n'
        string += Key_class.__sp1_mol  + '\n'
        string += Key_class.__sp2_atom + '\n'
        string += Key_class.__sp2_mol  + '\n'
        string += Key_class.__Eele     + '\n'
        string += Key_class.__E0K      + '\n'
        string += Key_class.__name     + '\n'
        return string

    def printAtomKeywords():
        string  = '\nThe available keywords for atomic subsection are : \n'
        string += Key_class.__name + '\n'
        string += Key_class.__symbol + '\n'
        return string
    
    def printInvLaplaceKeywords():
        string  = '\nThe available keywords for atomic subsection are : \n'
        string += Key_class.__expression + '\n'
        string += Key_class.__A1 + '\n'
        string += Key_class.__T1 + '\n'
        string += Key_class.__n1 + '\n'
        string += Key_class.__Ea1 + '\n'
        string += Key_class.__A2 + '\n'
        string += Key_class.__T2 + '\n'
        string += Key_class.__n2 + '\n'
        string += Key_class.__Ea2 + '\n'
        return string
    
