#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import numpy  as np
import qd_lib as qd 
import dd_lib as dd
from const_lib import cst

class MasterEquation():
    
    def __init__(self):
        self.__T            =   None
        self.__p            =   None
        self.__ener_step    =   None
        self.__ener_ref     =   None
        self.__Z            =   None
        self.__size         =   None
        self.__well_sz      =   []
        self.__well_kernel  =   []
        self.__Ncse         =   None
        self.__W            =   None
        self.__G            =   None
        self.__L            =   None
        self.__U            =   None
        self.__L_p          =   None
        self.__U_p          =   None
        self.__UC           =   None
        self.__LC           =   None
        self.__UI           =   None
        self.__LI           =   None
        self.__M            =   None
        self.__B            =   None
        self.__D            =   None
        self.__F            =   None
        self.__iF           =   None
        self.__h            =   None
        self.__kappa        =   None
        self.__epcs         =   None
        self.__epsrs        =   None
        self.__well_eig     =   None
        self.__merge_well   =   []
        self.__K            =   None
        self.__y            =   None
        self.__kE_map       =   []
        self.__charac_t     =   None


    @property
    def T      (self): return   self.__T
    @property
    def p      (self): return   self.__p
    @property
    def ener_step (self): return self.__ener_step
    @property
    def ener_ref  (self): return self.__ener_ref
    @property
    def kE_map (self): return   self.__kE_map
    @property
    def Z      (self): return   self.__Z
    @property
    def size   (self): return   self.__size
    @property
    def well_sz(self): return   self.__well_sz
    @property
    def well_kernel(self): return self.__well_kernel
    @property
    def Ncse   (self): return   self.__Ncse
    @property
    def W      (self): return   self.__W
    @property
    def G      (self): return   self.__G     
    @property
    def L      (self): return   self.__L 
    @property
    def U      (self): return   self.__U
    @property
    def L_p    (self): return   self.__L_p 
    @property
    def U_p    (self): return   self.__U_p
    @property
    def UC     (self): return   self.__UC   
    @property
    def LC     (self): return   self.__LC   
    @property
    def UI     (self): return   self.__UI   
    @property
    def LI     (self): return   self.__LI   
    @property
    def M      (self): return   self.__M    
    @property
    def B      (self): return   self.__B
    @property
    def D      (self): return   self.__D    
    @property
    def F      (self): return   self.__F     
    @property
    def iF     (self): return   self.__iF    
    @property
    def h      (self): return   self.__h    
    @property
    def kappa  (self): return   self.__kappa
    @property
    def epcs   (self): return  self.__epcs  
    @property
    def epsrs  (self): return  self.__epsrs 
    @property
    def K      (self): return  self.__K
    @property
    def y      (self): return  self.__y
    @property
    def well_eig  (self): return  self.__well_eig
    @property
    def merge_well(self): return  self.__merge_well
    @property
    def charac_t  (self): return  self.__charac_t


    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #        copy master equation
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
    #  
    def set_ME(self, **kwargs):
        for key in kwargs.keys():
            if     key        ==      '_T':
                self.__T            = kwargs[key]
            elif   key        ==      '_p':
                self.__p            = kwargs[key]
            elif   key        ==      '_ener_step':
                self.__ener_step    = kwargs[key]
            elif   key        ==      '_ener_ref':
                self.__ener_ref     = kwargs[key]
            elif   key        ==      '_Z':
                self.__Z            = kwargs[key]
            elif   key        ==      '_size':
                self.__size         = kwargs[key]
            elif   key        ==      '_well_sz':
                self.__well_sz      = kwargs[key]
            elif   key        ==      '_well_kernel':
                self.__well_kernel  = kwargs[key]
            elif   key        ==      '_Ncse':
                self.__Ncse         = kwargs[key]
            elif   key        ==      '_well_eig':
                self.__well_eig     = kwargs[key]
            elif   key        ==      '_merge_well':
                self.__merge_well   = kwargs[key]
            elif   key        ==      '_M':
                self.__M            = kwargs[key]
            elif   key        ==      '_epsrs':
                self.__epsrs        = kwargs[key]
            elif   key        ==      '_kappa':
                self.__kappa        = kwargs[key]
            elif   key        ==      '_epcs':
                self.__epcs         = kwargs[key]
            elif   key        ==      '_kE_map':
                self.__kE_map       = kwargs[key]
            elif   key        ==      '_L':
                self.__L            = kwargs[key]
            elif   key        ==      '_U':
                self.__U            = kwargs[key]
            elif   key        ==      '_L_p':
                self.__L_p          = kwargs[key]
            elif   key        ==      '_U_p':
                self.__U_p          = kwargs[key]
            elif   key        ==      '_F':
                self.__F            = kwargs[key]
            elif   key        ==      '_y':
                self.__y            = kwargs[key]
            elif   key        ==      '_K':
                self.__K            = kwargs[key]
            elif   key        ==      '_charac_t':
                self.__charac_t     = kwargs[key]
            else:
                print("wrong key ", key)
        self.__UC           =   self.__U[:,:self.__Ncse]
        self.__LC           =   self.__L[:self.__Ncse]
        self.__UI           =   self.__U[:,self.__Ncse:]
        self.__LI           =   self.__L[self.__Ncse:]



    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                  clear  memories
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
    #  
    def clear (self):
        # free memory
        self.__G            =   None
        self.__W            =   None
        self.__B            =   None
        self.__D            =   None
        self.__iF           =   None
        self.__h            =   None
        # save memory
        self.__F            =   np.array( np.diag(self.__F), order='C')
        


    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #           solve phenomenological rate constants
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #  
    #  
    def solveRateConstant( self, gvr, out ):
  
        well_stack  =   gvr.well_stack
        bim_stack   =   gvr.bim_stack
        Nbim        =   len( bim_stack  )
        Nwell       =   len( well_stack )
        Ncse        =   Nwell
        rxn_map     =   gvr.rxn_map
        idx         =   np.concatenate([0, np.cumsum( [x.size for x in well_stack ] )], axis = None )
        Z           =   [ x.collid.Z for x in gvr.well_stack ]

        #set Ncse
        self.__Ncse = Ncse

        #set temperature
        self.__T    =   gvr.cycle_param.T 

        #set pressure
        self.__p    =   gvr.cycle_param.p 

        #set ener_step
        self.__ener_step = gvr.cycle_param.ener_step

        #set ener_ref
        self.__ener_ref  = gvr.cycle_param.ener_ref

        #set collision rate
        self.__Z    =   Z

        #set kinetic matrix size
        self.__size =   np.sum ( [ x.size for x in well_stack ] ) 

        #set well sizes
        self.__well_sz   = [ x.size for x in well_stack ]

        #store well relaxation kernel
        if gvr.printkernelname is not None:
            self.__well_kernel = [ gvr.well_dict[x].kernel for x in gvr.printkernelname ]

        #set  F consisting of sqrt of boltnmann 
        self.__F    =   np.diag( np.concatenate( [   x.sqrt_bolztmann for x in well_stack ], axis = None ) ) 

        #set inverse of F
        self.__iF   =   np.diag( np.concatenate( [ 1/x.sqrt_bolztmann for x in well_stack ], axis = None ) ) 

        #set kE_map
        for i in range(Nwell+Nbim):
            kE_list = []
            for rxn in rxn_map[i]:
                if rxn:
                    if i < Nwell:
                        kE_list.append( np.array(rxn.kE) )
                    else:
                        kE_list.append( np.array(rxn.dk) )
            self.__kE_map.append ( kE_list )

        #set B and D
        B       =   np.zeros( (self.__size,  Nbim) )
        D       =   np.zeros( (self.__size,  Nbim) )
        for j in range(Nwell):
            for i in range(Nbim):
                if rxn_map[ i + Nwell ][j] and rxn_map[j][ i + Nwell ]:
                    B [ idx[j] : idx[j+1] ,i]     =  rxn_map[ i + Nwell ][j].dk
                    D [ idx[j] : idx[j+1] ,i]     =  rxn_map[j][ i + Nwell ].kE
        self.__B    =   B 
        self.__D    =   D 
  
        # construct W and G and eigenpairs
        if gvr.precision == 'double':
            #set asymmetrical kinetic matrix W
            W = np.zeros ( (self.__size, self.__size) )
            for i, welli in enumerate(well_stack):
                i0      =   idx [ i ]
                for j, wellj in enumerate(well_stack):
                    j0  =   idx [ j ]
                    #diagnal block
                    if i == j :
                        #kernel
                        kernel  =   welli.kernel
                        #kE
                        kE = 0
                        for rxn in rxn_map[i]:
                            if rxn:
                                kE += rxn.kE
                        #set diagnal block
                        for ii in range(welli.size):
                            for jj in range(welli.size):
                                #set non-diag elem
                                if ii != jj:
                                    W [ i0 + ii, j0 + jj ] = -Z[i] * kernel[ ii, jj ]
                                #set diag elem
                                else:
                                    W [ i0 + ii, j0 + jj ] = Z[i] * ( 1- kernel[ ii, jj ] ) + kE[ ii ]
                    #non-diagnal block
                    else:
                        if rxn_map[j][i]:
                            kE =  rxn_map[j][i].kE
                            for ii in range( min( welli.size, wellj.size ) ):
                                W [ i0 + ii, j0 + ii ] = - kE [ ii ]

            #set symmetrical kinetic matrix G
            self.__W    =   W
            self.__G    =   self.__iF @ W @ self.__F 
    
            #set eigen pair
            Ltmp, Utmp  =   np.linalg.eigh( self.__G )

            #check Nan
            if not np.isfinite(Ltmp).all():
                out.error(" Wrong! NaN countered in the transition matrix G in T=" + \
                str(self.__T*cst.au2K)+"K and p="+str(self.__p*cst.au2torr)+" torr, you should carefully check your input! ")

        elif gvr.precision == 'quadruple':
            Ltmp, Utmp  =   dd.ME_dd( gvr )
        elif gvr.precision == 'octuple':
            Ltmp, Utmp  =   qd.ME_qd( gvr )

        #set LC UC LI UI
        self.__L        =   np.sort(  Ltmp ) 
        self.__U        =   np.array( Utmp[ :, np.argsort(  Ltmp ) ], order='C' ) 
        self.__LC       =   self.__L [ :Ncse ] 
        self.__LI       =   self.__L [ Ncse: ] 
        self.__UC       =   self.__U [ :, :Ncse ] 
        self.__UI       =   self.__U [ :, Ncse: ] 
        
        #deal with 0 eigenvalue
        if len(gvr.bim_stack) == 0:
            self.__L[0] = 0

        #set h
        h   =   np.zeros([  Nwell, self.__size ])
        for  i in range(Nwell):
            h[ i, idx[i]:idx[i+1] ] = well_stack[i].sqrt_bolztmann
        self.__h    =   h 

        #set M
        self.__M    =   self.__h @ self.__UC 

        #set kappa
        kappa = self.__h @ self.__UI @ np.diag ( 1 / self.__LI ) @ self.__UI.T @ self.__iF @ self.__B
        for i in range(Nwell):
            for j in range(Nbim):
                kappa [ i, j ] = kappa [ i, j ] * bim_stack[j].Q_real / well_stack[i].Q_real
        self.__kappa  =   kappa 

        #set dictionary well_eig {well_index: cse_index }
        well_eig    =   {}
        normal_C    =   np.array( [ [ np.sqrt( well_stack[ i ].Q_static ) for j in range(Ncse)  ] for i in range (Nwell) ] )
        epcs        =   self.__M /  normal_C
        epsrs       =   1 - np.sum ( epcs**2, axis = 0 )

        cse_idx     =   0
        epce_abs    =   np.fabs ( epcs )
        while cse_idx  <  Nwell:
            if cse_idx not in well_eig.values():
                well_idx   =   np.argmax ( epce_abs[:, cse_idx] )
                n          =   -1
                while well_idx in well_eig.keys():
                    if epce_abs[ well_idx, cse_idx ] <= epce_abs [ well_idx, well_eig [ well_idx] ]:
                        n            =   n - 1
                        well_idx     =   np.argsort ( epce_abs[ :, cse_idx ] )[ n ]
                        if n < - Ncse:
                            out.error( " Something went wrong when associate eigenmodes to wells ! " )
                    else:
                        well_eig [ well_idx ], cse_idx  =   cse_idx, well_eig[ well_idx ]    
                well_eig.update( { well_idx: cse_idx } )
            cse_idx  =  cse_idx + 1

        self.__well_eig     =   well_eig 
        self.__epcs         =   epcs
        self.__epsrs        =   epsrs

        #check merge
        if epsrs [ Ncse - 1 ] > gvr.merge_threshold:
            #normlize M
            epcs_norm = np.zeros ( (Nwell, Ncse) )
            for j in range( Ncse ):
                neg_pop = 0; pos_pop =0
                for m in self.__M[ :, j ]:
                    if m > 0:
                        pos_pop = pos_pop + m
                    else:
                        neg_pop = neg_pop + m
                epcs_norm [ :, j ] = self.__M[ :, j ] / max(pos_pop, -neg_pop)
            #reset Ncse
            while epsrs [ Ncse - 1 ] > gvr.merge_threshold:
                Ncse    =   Ncse - 1
                if  Ncse == 0:
                    break
            self.__Ncse = Ncse

            #reset self.LC, self.LI, self.UC, self.UI
            self.__LC   =   self.__L[ :Ncse ] 
            self.__LI   =   self.__L[ Ncse: ] 
            self.__UC   =   self.__U[ :, :Ncse ]
            self.__UI   =   self.__U[ :, Ncse: ]

            #reset kappa
            kappa = self.__h @ self.__UI @ np.diag ( 1 / self.__LI ) @ self.__UI.T @ self.__iF @ self.__B
            for i in range(Nwell):
                for j in range(Nbim):
                    kappa [ i, j ] = kappa [ i, j ] * bim_stack[j].Q_real / well_stack[i].Q_real
            self.__kappa  =   kappa 

            # set merge_well
            merge_well = {}
            for k in range( Ncse, Nwell ):
                epcs_norm_v   =   epcs_norm[ :, k ]
                idx_w       =   np.array([x for x in well_eig.keys()])[ np.argwhere( ( np.array([x for x in well_eig.values()] ) == k ) )[0,0] ]
                tmp         =   [np.argwhere(epcs_norm_v==x)[0,0]  for x in epcs_norm_v if x < -0.5 or x > 0.5] 
                if len(tmp) >= 2 and epcs_norm_v[tmp[0]]*epcs_norm_v[tmp[1]] < 0:
                    if   idx_w == tmp[0]:    idx_ph =  tmp[1]
                    elif idx_w == tmp[1]:    idx_ph =  tmp[0]
                    else:
                        idx_ph = Nwell + np.argmax( self.__kappa[ idx_w, : ])
                else:
                    idx_ph = Nwell + np.argmax( self.__kappa[ idx_w, : ])
                merge_well.update (  { idx_w: idx_ph } )
            self.__merge_well   =   merge_well

            # reset M
            M_tmp   =   np.array ( self.__M )
            for  i  in  self.__merge_well:
                ii = self.__merge_well[ i ]
                if ii < Nwell:
                    M_tmp [ ii, : ] = M_tmp [ ii, : ] + self.__M[ i, : ]        
            M_tmp_m =   np.empty([Ncse,Ncse] )
            ii      =   0
            for i in range(Nwell):
                if i not in self.__merge_well:
                    M_tmp_m[ ii, :] = M_tmp[ i, :Ncse]
                    ii  =   ii + 1
            self.__M    =   M_tmp_m 


        #rate constant
        Kww  =  -( self.__M   @ np.diag(self.__LC)    @ np.linalg.inv ( self.__M ) ).T         
        if Nbim != 0:
            Kbw  =     self.__B.T @ self.__iF @ self.__UC @ self.__M.T                        
            Kwb  =   ( self.__D.T @ self.__F  @ self.__UC @ np.linalg.inv ( self.__M ) ).T    
            Kbb  =   ( self.__D.T @ self.__F  @ self.__UI @ np.diag(  1 / self.__LI ) @ self.__UI.T @ self.__iF @ self.__B ).T 

        Nph         =   Nbim + Nwell
        K           =   np.zeros( ( Nph, Nph ) )
        I  = 0 
        for i in range( Nph ):
            J =   0
            if i not in self.__merge_well:
                for j in range( Nph ):
                    if j not in self.__merge_well:
                        if I < Ncse:
                            if J < Ncse:
                                if I != J:
                                    K[ i, j ] =  Kww[ I, J ]
                            else:
                                K[ i, j ] =  Kwb[ I, J - Ncse ]
                        else:
                            if J < Ncse:
                                K[ i, j ] =  Kbw[ I - Ncse, J ]
                            else:
                                if I != J:
                                    K[ i, j ] = Kbb[ I - Ncse, J - Ncse ]  
                        J = J + 1
                I = I + 1
        K   =   K + np.diag( np.sum( K, axis=1) )
        self.__K = K

        
    
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                 time evolution of solutions
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #  
    # 
    def solveTimeEvolution( self, gvr):

        if gvr.timeevolution_sw:
            # pseudo-first-order condition 
            if gvr.pseubimname is not None:
                #set local variables
                sz              = self.__size 
                r               = len(gvr.pseubimname)
                sz_p            = sz + r       
                rxn_map         = gvr.rxn_map
                well_stack      = gvr.well_stack
                bim_stack       = gvr.bim_stack
                well_name       = [ x.name for x in well_stack ]
                bim_name        = [ x.name for x in bim_stack ]
                pseubimname     = gvr.pseubimname
                excessconc      = gvr.excessconc
                Nwell           = len(well_stack)
                idx             = np.concatenate([0, np.cumsum( [x.size for x in well_stack ] )], axis = None )
                
                #set F matrix
                F_diag                  =   np.empty(sz_p)
                F_diag[ :sz ]           =   np.concatenate( [   x.sqrt_bolztmann for x in well_stack ], axis = None )
                for i in range(r):
                    bim_i               =   bim_name.index( pseubimname[i] ) 
                    F_diag[ sz+i ]      =   np.sqrt(  bim_stack[bim_i].Q_real / excessconc[i] / self.__ener_step  )
                F                       =   np.diag( F_diag )
                iF                      =   np.diag( 1 / F_diag )
                
                #double precision
                if gvr.precision == 'double':
                    # construct pseudo-first-order transition metrix W
                    W               = np.zeros((sz_p, sz_p ), dtype = 'float', order='C' )
                    W[ :sz, :sz ]   = self.__W                
                    for i in range(r):    
                        bim_i               =  bim_name.index( pseubimname[i] ) 
                        W[ :sz, sz+i ]      =  0-self.__B[ : , bim_i ] * excessconc[i]
                        for well_i in range(Nwell):
                            if rxn_map[well_i][ Nwell+bim_i ] is not None:
                                W[ sz+i, idx[well_i]:idx[well_i+1] ] = 0-rxn_map[well_i][ Nwell+bim_i ].kE
                        W[ sz_p-1, sz_p-1 ] =  np.sum(self.__B[ :, bim_i]) * excessconc[i] 
                    
                    # construct symmetrized matrix G and calc eigenpairs
                    G                       =   iF @ W @ F
                    Ltmp, Utmp              =   np.linalg.eigh( G )

                #quadruple precision
                elif gvr.precision == 'quadruple':
                    Ltmp, Utmp              =   dd.pseudo_ME_dd( gvr )
            
                #octuple precision
                elif gvr.precision == 'octuple':
                    Ltmp, Utmp              =   qd.pseudo_ME_qd( gvr )
                
                # set eigen pairs
                L   =   self.__L_p  =   np.sort(  Ltmp ) 
                U   =   self.__U_p  =   np.array( Utmp[ :, np.argsort(  Ltmp ) ], order='C' )
                
                # deal with 0 eigenvalue and set characteristic time scale
                if len(gvr.pseubimname) == len(gvr.bim_stack):
                    L[0]        =   0
                    self.__charac_t    =   1 / L[1]
                else:
                    self.__charac_t    =   1 / L[0]

                # set index of bins with initial population (ground state) 
                I                   =   np.empty( len(gvr.initspecies), dtype=np.int )
                for i,species in enumerate(gvr.initspecies):
                    if species in pseubimname:
                        I[i]        =  sz + pseubimname.index( species )
                    else:
                        I[i]        =  int(np.sum( self.__well_sz [ : well_name.index(species) + 1 ] ) - 1)
                
                # size of vector y
                sz_y                =  sz_p
            
            # all bimoleculars are viewed as sinks
            else:
                L           =   self.__L 
                U           =   self.__U
                F           =   self.__F
                iF          =   self.__iF
                well_name   =   [x.name for x in gvr.well_stack]

                # set characterisitic time scale
                if gvr.bim_stack:
                    self.__charac_t =   1 / L[0]
                else:
                    self.__charac_t =   1 / L[1]

                # set index of bins with initial population (ground state)
                I                   =   np.empty( len(gvr.initspecies), dtype=np.int )
                for i,species in enumerate(gvr.initspecies):
                    I[i]            =   int( np.sum( self.__well_sz [ : well_name.index(species) + 1 ] ) - 1 )
                
                # size of vector y
                sz_y                =   self.__size

            # time step
            if gvr.time_step    ==  None:
                time_step       =   self.__charac_t / 300
            else:
                time_step       =   gvr.time_step

            # numbers of step
            if gvr.time_num     ==  None:
                Ntime           =   50  
            else:
                Ntime           =   int(gvr.time_num)
            
            # delta initial condition
            y                   =   np.empty( ( Ntime, sz_y ), dtype = 'float', order='C' )
            y0                  =   np.zeros( ( sz_y, 1 ) )
            for i in I: 
                y0[ i,0 ]       =   1
            
            
            # time evolution
            for t in range( Ntime ):
                y[ t, : ]  = (F @ U @ np.diag( np.exp( 0 - L * t * time_step ) ) @ U.T @ iF @ y0 )[:, 0] 
            self.__y = y
        
