#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import  numpy           as      np
import  species_lib     as      sl
from    const_lib       import  cst
import  re
      

class MEP:
    def __init__(self, _R_name, _TS_name, _P_name):
        self.__reactant     =   _R_name
        self.__ts           =   _TS_name
        self.__prodcut      =   _P_name
        self.__sctPE        =   []
        self.__zctPE        =   []
        self.__lctPE        =   []
        self.__omtPE        =   []
        self.__gts_dict     =   {}
        self.__barrier      =   None
        self.__rxn_ener     =   None
        
    @property
    def gts_dict (self):   return self.__gts_dict
    @property
    def reactant  (self):   return self.__reactant
    @property   
    def ts        (self):   return self.__ts
    @property
    def product   (self):   return self.__prodcut     
    @property
    def sctPE     (self):   return self.__sctPE
    @property
    def zctPE     (self):   return self.__zctPE
    @property
    def lctPE     (self):   return self.__lctPE
    @property
    def omtPE     (self):   return self.__omtPE
    @property
    def barrier   (self):   return self.__barrier
    @property
    def rxn_ener  (self):   return self.__rxn_ener

    
    def readPYR( self, _pyr_file, _s_lower, _s_upper, gvr, errorf ):
        symbol_key       =   'Atomic information:'
        R1_obj_key       =   'Reactant #1'
        R2_obj_key       =   'Reactant #2'
        P1_obj_key       =   'Product #1'
        P2_obj_key       =   'Product #2'
        TS_obj_key       =   'Starting point '
        obj_off          =   6
        Natom_off        =   3
        islinear_key     =   'SPECIES'
        eleLevel_key     =   'Electronic degeneracies and energies'
        R_geom_key       =   '********** Reactants:'
        P_geom_key       =   '********** Products:'
        TS_geom_key      =   '********* Saddle point:'
        geom_off         =   5
        R1_info_key      =   '********* Reactant 1 ************'
        R2_info_key      =   '********* Reactant 2 ************'
        P1_info_key      =   '*********** Product 1 *************'
        P2_info_key      =   '*********** Product 2 *************'
        TS_info_key      =   '********** Saddle point ***********'
        freq_key         =   '    Harmonic Frequencies'
        freq_off         =   4
        freq_end_key     =   'Saddle point energetics'
        rxn_ener_key     =   'Reaction energetics'
        rxn_ener_off     =   16
        barrier_key      =   'V+ZPE w/re reactant V+ZPE'
        gts_freq_key     =   's(angstrom) VMEP     Va^G    mu^CD-SC   frequencies'
        gts_freq_off     =   2
        gts_geom_key     =   'Space-fixed cartesian coordinates vs reaction coordinate'
        gts_geom_off     =   4
        tunnel_sct_key   =   'Transmission probabilities'
        tunnel_sct_off   =   4
        tunnel_lct_key   =   'Table of Tunneling Probabilities'
        tunnel_lct_off   =   7

        class mol():
            def __init__(self):
                self.name       =   None
                self.Natom      =   None
                self.seq        =   None
                self.geom       =   None
                self.freq       =   None
                self.imagefreq  =   None
                self.islinear   =   None
                self.eleLevel   =   None
                self.symbol     =   None
                self.rotsigma   =   1
                self.opticalnum =   1
                self.freqscale  =   1
                self.diameter   =   None
                self.epsilon    =   None
                self.bmax       =   None
                self.mst_file   =   None
                self.g09_file   =   None
                self.source     =   "Polyrate standard output"


        def __read_symbol( I, cdata ):
            symbol  =   []
            I       =   I   +   3
            line    =   cdata[I].strip()
            while line:
                symbol.append( line.split()[2].strip() )
                I       =   I + 1
                line    =   cdata[I].strip()
            return  I, np.array(symbol)

        def __read_obj( I, cdata ):
            obj     =   mol()
            I       =   I + obj_off
            line    =   cdata[I].strip()
            while  line:
                if islinear_key in cdata[I]:
                    tmp = cdata[I].strip().split()[3]
                    if tmp == 'linrp':
                        obj.islinear =   True
                    else:
                        obj.islinear =   False
                I       =       I   +    1
                line    =       cdata[I].strip()
            return I, obj

        def __read_geom( I, cdata ):
            geom    =   []
            I       =   I + geom_off
            line    =   cdata[I].strip()
            while line:
                geom.append( [ float(x) for x in line.split()[-3:] ] )
                I       =   I + 1
                line    =   cdata[I].strip()
            return I, np.array(geom)*cst.A2au

        def __read_info( I, cdata, symbol, islinear ):
            #number of atom and their sequences
            if 'Species' in cdata[ I + Natom_off ]:
                if '-' in cdata[ I + Natom_off ]:
                    Natom = int(cdata[ I + Natom_off ].strip().split('-')[0].split()[-1])
                else:
                    Natom = 1

                i = 0
                if Natom > 1:
                    atom_seq = [int(x) for x in cdata[ I + Natom_off ].strip().split(':')[1].split() ]
                    while len(atom_seq) != Natom:
                        i = i + 1
                        atom_seq = atom_seq + [ int(x) for x in cdata[ I + Natom_off + i ].strip().split() ]
                else:
                    atom_seq = [int(x) for x in cdata[ I + Natom_off ].strip().split('#')[1].split() ]
            else:
                Natom = len(symbol)
                atom_seq = range(int(Natom))
            #degree of freedom
            if islinear:
                dof = max(3 * Natom - 5, 0)
            else:
                dof = max(3 * Natom - 6, 0)
            #eletronic levels
            ele_tmp =   []
            while True:
                I   =   I   +   1
                if eleLevel_key in cdata[I]:
                    ele_tmp.append( [ float(x) for x in cdata[I].split()[-2:]] )
                    I       =   I + 1
                    line    =   cdata[I].strip().split()
                    while len(line) == 2:
                        tmp = [ float(x) for x in line ]
                        if tmp [0]:
                            ele_tmp.append( tmp )
                        I        =   I   +   1
                        line     =   cdata[I].strip().split()
                    break
                if I >= len(cdata):
                    raise
            #freq
            freq    =   []
            I       =   I + 1
            while Natom > 1:
                line = cdata[I]
                if freq_key in line:
                    I    =   I + freq_off
                    line =   cdata[I].strip().split()
                    for i in range( dof ):
                        if 'i' in line:
                            freq.append( 0-float(line[6]) )
                        else:
                            freq.append(float(line[-2]))
                        I       =   I +1
                        line    =   cdata[I].strip().split()
                    break
                I = I + 1
                if freq_end_key in line:
                  break
            return I, Natom, ele_tmp, np.array(freq)*cst.cm2au, np.array(atom_seq)

        def __read_rxn_ener( I, cdata ):
            I   =   I   +   rxn_ener_off
            return I, float(cdata[I].strip().split()[6])

        def __read_gts_geom( I, cdata, symbol ):
            Natom       =   len(symbol)
            I           =   I   +   gts_geom_off
            line        =   cdata[I].strip()
            tmp_geom    =   []
            gts_geom    =   []
            while gts_freq_key not in line:
                if 'Atom' in line:
                    for n in range(Natom):
                        I   =   I   +   1
                        tmp_geom.append( [ float(x) for x in cdata[I].strip().split()[1:] ] )
                    gts_geom.append( np.array( tmp_geom ) )
                    tmp_geom=[]
                I = I + 1
                line = cdata[I].strip()
            return I, gts_geom

        def __read_gts_freq( I, cdata, symbol ):
            dI  =  ( 3*len(symbol) - 6) // 7 + 1
            tmp_freq        =   []
            gts_freq        =   []
            gts_E0          =   []
            gts_s           =   []

            I               =   I + gts_freq_off
            line            =   cdata[I]
#            while gts_freq_1_end not in line and gts_freq_2_end not in line:
            while bool(re.search(r'\d',line)):
                line        =   cdata[I].strip().split()
                if line:
                    gts_s.append    ( float(line[0])*cst.A2au )
                    gts_E0.append   ( float(line[2])*cst.kcalmol2au )
                    tmp_freq    =   tmp_freq+[float(x) for x in line[4:]]
                    for n in range( 1, dI ):
                        I       =   I + 1
                        for x in cdata[I].strip().split():
                            if "i" in x:
                                tmp_freq.append( 0-float(x[:-1]) )
                            else:
                                tmp_freq.append( float(x) )
                    gts_freq.append( np.array(tmp_freq)*cst.cm2au )
                    tmp_freq    =   []
                I           =   I   +   1
                line        =   cdata[I]
                
            return I, gts_s, gts_E0, gts_freq

        def __read_tunnel_sct( I, cdata ):
            sctPE       =   []
            zctPE       =   []
            # determine index of sct and zct
            sct_idx     =   0
            zct_idx     =   0
            line        =   cdata[I+2].strip().split()
            for i in range(len(line)):
                if   line[i] == 'ZCT':
                    zct_idx = i
                elif line[i] == 'SCT':
                    sct_idx = i
            # read sct and zct PE
            I           =   I + tunnel_sct_off
            line        =   cdata[I].strip()
            while line != '':
                line = line.split()
                if len(line) > 2:
                    if zct_idx != 0:
                        zctPE.append( [ float(line[0])*cst.kcalmol2au, float(line[zct_idx]) ] )
                    if sct_idx != 0:
                        sctPE.append( [ float(line[0])*cst.kcalmol2au, float(line[sct_idx]) ] )
                I       =   I   +   1
                line    = cdata[I].strip()
            return I, np.array(sctPE), np.array(zctPE)

        def __read_tunnel_lct( I, cdata ):
            lctPE       =   []
            omtPE       =   []
            # determine index of sct and zct
            lct_idx     =   0
            omt_idx     =   0
            line        =   cdata[I+5].strip().split()
            for i in range(len(line)):
                if   line[i] == 'LCG4':
                    lct_idx = i
                elif line[i] == 'muOMT':
                    omt_idx = i
            # read lct and omt PE
            I           =   I + tunnel_lct_off
            line        =   cdata[I].strip()
            while line != '':
                line = line.split()
                if lct_idx != 0:
                    lctPE.append( [ float(line[1])*cst.kcalmol2au, float(line[lct_idx]) ] )
                if omt_idx != 0:
                    omtPE.append( [ float(line[1])*cst.kcalmol2au, float(line[omt_idx]) ] )
                I       =   I   +   1
                line    = cdata[I].strip()
            return I, np.array(lctPE), np.array(omtPE)

        if _pyr_file:
            #init variables
            R1          =   None;           R2          =   None
            P1          =   None;           P2          =   None
            TS          =   None;           Rname       =   self.__reactant
            TSname      =   self.__ts;      Pname       =   self.__prodcut          
            gts_s       =   []

            try:
                f           =   open( _pyr_file )
            except:
                errorf( ' Error occurs when open polyrate file ' + _pyr_file )

            cdata       =   f.readlines()         

            #read polyrate output
            I = 0
            while I < len( cdata ) :
                if   symbol_key    in  cdata[I]:
                    try:
                        I, symbol                           =   __read_symbol( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ symbol_key +'\" in '+ str(_pyr_file) )
                elif  R1_obj_key   in  cdata[I]:
                    try:
                        I, R1                               =   __read_obj  ( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ R1_obj_key +'\" in '+ str(_pyr_file) )
                elif  R2_obj_key   in  cdata[I]:
                    try:
                        I, R2                               =   __read_obj  ( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ R2_obj_key +'\" in '+ str(_pyr_file) )
                elif  P1_obj_key   in  cdata[I]:
                    try:
                        I, P1                               =   __read_obj  ( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ P1_obj_key +'\" in '+ str(_pyr_file) )
                elif  P2_obj_key   in  cdata[I]:
                    try:
                        I, P2                               =   __read_obj  ( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ P2_obj_key +'\" in '+ str(_pyr_file) )
                elif  TS_obj_key   in  cdata[I]:
                    try:
                        I, TS                               =   __read_obj  ( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ TS_obj_key +'\" in '+ str(_pyr_file) )
                elif  R_geom_key   in  cdata[I]:
                    try:
                        I, R_geom                           =   __read_geom( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ R_geom_key +'\" in '+ str(_pyr_file) )
                elif  R1_info_key  in  cdata[I]:
                    try:
                        I, R1.Natom, R1.eleLevel, R1.freq, R1.seq =   __read_info( I, cdata, symbol, R1.islinear)
                        R1.geom                             =   R_geom[ R1.seq-1 , : ]
                        R1.symbol                           =   symbol[ R1.seq-1  ]
                        if len(R1.freq) == 0 and R1.Natom != 1:
                          raise
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ R1_info_key +'\" in '+ str(_pyr_file) )
                elif  R2_info_key  in  cdata[I]:
                    try:
                        I, R2.Natom, R2.eleLevel, R2.freq, R2.seq =   __read_info( I, cdata, symbol, R2.islinear)
                        R2.geom                             =   R_geom[ R2.seq-1 , : ]
                        R2.symbol                           =   symbol[ R2.seq-1  ]
                        if len(R2.freq) == 0 and R2.Natom != 1:
                          raise
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ R2_info_key +'\" in '+ str(_pyr_file) )
                elif  P_geom_key   in  cdata[I]:
                    try:
                        I, P_geom                           =   __read_geom( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ P_geom_key +'\" in '+ str(_pyr_file) )
                elif  P1_info_key  in  cdata[I]:
                    try:
                        I, P1.Natom, P1.eleLevel, P1.freq, P1.seq =   __read_info( I, cdata, symbol, P1.islinear)
                        P1.geom                             =   P_geom[ P1.seq-1 , : ]
                        P1.symbol                           =   symbol[ P1.seq-1  ]
                        if len(P1.freq) == 0 and P1.Natom != 1:
                          raise
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ P1_info_key +'\" in '+ str(_pyr_file) )
                elif  P2_info_key  in  cdata[I]:
                    try:
                        I, P2.Natom, P2.eleLevel, P2.freq, P2.seq =   __read_info( I, cdata, symbol, P2.islinear)
                        P2.geom                             =   P_geom[ P2.seq-1 , : ]
                        P2.symbol                           =   symbol[ P2.seq-1  ]
                        if len(P2.freq) == 0 and P2.Natom != 1:
                          raise
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ P2_info_key +'\" in '+ str(_pyr_file) )
                elif  TS_geom_key  in  cdata[I]:
                    try:
                        I, TS.geom                          =   __read_geom( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ TS_geom_key +'\" in '+ str(_pyr_file) )
                elif  TS_info_key  in  cdata[I]:
                    try:
                        I,TS.Natom, TS.eleLevel, TS.freq, TS.seq =   __read_info( I, cdata, symbol, TS.islinear)
                        TS.symbol                           =   symbol
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ TS_info_key +'\" in '+ str(_pyr_file) )
                elif  rxn_ener_key in  cdata[I]:
                    try:
                        I, self.__rxn_ener                  =   __read_rxn_ener( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ rxn_ener_key +'\" in '+ str(_pyr_file) )
                elif  barrier_key  in  cdata[I]:
                    try:
                        self.__barrier                      =   float(cdata[I].strip().split()[6]) * cst.cm2au;  I = I + 1
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ barrier_key +'\" in '+ str(_pyr_file) )
                elif  gts_geom_key in  cdata[I]:
                    try:
                        I, gts_geom                         =   __read_gts_geom( I, cdata, symbol )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ gts_geom_key +'\" in '+ str(_pyr_file) )
                elif gts_freq_key  in  cdata[I]:
                    try:
                        I, gts_s,gts_E0,gts_freq            =   __read_gts_freq( I, cdata, symbol )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ gts_freq_key +'\" in '+ str(_pyr_file) )
                elif tunnel_sct_key  in  cdata[I]:
                    try:
                        I, self.__sctPE, self.__zctPE       =   __read_tunnel_sct( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ tunnel_sct_key +'\" in '+ str(_pyr_file) )
                elif tunnel_lct_key  in  cdata[I]:
                    try:
                        I, self.__lctPE, self.__omtPE       =   __read_tunnel_lct( I, cdata )
                    except:
                        errorf( 'Polyrate file format not matched.\n\nError occurs after \"'+ tunnel_lct_key +'\" in '+ str(_pyr_file) )
                else:
                    I   =   I   +   1
            # check TS freq
            try:
              if len(TS.freq) == 0:
                for i in range(len(gts_s)):
                  if gts_s[i] == 0.0 :
                    TS.freq = np.array( gts_freq[i] )
              else:
                TS.freq, TS.imagefreq = TS.freq[:-1], TS.freq[-1]
            except:
              errorf( 'Failed to read TS frequencies neither from the information of transition states nor s=0 in the MEP' )

            #assign name
            if R1 and R2:
                R1.name     = Rname + '_1' 
                R2.name     = Rname + '_2' 
            else:
                R1.name     = Rname
            TS.name         = TSname
            if P1 and P2:
                P1.name     = Pname  + '_1' 
                P2.name     = Pname  + '_2' 
            else:
                P1.name     = Pname 

            #reactant
            if R1 and R2:
                if R1.Natom == 1:
                    sp1         =   sl.FragmentAtom_class   ( R1.name, 0, R1.eleLevel, R1.symbol[0], R1.source, errorf)
                else:
                    sp1         =   sl.FragmentMol_class    ( R1.name, R1.symbol, R1.geom, R1.freq, R1.freqscale, 'E0K', 0, R1.eleLevel, R1.rotsigma, R1.opticalnum, R1.mst_file, R1.g09_file, R1.source, errorf)
                if R2.Natom == 1:
                    sp2         =   sl.FragmentAtom_class   ( R2.name, 0, R2.eleLevel, R2.symbol[0], R2.source, errorf)
                else:
                    sp2         =   sl.FragmentMol_class    ( R2.name, R2.symbol, R2.geom, R2.freq, R2.freqscale, 'E0K', 0, R2.eleLevel, R2.rotsigma,  R2.opticalnum, R2.mst_file, R2.g09_file, R2.source, errorf)
                self.__reactant =   sl.Bimolecular_class    ( Rname, 'E0K', 0, sp1, sp2, errorf)
            else:
                self.__reactant =   sl.Well_class           ( R1.name, R1.symbol, R1.geom, R1.freq,R1.freqscale, 'E0K', 0, R1.eleLevel, R1.rotsigma, R1.opticalnum, R1.epsilon, R1.diameter, R1.bmax, gvr.collid, R1.mst_file, R1.g09_file, R1.source, errorf)
            #product
            if P1 and P2:
                if P1.Natom == 1:
                    sp1         =   sl.FragmentAtom_class   ( P1.name,  0, P1.eleLevel, P1.symbol[0], P1.source, errorf)
                else:
                    sp1         =   sl.FragmentMol_class    ( P1.name, P1.symbol, P1.geom, P1.freq, P1.freqscale, 'E0K', 0, P1.eleLevel, P1.rotsigma, P1.opticalnum, P1.mst_file, P1.g09_file, P1.source, errorf)
                if P2.Natom == 1:
                    sp2         =   sl.FragmentAtom_class   ( P2.name,  0, P2.eleLevel, P2.symbol[0], P2.source, errorf)
                else:
                    sp2         =   sl.FragmentMol_class    ( P2.name, P2.symbol, P2.geom, P2.freq, P2.freqscale, 'E0K', 0, P2.eleLevel, P2.rotsigma, P2.opticalnum, P2.mst_file, P2.g09_file, P2.source, errorf)
                self.__prodcut  =   sl.Bimolecular_class    ( Pname, 'E0K', 0, sp1, sp2, errorf )
            else:
                self.__prodcut  =   sl.Well_class           ( P1.name, P1.symbol, P1.geom, P1.freq, P1.freqscale, 'E0K', 0, P1.eleLevel, P1.rotsigma, P1.opticalnum, P1.epsilon, P1.diameter, P1.bmax, gvr.collid, P1.mst_file, P1.g09_file, P1.source, errorf)
            #transition state
            self.__ts           =   sl.TS_class             ( TS.name, TS.symbol, TS.geom, TS.freq, TS.freqscale, 'E0K', 0, TS.eleLevel, TS.rotsigma, TS.opticalnum, TS.mst_file, TS.g09_file, TS.imagefreq, TS.source ,errorf)

            #general transition state
            if len(gts_s) > 0:
                if _s_lower is None:
                    s_lower = min(gts_s)
                else:
                    s_lower = _s_lower

                if _s_upper is None:
                    s_upper = max(gts_s)
                else:
                    s_upper = _s_upper

                for i in range(len(gts_s)):
                    if gts_s[i] >= s_lower and gts_s[i] <= s_upper:
                        if gts_freq[i][-1] > 0 and gts_freq[i][-2] > 0 :
                            self.__gts_dict.update ( { gts_s[i]: sl.TS_class(str(gts_s[i]), TS.symbol, gts_geom[i], gts_freq[i], TS.freqscale, 'E0K', gts_E0[i], TS.eleLevel, TS.rotsigma, TS.opticalnum, TS.mst_file, TS.g09_file, TS.imagefreq, TS.source ,errorf) } )

