#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import numpy         as np
import global_lib    as gl
import const_lib     as cl
import numba         as nb
import data_lib      as dl

class Atom_class(gl.CycleParam_class):
    'class for single atom'
    __Natom  =  1
    def __init__(
        self,
        _name,
        _energy,
        _ele_level,
        _symbol,
        _source
):
        self.__name         =   _name
        self.__symbol       =   _symbol   
        self.__mass         =   cl.cst.mass_dict[_symbol]
        self.__Qe           =   1.0
        self.__Qr           =   1.0
        self.__Qv           =   1.0
        self.__Qevr         =   1.0
        self.__Qt           =   1.0
        self.__opticalnum   =   1
        self.__ZPVE         =   0
        self.__E0           =   _energy
        self.__dos          =   None
        self.__ele_level    =   _ele_level
        self.__source       =   _source
    
    @property
    def name(self):                 return  self.__name
    @property
    def symbol(self):               return  self.__symbol
    @property
    def mass(self):                 return  self.__mass
    @property
    def Natom(self):                return  self.__Natom
    @property
    def E0(self):                   return  self.__E0
    @property
    def Qe(self):                   return  self.__Qe
    @property
    def Qr(self):                   return  self.__Qr
    @property
    def Qv(self):                   return  self.__Qv
    @property
    def Qevr(self):                 return  self.__Qevr
    @property
    def Qt(self):                   return  self.__Qt
    @property
    def ZPVE(self):                 return  self.__ZPVE
    @property
    def ele_level(self):            return  self.__ele_level
    @property
    def opticalnum(self):           return  self.__opticalnum
    @property
    def source(self):               return  self.__source
    @property
    def dos(self):                  return  self.__dos


    def setState( self, dE, Emax, errorf = print ):
        MAX           = int( round( Emax / dE ) + 1 )
        self.__dos    = np.zeros( MAX )
        self.__dos[0] = 1 / dE
        # only count electronic dos
        for x in self.__ele_level:
            degen                   =   x[0]
            ener                    =   x[1]
            i0                      =   int( round( ener /  dE ) )
            self.__dos[i0 : ]       =   self.__dos[ : MAX - i0 ] *  degen


    def setPartitionfunc(self):
        T = self.T
        #electronic Q
        Qe = 0
        for x in self.__ele_level:
             Qe += x[0] * np.exp( -x[1] / cl.cst.kb / T )
        self.__Qe   = Qe
        #translational Q
        self.__Qt   =   np.power( 2 * np.pi * self.__mass * cl.cst.kb * T / cl.cst.h**2, 3/2 )
        #electronic-vibrational-rotational Q
        self.__Qevr = Qe


    def printProperty(self, printf):
        printf(' [ATOM PROPERTIES]')
        printf('%-30s'%('  Name:')+self.__name   )
        printf('%-30s'%('  Read from: ')+self.__source )
        printf('%-30s'%('  Atom:')+self.__symbol )
        printf('%-30s'%('  E(ele) [a.u.]:')+'%-15s'%(str(self.__E0)))
        printf('%-30s'%('  E(0K) [a.u.]:')+'%-15s'%(str(self.__E0)))
        printf('%-30s'%('  Eletronic level [a.u.]:'))
        for x in self.__ele_level:
            printf('%3s'%(' ')+'%17.5E'%(x[0]) + '%17.5E'%(x[1]) )
        printf(' [END OF ATOM]')
        printf(' ')



class Molecule_class(gl.CycleParam_class):
    'father class for molecule'
    def __init__(
        self,
        _name,
        _geom,
        _symbol,
        _freq,
        _freqscale,
        _ener_type,
        _energy,
        _ele_level,                
        _rotsigma,    
        _opticalnum,           
        _mst_file,                 
        _g09_file,                 
        _image_freq,            
        _source, 
        errorf = print 
):

        self.__name               =   _name
        self.__source             =   _source
        self.__opticalnum         =   _opticalnum
        self.__mst_file           =   _mst_file
        self.__freqscale          =   _freqscale
        
        if not _g09_file:
            self.__geom           =   _geom
            self.__symbol         =   _symbol
            self.__freq           =   _freq
            self.__ZPVE           =   self.__calcZPVE()
            self.__ele_level      =   _ele_level
            self.__Natom          =   _symbol.shape[0]
            self.__E0             =   self.__setE0(_ener_type, _energy, errorf)
            self.__image_freq     =   _image_freq
            self.__rotsigma       =   _rotsigma
        else:
            self.__ele_level      =  _ele_level
            self.__readG09File( _g09_file, errorf )
            self.__ZPVE           =   self.__calcZPVE()
            
        self.__freq               =   self.__freq * self.__freqscale
        self.__dos                =   []
        self.__sos                =   []
        self.__dos_ho             =   []
        self.__sos_ho             =   []
        self.__Qe                 =   1
        self.__Qr                 =   1
        self.__Qv                 =   1
        self.__Qevr               =   1
        self.__Qt                 =   1
        self.__amass              =   self.__setAmass() 
        self.__mass               =   self.__setMass() 
        self.__islinear           =   self.__setIslinear()
        self.__IABC               =   self.__principalInertiaMoment()
        

    @property
    def name(self):             return self.__name
    @property
    def geom(self):             return self.__geom
    @property
    def symbol(self):           return self.__symbol
    @property
    def freq(self):             return self.__freq
    @property
    def freqscale(self):        return self.__freqscale
    @property
    def ele_level(self):        return self.__ele_level
    @property
    def Natom(self):            return self.__Natom
    @property
    def rotsigma(self):         return self.__rotsigma
    @property
    def opticalnum(self):       return self.__opticalnum
    @property
    def islinear(self):         return self.__islinear
    @property
    def image_freq(self):       return self.__image_freq
    @property
    def mst_file(self):         return  self.__mst_file
    @property
    def dos(self):              return  self.__dos
    @property
    def sos(self):              return  self.__sos
    @property
    def dos_ho(self):           return  self.__dos_ho
    @property
    def sos_ho(self):           return  self.__sos_ho
    @property
    def Qe(self):               return  self.__Qe
    @property
    def Qr(self):               return  self.__Qr
    @property
    def Qv(self):               return  self.__Qv
    @property
    def Qevr(self):             return self.__Qevr     
    @property
    def Qt(self):               return  self.__Qt
    @property
    def amass(self):            return self.__amass
    @property
    def mass(self):             return self.__mass
    @property
    def IABC(self):             return self.__IABC
    @property
    def ZPVE(self):             return self.__ZPVE
    @property
    def E0(self):               return self.__E0
    @property
    def source(self):           return self.__source


    def __setIslinear( self ):
        geom = self.__geom
        if len(geom) > 2:
            dx0 = geom[1,:] - geom[0,:]
            for i in range(2, len(geom)):
                dx1 = geom[i,:] - geom[0,:]
                sin_theta = 1 - ( np.dot(dx0,dx1)/np.linalg.norm(dx0)/np.linalg.norm(dx1) )**2
                if sin_theta > 0.01:
                    return False
        return True  


    def __setAmass(self):
        tmp = np.empty([self.__Natom])
        for i in range(self.__Natom):
            tmp[i]  =  cl.cst.mass_dict[self.__symbol[i]]
        return tmp


    def __setMass(self):
        return np.sum(self.__amass)
        

    def __principalInertiaMoment(self):
        amass    =  self.__amass
        xyz      =  self.__geom
        if not self.__islinear:
            I        =  np.empty([3,3])
            I[0,0] = np.sum(amass*(xyz[:,1]**2+xyz[:,2]**2)); I[0,1] = -np.sum(amass*xyz[:,0]*xyz[:,1]);         I[0,2] = -np.sum(amass*xyz[:,0]*xyz[:,2])
            I[1,0] = I[0,1];                                  I[1,1] =  np.sum(amass*(xyz[:,0]**2+xyz[:,2]**2)); I[1,2] = -np.sum(amass*xyz[:,1]*xyz[:,2])
            I[2,0] = I[0,2];                                  I[2,1] =  I[1,2];                                  I[2,2] =  np.sum(amass*(xyz[:,0]**2+xyz[:,1]**2))   
            IABC   =  np.linalg.det(I)
        else:
            IABC    = 0
            com_xyz = np.dot(amass, xyz) / np.sum(amass)
            for i in range(len(amass)):
                IABC += amass[i] * np.sum((xyz[i,:] - com_xyz)**2)
        return IABC


    def __calcZPVE(self):
        return np.sum(self.__freq) / 2
    

    def __setE0( self, _ener_type, _energy, errorf ):
        if _ener_type == "Eele":
            return _energy  +   self.__ZPVE
        elif _ener_type == "E0K":
            return _energy
        else:
            errorf ( 'Error, the energy of ' + self.__name + ' is not defined.' )



    def __readG09File( self, _g09_file, errorf ):
        geom_key        =   'Standard orientation'
        rotsigma_key    =   'Rotational symmetry number'
        freq_key        =   'Frequencies'
        E0_key          =   'Sum of electronic and zero-point Energies'
        geom_tmp        =   []
        freq_tmp        =   []
        symbol_tmp      =   []
        en_first        =   True
        en_freq         =   True
        en_geom         =   True
        en_rotsigma     =   True
        en_E0           =   True
        ndof            =   0
        CNT             =   0
        self.__source   =   " Gaussian standard output "
        
        try:
            f               =   open( _g09_file, mode='r' )
        except:
            errorf ( ' Error when open '+_g09_file )
        cdata           =   f.readlines()
        Nimage          =   0
        for I in range(len(cdata)-1,0,-1):
            if geom_key in cdata[I] and en_geom :
                try:
                    i  =  I + 5
                    while '----' not in cdata[i]:
                        line    =   cdata[i].strip().split()
                        symbol_tmp.append(cl.cst.symbol_dict[ float(line[1])] )
                        geom_tmp.append( [float(x) for x in line[-3:]] )
                        i           =   i + 1
                    self.__symbol   =   np.array(symbol_tmp)
                    self.__Natom    =   len(self.__symbol)
                    self.__geom     =   np.array(geom_tmp)*cl.cst.A2au
                    CNT             =   CNT + 1
                    en_geom         =   False
                except:
                    errorf(' Error when read the geometry of the Gaussian09 output file '+_g09_file)
            elif rotsigma_key in cdata[I] and en_rotsigma:
                try:
                    self.__rotsigma =   float(cdata[I].strip().split()[-1])
                    CNT             =   CNT + 1
                    en_rotsigma     =   False
                except:
                    errorf(' Error when read the rotational symmetry number of the Gaussian09 output file '+_g09_file)
            elif freq_key in  cdata[I] and en_freq:
                try:
                    for x in cdata[I].strip().split('--')[-1].split():
                        if float(x) > 0:
                            freq_tmp.append( float(x)*cl.cst.cm2au )
                            ndof += 1
                        else:
                            Nimage += 1
                            ndof   += 1
                            self.__image_freq = 0 - float(x)*cl.cst.cm2au
                            if Nimage > 1:
                                errorf(_g09_file," have more than one image frequency")
                except:
                    errorf(' Error when read the frequencies of the Gaussian09 output file '+_g09_file)
                if en_first:
                    Ndof     =  int(cdata[ I - 2 ].strip().split()[-1])
                    en_first =  False

                if ndof == Ndof:    
                    self.__freq     =   np.array(freq_tmp)
                    en_freq         =   False
                    CNT             =   CNT + 1

            elif E0_key in cdata[I] and en_E0:
                try:
                    self.__E0       =   float(cdata[I].strip().split()[-1])
                    CNT             =   CNT + 1
                    en_E0           =   False
                except:
                    errorf(' Error when read the OK energy with ZPE of the Gaussian09 output file '+_g09_file)
            if  CNT == 4:
                break

        #check if read sucessfully. if rotsigma is not read, it will be the default value 1. 
        if en_geom:
            errorf( _g09_file + " has no geometry info. ")
        if en_freq is None:
            errorf( _g09_file + " has no frequency info. ")
        if en_E0 is None:
            errorf( _g09_file + " has no E0 info. ")
        

    
    def set_E0  (self,x):           self.__E0           = x
    def set_dos (self,x):           self.__dos          = x
    def set_sos (self,x):           self.__sos          = x
    def set_rotsigma(self,x):       self.__rotsigma     = x
    def set_mst_file(self,x):       self.__mst_file     = x
    def set_opticalnum(self,x):     self.__opticalnum   = x


    def setState( self, dE, Emax, errorf = print, HO = False ):
        def inlineRotSoS(E):
            'calculate Nrot(E) from classical result'
            if self.__islinear:
                Nr  =   E   *2  *self.__IABC   /cl.cst.hba**2
            else:
                Nr  =   (4.0 / 3.0) * np.sqrt( ( 2 *E /cl.cst.hba**2 )**3 * self.__IABC)
            return Nr

        @nb.jit(nopython=True)
        def inlineVibDoS(_dos, _freq_bin, _MAX):
            for k in range(_freq_bin.shape[0]):
                for i in range( _MAX - _freq_bin[k] ):
                    _dos[ _freq_bin[k] + i] += _dos[i]
            return _dos

        def inlineState( _freq, _rotsigma, _ele_level, _Emax, _dE):
            #calculate dos
            MAX         =   int( round( _Emax / _dE ) + 1 )
            freq_bin    =   np.sort( [ int( round( x / _dE) )  for x in _freq ] )
            #rotation convovle
            arr_1 = np.arange( 0,MAX  ,1) * _dE
            arr_2 = np.arange(-1,MAX-1,1) * _dE    ;   arr_2[0] = 0
            _dos  =   (inlineRotSoS(arr_1) - inlineRotSoS(arr_2)) / _rotsigma / _dE
            #vibration convovle
            _dos  =   inlineVibDoS(_dos, freq_bin, MAX)
            #eletronic convovle
            for x in _ele_level:
                degen                   =   x[0]
                ener                    =   x[1]
                i0                      =   int(round( ener /  _dE ))
                _dos[ i0 : ]            =   _dos[  : MAX-i0 ] *  degen
            #optical number
            _dos = _dos * self.__opticalnum
            #sos
            return _dos,  np.cumsum( _dos ) * _dE

        if HO:
            if self.__mst_file == None:
                self.__sos_ho           =  self.__sos
                self.__dos_ho           =  self.__dos
            else:
                self.__dos_ho, self.__sos_ho   =   inlineState( self.__freq, self.__rotsigma, self.__ele_level, Emax, dE)
        else:
            if self.__mst_file == None:
                self.__dos, self.__sos      =  inlineState( self.__freq, self.__rotsigma, self.__ele_level, Emax, dE)
            else:
                #read dos from mst
                try:
                    f = open(self.__mst_file)
                except IOError or PermissionError:
                    errorf( ' Can not open MSTor output file '+self.__mst_file )

                line = f.readline()
                i    = 0 
                while 'Sum of states and density of states' not in line:
                    i        =  i + 1
                    line     =  f.readline()
                    if not line:
                        errorf( self.__mst_file + ' does not have the density of state.' )
                        exit(1)
                f.close()

                cdata        = np.loadtxt( self.__mst_file, comments='=', skiprows = i + 8, usecols = [ 6 ])
                self.__dos   = np.append( [0], cdata[:] )
                
                #check
                if len(self.__dos) > int(Emax//dE * 1.1 ):
                    errorf( ' Number of density of states loaded from MSTor is much more than expected. User should check whether dE and Emax are the same in TUMME and MSTor.' )
                elif len(self.__dos) < int(Emax//dE * 0.9 ):
                    errorf( ' Number of density of states loaded from MSTor is much less than expected. User should check whether dE and Emax are the same in TUMME and MSTor.' )

                #check NaN
                dl.CheckNaN( self.__dos )

                #count electronic dos
                for x in self.__ele_level:
                    degen                   =   x[0]
                    ener                    =   x[1]
                    i0                      =   int(round( ener /  dE ))
                    self.__dos[i0 + 1:-1 ]  =   self.__dos[ 1 :-(i0+1) ] *  degen
                
                #optical number
                self.__dos    =  self.__dos * self.__opticalnum
                
                #sos
                self.__sos    =  np.cumsum( self.__dos ) * dE


    def setPartitionfunc(self, dE):
        T   =   self.T
        #roational Q
        if self.islinear:
            self.__Qr = 8*(np.pi**2)*cl.cst.kb*T*self.__IABC/cl.cst.h**2/self.__rotsigma
        else:
            self.__Qr = (np.sqrt(np.pi)/cl.cst.h**3) * np.power(8*(np.pi**2)*cl.cst.kb*self.T,3/2)*np.sqrt(self.__IABC) / self.__rotsigma
        #vibrational Q
        self.__Qv     = np.prod( 1.0 / ( 1 - np.exp( 0 - np.array(self.__freq) / (cl.cst.kb * T) ) ))
        #electronic Q
        Qe = 0
        for x in self.__ele_level:
             Qe += x[0] * np.exp( -x[1] / cl.cst.kb / T )
        self.__Qe =Qe
        #translational Q
        self.__Qt   =   np.power( 2 * np.pi * self.__mass * cl.cst.kb * T / cl.cst.h**2, 3/2 )
        #eletronic-vibrational-rotational partition function Qevr
        if self.__mst_file is not None:
            self.__Qevr = np.dot( self.__dos, np.exp( -np.arange( self.__dos.size ) * dE / (cl.cst.kb * T) ) ) * dE
        else:
            self.__Qevr = self.__Qe * self.__Qv * self.__Qr
    

    def printProperty(self, printf):
        printf(' [MOLECULE PROPERTIES]')
        printf('%-30s'%('  Name:')+self.__name )
        printf('%-30s'%('  Read from: ')+self.__source )
        printf('%-30s'%('  Number of atoms:')+str(self.__Natom))
        printf('%-30s'%('  Geometry [a.u.]:') )
        for i in range(self.__Natom):
            printf( '%10s'%(self.__symbol[i])+'%17.5E'%(self.__geom[i,0])+'%17.5E'%(self.__geom[i,1])+'%17.5E'%(self.__geom[i,2]) )
        printf('%-30s'%('  Freuencies scaling factor:')+'%-15.5E'%(self.__freqscale))
        printf('%-30s'%('  Frequencies (scaled) [cm-1]:'))
        freq_str = '%3s'%(' ')
        for i in range(self.__freq.size):
            freq_str += '%17.5E'%(self.__freq[i]*cl.cst.au2cm)
            if  i % 4 == 3:
                freq_str += '\n'+'%3s'%(' ')
        printf(freq_str)
        printf('%-30s'%('  E(ele) [ a.u.]:')+'%-15s'%(str(self.__E0 - self.__ZPVE)))
        printf('%-30s'%('  ZPVE  [ a.u.]:')+'%-15s'%(str(self.__ZPVE)))
        printf('%-30s'%('  E(0K) [ a.u.]:')+'%-15s'%(str(self.__E0)))
        printf('%-30s'%('  Eletronic level [a.u.]:'))
        for x in self.__ele_level:
            printf('%3s'%(' ')+'%17.5E'%(x[0]) + '%17.5E'%(x[1]) )
        printf('%-30s'%('  Islinear:')+'%-15s'%(str(self.__islinear)) )
        printf('%-30s'%('  IABC [a.u.]:')+'%-15.5E'%(self.__IABC) )
        printf('%-30s'%('  Symmetry number:')+'%-15.5E'%(self.__rotsigma) )
        printf('%-30s'%('  Optical-isomer number:')+'%-15.5E'%(self.__opticalnum) )
        printf('%-30s'%('  MSTor file:')+'%-15s'%(str(self.__mst_file)) )
        printf(' [END OF MOLECULE]')
        printf(' ')
            
        