#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import numpy           as      np
from   const_lib       import  cst
import time

class std_out_class():
    __cycle_cnt = 0

    def __init__(self):
        self.__head            =    '''# 
#   TUMME version 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@mails.tsinghua.edu.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#
#   
'''

        self.__head           +=    '*   Run at ' + time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()) + '\n'
        self.__file_stdout     =    None 
        self.__file_rate       =    None 
        self.__file_eval       =    None
        self.__file_evec       =    None
        self.__file_partifunc  =    None
        self.__file_kE         =    None  
        self.__file_timeevol   =    None
        self.__file_relaxkernel=    None
        self.__stdout_buff     =    []


    def set_stdout_file     (self,x):           self.__file_stdout      =   open(x+'.out', mode='w+')
    def set_rate_file       (self,x):           self.__file_rate        =   open(x+'.rate', mode='w+')
    def set_evals_file      (self,x):           self.__file_eval        =   open(x,mode='w+')
    def set_evecs_file      (self,x):           self.__file_evec        =   open(x,mode='w+')
    def set_partifunc_file  (self,x):           self.__file_partifunc   =   open(x,mode='w+')
    def set_kE_file         (self,x):           self.__file_kE          =   open(x,mode='w+')
    def set_timeevol_file   (self,x):           self.__file_timeevol    =   open(x,mode='w+')
    def set_relaxkernel_file(self,x):           self.__file_relaxkernel =   open(x,mode='w+')
    
    @property
    def head          (self): return self.__head 
    @property
    def file_stdout   (self): return self.__file_stdout   
    @property
    def file_rate     (self): return self.__file_rate 
    @property
    def file_eval     (self): return self.__file_eval       
    @property
    def file_evec     (self): return self.__file_evec      
    @property
    def file_partifunc(self): return self.__file_partifunc 
    @property
    def file_kE       (self): return self.__file_kE     
    @property
    def file_timeevol (self): return self.__file_timeevol   
    @property
    def file_relaxkernel(self): return self.__file_relaxkernel
    @property
    def stdout_buff   (self): return self.__stdout_buff   

    def stdout(self, _str):
        self.__stdout_buff.append ( _str + '\n' )

    def flush_stdout(self, _content=None):
        'print out all stdout'
        if _content is None:
            _content = self.__stdout_buff
        self.__file_stdout.write( self.__head )
        for line_list in _content:
            for line in line_list:
                self.__file_stdout.write( line )
        self.__file_stdout.flush()


    def printGlobalParameter( self, gvr):
        'print out global parameters'
        self.stdout( '*********************** Global Parameter ***********************' )
        self.stdout('%-30s'%('  Precision:')          + gvr.precision        )
        self.stdout('%-30s'%('  MergeThreshold:')     + '%.5E'%(gvr.merge_threshold) )
        self.stdout('%-30s'%('  GroundSpecies:')      + gvr.gndSpecies       )
        self.stdout('%-30s'%('  <DeltaE_d>[a.u.]  =  (T/Tc)^(exponent)[ slope*( E - E_gamma,0) + EdConst ]') )
        self.stdout('%-30s'%('     Tc[a.u.]:')        + '%.5E'%(gvr.Tc) )
        self.stdout('%-30s'%('     exponent:')        + '%.5E'%(gvr.exponent) )
        self.stdout('%-30s'%('     slope:')           + '%.5E'%(gvr.slope) )
        self.stdout('%-30s'%('     EdConst[a.u.]:')   + '%.5E'%(gvr.EdConst) )
        if gvr.collid.name == 'LJ':
            self.stdout('%-30s'%('  LJCollision')  )
            self.stdout('%-30s'%('     EpsilonM[a.u.]:')        + '%.5E'%(gvr.collid.epsilonM)  )
            if gvr.collid.epsilonA is None:
                self.stdout('%-30s'%('     EpsilonA[a.u.]:')    + 'None')
            else:
                self.stdout('%-30s'%('     EpsilonA[a.u.]:')    + '%.5E'%(gvr.collid.epsilonA))
            self.stdout('%-30s'%('     diameterM[a.u.]:')       + '%.5E'%(gvr.collid.diameterM) )
            if gvr.collid.diameterA is None:
                self.stdout('%-30s'%('     diameterA[a.u.]:')   + 'None')
            else:
                self.stdout('%-30s'%('     diameterA[a.u.]:')   + '%.5E'%(gvr.collid.epsilonA))
            self.stdout('%-30s'%('     massM[a.u.]:')           + '%.5E'%(gvr.collid.massM   )  )
            if gvr.collid.massA is None:
                self.stdout('%-30s'%('     massA[a.u.]:')       + 'None')
            else:
                self.stdout('%-30s'%('     massA[a.u.]:')       + '%.5E'%(gvr.collid.massA))
            self.stdout('')
            
        if gvr.collid.name == 'HS':
            self.stdout('%-30s'%('  HSCollision')  )           
            if gvr.collid.bmax is None:
                self.stdout('%-30s'%('     bmax[a.u.]:')        + 'None')
            else:
                self.stdout('%-30s'%('     bmax[a.u.]:')        + '%.5E'%(gvr.collid.bmax) )
            self.stdout('%-30s'%('     massM[a.u.]:')           + '%.5E'%(gvr.collid.massM))
            if gvr.collid.massA is None:
                self.stdout('%-30s'%('     massA[a.u.]:')       + 'None')
            else:
                self.stdout('%-30s'%('     massA[a.u.]:')       + '%.5E'%(gvr.collid.massA))
            self.stdout('')


    def printRxnStack( self, rxn_stack ):
        'print out rxn_stack'
        self.stdout( '************************ Reaction Stack ************************' )
        for rxn in rxn_stack:
            rxn.printProperty( self.stdout )


    def printRxnMap( self, gvr):
        'print out rxn_map'
        rxn_map     =   gvr.rxn_map
        ph_name     =   [x.name for x in gvr.phase_stack]
        self.stdout( '************************* Reaction Map *************************' )
        self.stdout('( Barrier with ZPVE in a.u.; X denotes none reaction )')
        name_str = '%12s'%('From/to')
        for x in ph_name:
            name_str += '%16s'%x
        self.stdout(name_str)
        for i in range( len(rxn_map) ):
            line_str = '%-12s'%(ph_name[i])
            for j in range( len(rxn_map) ):
                if rxn_map[i][j]:
                    line_str += '%16.5E'%(rxn_map[i][j].barrier)
                else:
                    line_str += '%16s'%('X')
            self.stdout(line_str)
        self.stdout('')    

        self.stdout('(  Reaction energy with ZPVE in a.u.; X denotes none reaction )')
        name_str = '%12s'%('From/to')
        for x in ph_name:
            name_str += '%16s'%x
        self.stdout(name_str)
        for i in range( len(rxn_map) ):
            line_str = '%-12s'%(ph_name[i])
            for j in range( len(rxn_map) ):
                if rxn_map[i][j]:
                    line_str += '%16.5E'%(rxn_map[i][j].rxn_ener)
                else:
                    line_str += '%16s'%('X')
            self.stdout(line_str)
        self.stdout('\n')
        self.stdout( '********************* Solve master equation *********************\n'  )


    def printIsomer( self, well_stack ):
        'print out isomers property'
        self.stdout( '**************************** Isomers ****************************' )
        for well in well_stack:
            well.printProperty( self.stdout )
        self.stdout('')


    def printTransState( self, ts_stack ):
        'print out transition state property'
        self.stdout( '*********************** Transition States ***********************' )
        for ts in ts_stack:
            ts.printProperty( self.stdout )
        self.stdout('')


    def printBimolecule(self, bim_stack):
        'print out bimolecule property'
        self.stdout( '************************** Bimolecular **************************' )
        for bim in bim_stack:
            self.stdout('[BIMOLECULAR PROPERTIES]')
            self.stdout('%-30s'%('  Name:')+bim.name)
            self.stdout('%-30s'%('  E(ele) [ a.u.]:')+'%-15s'%(str(bim.sp1.E0+bim.sp2.E0-bim.sp1.ZPVE-bim.sp2.ZPVE)))
            self.stdout('%-30s'%('  E(0K) [ a.u.]:')+'%-15s'%(str(bim.E0)))
            self.stdout('%-30s'%('  Reduced mass [ a.u.]:')+'%-15s'%(str(bim.mu)))
            bim.sp1.printProperty( self.stdout )
            bim.sp2.printProperty( self.stdout )
            self.stdout('[END OF BIMOLECULAR]')
        
        if len(bim_stack) == 0:
            self.stdout( '  No bimolecular pair       ' )
        self.stdout('')


    def printCycleparam(self, gvr):
        T           =   gvr.cycle_param.T
        P           =   gvr.cycle_param.p
        ener_step   =   gvr.cycle_param.ener_step
        ener_ref    =   gvr.cycle_param.ener_ref

        self.stdout( '==================  Cycle  ==================\n'  )
        self.stdout('%-30s'%('  Temperature[K]:')           +'%.3E'%(T * cst.au2K)       )
        self.stdout('%-30s'%('  Pressure[Torr]:')           +'%.3E'%(P * cst.au2torr)    )
        self.stdout('%-30s'%('  Energy step[cm-1]:')        +'%.5E'%(ener_step*cst.au2cm))
        self.stdout('%-30s'%('  Energy reference[cm-1]:')   +'%.5E'%(ener_ref*cst.au2cm) )
        self.stdout('')
        std_out_class.__cycle_cnt += 1


    def printKernelinfo(self, well_stack):
        'print kernel of all isomers'
        for well in well_stack:
            self.stdout('[ RELAXATION INFO ]')
            self.stdout('%-30s'%('  Name:')                     +'%-8s'% well.name               )
            self.stdout('%-30s'%('  Size:')                     +'%-8d'%(well.size)              )
            if well.epsilon  is not None:
                self.stdout('%-30s'%('  Epsilon[a.u.]:')        +'%-8f'%(well.epsilon)           )
            if well.diameter is not None:
                self.stdout('%-30s'%('  Diameter[a.u.]:')       +'%-8f'%(well.diameter)          )
            if well.bmax     is not None:
                self.stdout('%-30s'%('  bmax[a.u.]:')           +'%-8f'%(well.bmax)              )
            self.stdout('%-30s'%('  Z[1/s]:')                   +'%-8E'%(well.collid.Z/cst.au2s) )
            self.stdout('%-30s'%('  Truncated well E0[cm-1]:')  +'%-8.3f'%(well.E0_t*cst.au2cm)  )
            self.stdout('%-30s'%('  True well E0:[cm-1]')       +'%-8.3f'%(well.E0*cst.au2cm)    )
            self.stdout('[ END OF INFO ]\n')


    def printMEinfo( self, ME, ph_name ):
        Nwell    =   ME.epcs.shape[0]
        Nbim     =   len(ph_name) - Nwell

        self.stdout('[ ME INFO ]')
        self.stdout('%-30s'%('  Size')                      + '%d'%(ME.size)             )
        self.stdout('%-30s'%('  Min CSE eigenvalue[1/s]:')  + '%.5E'%(ME.L[0]/cst.au2s) )
        self.stdout('%-30s'%('  Max CSE eigenvalue[1/s]:')  + '%.5E'%(ME.L[Nwell-1]/cst.au2s))
        self.stdout('%-30s'%('  Min IERE eigenvalue[1/s]:') + '%.5E'%(ME.L[Nwell]/cst.au2s) )
        self.stdout('%-30s'%('  Max IERE eigenvalue[1/s]:') + '%.5E'%(ME.L[-1]/cst.au2s))

        #print EPCS
        self.stdout('')
        self.stdout('%-30s'%('  EPCS:') )
        cse_tmp  =   '%-30s'%('    CSE No.')
        for i in range(Nwell):
            cse_tmp  += '%-14d'%(i+1)
        self.stdout(cse_tmp)
        for i in range(Nwell):
            val_tmp =   '    '+'%-26s'%(ph_name[i])
            for j in range(Nwell):
                val_tmp += '%-14.5E'%(ME.epcs[i,j])
            self.stdout(val_tmp)

        #print kappa
        if Nbim:
            self.stdout('')
            self.stdout('%-30s'%('  Kappa:') )
            bim_tmp  =  '%-30s'%('    Bim Name')
            for j in range(Nbim):
                bim_tmp += '%-14s'%(ph_name[Nwell+j])
            self.stdout( bim_tmp )
            for i in range(Nwell):
                val_tmp =   '    '+'%-26s'%(ph_name[i])
                for j in range(Nbim):
                    val_tmp += '%-14.5E'%(ME.kappa[i,j])
                self.stdout( val_tmp )

        #print eigenmode and well correspondence 
        self.stdout('')
        self.stdout( '  Binding:' )
        for idx_w in ME.well_eig:
            self.stdout( '    %s'%(ph_name[idx_w]) +' well corresponds to eigen mode %d'%( ME.well_eig[idx_w] + 1 ) )

        #print merge pairs
        if ME.merge_well:
            self.stdout('')
            self.stdout( '  Merge Pairs:' )
            for x in ME.merge_well:
                self.stdout( '    '+'%5s'%( ph_name[x] )+' - '+'%5s'%( ph_name[ ME.merge_well[x] ] ) )
        self.stdout('')
        self.stdout('[ END OF INFO ]\n')
        
        
    def printRate( self, gvr ):
        ME_stack    =   gvr.ME_stack
        NT          =   len(gvr.temperature)
        NP          =   len(gvr.pressure)
        p_list      =   gvr.pressure
        T_list      =   gvr.temperature
        ph_name     =   [x.name for x in gvr.phase_stack]
        Nwell       =   len(gvr.well_stack)
        Nph         =   len(ph_name)

        f           =   self.__file_rate
        f.write(self.__head)
        f.write('\n\n\n')
        f.write( '* The unit for isomer-to-isomer and isomer-to-bimolecular rate constants is s^(-1).\n')
        f.write( '* The unit for bimolecular-to-isomer and bimolecular-to-bimolecular rate constants is cm^3*molecule^(-1)*s^(-1).\n')
        f.write( '* # denotes this rate constant does not exist due to mergers or denotes this rate constant is approximately zero.\n')
        f.write( '* NAG denotes negtive value, this is cause by numerical problem, try high-precision option.')
        f.write('\n\n')

        f.write('********************* Rate constants map *********************\n')
        f.write('*   The diagonal denotes sum of the off-diagonal in a row.\n\n')
        name_str = '%-20s'%(' From/To')
        for x in ph_name:
            name_str += '%15s'%x

        for ME in ME_stack:
            J   =   0
            f.write(' T = %.2f'%(ME.T*cst.au2K)+' K, p = %.3E'%(ME.p*cst.au2torr)+'\n')
            f.write(name_str+'\n')
            for i in range(Nph):
                val_str = '%-20s'%(ph_name[i])

                if i < Nwell:
                    unit = 1 / cst.au2s
                else:
                    unit = (cst.au2m*cst.m2cm)**3 / cst.au2s 

                for  j in range(Nph):
                    if ME.K[i,j] == 0:
                        val_str += '%15s'%('#') 
                    elif ME.K[i,j] < 0:
                        val_str += '%15s'%('NAG')
                    else:
                        val_str += '%15.5E'%(ME.K[i,j]*unit)
                f.write(val_str+'\n')
            f.write('\n')
        
        f.write('\n\n********************* Rate constants in pressure *********************\n\n') 
        title = '%-20s'%('p[torr]/T[K]')
        for x in gvr.temperature:
            title += '%15.2f'%(x*cst.au2K)

        for i in range(Nph):
            if i < Nwell:
                unit = 1 / cst.au2s
            else:
                unit = (cst.au2m*cst.m2cm)**3 / cst.au2s 

            for j in range(Nph):
                if i == j:
                    f.write( '%-20s'%(ph_name[i]+' -> other species\n' ))
                else:
                    f.write( '%-20s'%(ph_name[i]+' -> '+ph_name[j])+'\n' )
                f.write(title +'\n')
                for ii in range(NP):
                    val_str =   '%-20.3E'%(p_list[ii]*cst.au2torr)
                    for jj in range(NT):
                        ME  =   ME_stack[ ii + jj*NP ]
                        if ME.K[i,j] == 0:
                            val_str += '%15s'%('#') 
                        elif ME.K[i,j] < 0:
                            val_str += '%15s'%('NAG')
                        else:
                            val_str += '%15.5E'%(ME.K[i,j]*unit)
                    f.write(val_str+'\n')
                f.write('\n')
        
        f.write('\n\n********************* Rate constants in temperature *********************\n\n') 
        title = '%-20s'%('T[K]/p[torr]')
        for x in gvr.pressure:
            title += '%15.3E'%(x*cst.au2torr)

        J = 0
        for i in range(Nph):
            if i < Nwell:
                unit = 1 / cst.au2s
            else:
                unit = (cst.au2m*cst.m2cm)**3 / cst.au2s 

            for j in range(Nph):
                if i == j:
                    f.write( '%-20s'%(ph_name[i]+' -> other species\n' ))
                else:
                    f.write( '%-20s'%(ph_name[i]+' -> '+ph_name[j])+'\n' )
                f.write(title +'\n')
                for ii in range(NT):
                    val_str =   '%-20.3E'%(T_list[ii]*cst.au2K)
                    for jj in range(NP):
                        ME      =   ME_stack[ jj + ii*NP ]
                        if ME.K[i,j] == 0:
                            val_str += '%15s'%('#')
                        elif ME.K[i,j] < 0:
                            val_str += '%15s'%('NAG') 
                        else:
                            val_str += '%15.5E'%(ME.K[i,j]*unit)
                    f.write(val_str+'\n')
                f.write('\n')                   

        f.write('\n\n********************* High-pressure limit flux coefficient *********************\n\n') 
        title = '%-20s'%('T[K]')
        for rxn_list in gvr.rxn_map:
            for rxn in rxn_list:
                if rxn is not None:
                    title += '%15s'%(rxn.Rname+'->'+rxn.Pname)
        f.write( title + '\n' )

        for t in range(NT):
            val_str         =   '%-20.3E'%(T_list[t]*cst.au2K)
            ME              =   ME_stack[ t*NP ]
            well_offset     =   np.zeros( Nwell+1, dtype=np.int32)
            well_offset[1:] =   np.cumsum(ME.well_sz,dtype=np.int32)  
            for i in range(len(gvr.rxn_map)):
                J = 0
                for j in range(len(gvr.rxn_map)):
                    if gvr.rxn_map[i][j] is not None:
                        if i <  Nwell:
                            unit = 1 / cst.au2s 
                            bolztmann   =   ME.F[ well_offset[i]:well_offset[i]+ME.well_sz[i] ]**2
                            k_hpl = unit * np.dot( ME.kE_map[i][J], bolztmann ) / np.sum( bolztmann )
                        else:
                            unit  = (cst.au2m*cst.m2cm)**3 / cst.au2s 
                            k_hpl = unit * np.sum( ME.kE_map[i][J] )       
                        val_str += '%15.5E'%(k_hpl)
                        J = J + 1
            f.write(val_str+'\n')
        f.write('\n')
        f.flush()  
        f.close()


    def printEval(self, gvr):
        if self.__file_eval:
            ME_stack    =   gvr.ME_stack
            Nwell       =   len( gvr.well_stack )

            f           =   self.file_eval
            f.write('#\n# Eigenvalue of the symmetrized transition matrix G (excluding accosiations of pseudo-first-order bimoleculars) \n#')
            f.write('\n')

            if gvr.printevalnum is None:
                f.write('\n***************************** Eigenvalue (1/s) *****************************\n\n')
                for I in range(Nwell):
                    f.write(' [ '+gvr.well_stack[I].name+' ] \n')
                    title       =   '%-10s'%('T(K)') + '%-10s'%('P(torr)') + '%15s'%('Z(1/s)')
                    for i in range(Nwell):
                        title  +=   '%15s'%('E'+str(i))
                    title      +=   '%15s'%('Rmin') + '%15s'%('Rmax')
                    f.write(title+'\n')
                    val_str     =   ''
                    for ME in ME_stack:
                        val_str+=   '%-10.2f'%(ME.T*cst.au2K)
                        val_str+=   '%-10.3E'%(ME.p*cst.au2torr)
                        val_str+=   '%15.5E'%(ME.Z[I]/cst.au2s)
                        for i in range(Nwell):
                            val_str += '%15.5E'%(ME.L[i]/cst.au2s)
                        val_str+=   '%15.5E'%(ME.L[Nwell]/cst.au2s)+'%15.5E'%(ME.L[-1]/cst.au2s)
                        f.write(val_str+'\n')
                        val_str =   ''
                    f.write('Ei denotes i-th CSE  eigenvalues \n')
                    f.write('Rmin denotes the minimum IERE eigenvalues \n')
                    f.write('Rmax denotes the maximum IERE eigenvalues \n')
                    f.write('\n')
                f.write('\n\n******************************* Eigenvalue/Z *******************************\n\n')
                for I in range(Nwell):
                    f.write(' [ '+gvr.well_stack[I].name+' ] \n')
                    title       =   '%-10s'%('T(K)') + '%-10s'%('P(torr)') + '%15s'%('Z(1/s)')
                    for i in range(Nwell):
                        title  +=   '%15s'%('E'+str(i)+'/Z')
                    title      +=   '%15s'%('Rmin/Z') + '%15s'%('Rmax/Z')
                    f.write(title+'\n')
                    val_str     =   ''
                    for ME in ME_stack:
                        val_str+=   '%-10.2f'%(ME.T*cst.au2K)
                        val_str+=   '%-10.3E'%(ME.p*cst.au2torr)
                        val_str+=   '%15.5E'%(ME.Z[I]/cst.au2s)
                        for i in range(Nwell):
                            val_str += '%15.5E'%(ME.L[i]/ME.Z[I])
                        val_str+=   '%15.5E'%(ME.L[Nwell]/ME.Z[I])+'%15.5E'%(ME.L[-1]/ME.Z[I])
                        f.write(val_str+'\n')
                        val_str =   ''
                    f.write('Ei denotes i-th CSE  eigenvalues \n')
                    f.write('Rmin denotes the minimum IERE eigenvalues \n')
                    f.write('Rmax denotes the maximum IERE eigenvalues \n')
                    f.write('\n')
            else:
                eig_num = int(gvr.printevalnum)
                f.write('\n\n***************************** Eigenvalue (1/s) *****************************\n\n')
                for I in range(Nwell):
                    f.write(' [ '+gvr.well_stack[I].name+' ] \n')
                    title       =   '%-10s'%('T(K)') + '%-10s'%('P(torr)') + '%15s'%('Z(1/s)')
                    for i in range(eig_num):
                        title  +=   '%15s'%('E'+str(i))
                    f.write(title+'\n')
                    val_str     =   ''
                    for ME in ME_stack:
                        val_str+=   '%-10.2f'%(ME.T*cst.au2K)
                        val_str+=   '%-10.3E'%(ME.p*cst.au2torr)
                        val_str+=   '%15.5E'%(ME.Z[I]/cst.au2s)
                        for i in range(eig_num):
                            val_str += '%15.5E'%(ME.L[i]/cst.au2s)
                        f.write(val_str+'\n')
                        val_str =   ''
                    f.write('Ei denotes i-th least eigenvalue \n')
                    f.write('\n')
                f.write('\n\n******************************* Eigenvalue/Z *******************************\n\n')
                for I in range(Nwell):
                    f.write(' [ '+gvr.well_stack[I].name+' ] \n')
                    title       =   '%-10s'%('T(K)') + '%-10s'%('P(torr)') + '%15s'%('Z(1/s)')
                    for i in range(eig_num):
                        title  +=   '%15s'%('E'+str(i)+'/Z')
                    f.write(title+'\n')
                    val_str     =   ''
                    for ME in ME_stack:
                        val_str+=   '%-10.2f'%(ME.T*cst.au2K)
                        val_str+=   '%-10.3E'%(ME.p*cst.au2torr)
                        val_str+=   '%15.5E'%(ME.Z[I]/cst.au2s)
                        for i in range(eig_num):
                            val_str += '%15.5E'%(ME.L[i]/ME.Z[I])
                        f.write(val_str+'\n')
                        val_str =   ''
                    f.write('Ei denotes i-th least eigenvalue \n')
                    f.write('\n')

            f.write('\n\n*******************  Projection to Relaxational Subspace  *******************\n\n')
            title       =   '%-10s'%('T(K)') + '%-10s'%('P(torr)')
            for i in range(Nwell):
                title  +=   '%15s'%('Prj'+str(i))
            f.write(title+'\n')
            val_str     =   ''
            for ME in ME_stack:
                val_str+=   '%-10.2f'%(ME.T*cst.au2K)
                val_str+=   '%-10.3E'%(ME.p*cst.au2torr)
                for i in range(Nwell):
                    val_str += '%15.5E'%(ME.epsrs[i])
                f.write(val_str+'\n')
                val_str = ''
            f.write('Pi denotes i-th CSE eigenvector projection squared on relaxation space.\n\n')

            if gvr.pseubimname:
                Ncse = Nwell+len(gvr.pseubimname)
                sp_name = [x.name for x in gvr.well_stack] + gvr.pseubimname
                f.write('#\n# Eigenvalue of the symmetrized transition matrix G* (including accosiations of pseudo-first-order bimoleculars) \n#\n\n')
                if gvr.printevalnum is None:
                    f.write('\n***************************** Eigenvalue (1/s) *****************************\n\n')
                    for I in range(Ncse):
                        f.write(' [ '+sp_name[I]+' ] \n')
                        title       =   '%-10s'%('T(K)') + '%-10s'%('P(torr)')
                        for i in range(Ncse):
                            title  +=   '%15s'%('E'+str(i))
                        title      +=   '%15s'%('Rmin') + '%15s'%('Rmax')
                        f.write(title+'\n')
                        val_str     =   ''
                        for ME in ME_stack:
                            val_str+=   '%-10.2f'%(ME.T*cst.au2K)
                            val_str+=   '%-10.3E'%(ME.p*cst.au2torr)
                            for i in range(Ncse):
                                val_str += '%15.5E'%(ME.L_p[i]/cst.au2s)
                            val_str+=   '%15.5E'%(ME.L_p[Ncse]/cst.au2s)+'%15.5E'%(ME.L_p[-1]/cst.au2s)
                            f.write(val_str+'\n')
                            val_str =   ''
                        f.write('Ei denotes i-th CSE  eigenvalues \n')
                        f.write('Rmin denotes the minimum IERE eigenvalues \n')
                        f.write('Rmax denotes the maximum IERE eigenvalues \n')
                        f.write('\n')
                else:
                    eig_num = int(gvr.printevalnum)
                    f.write('\n\n***************************** Eigenvalue (1/s) *****************************\n\n')
                    for I in range(Ncse):
                        f.write(' [ '+sp_name[I]+' ] \n')
                        title       =   '%-10s'%('T(K)') + '%-10s'%('P(torr)')
                        for i in range(eig_num):
                            title  +=   '%15s'%('E'+str(i))
                        f.write(title+'\n')
                        val_str     =   ''
                        for ME in ME_stack:
                            val_str+=   '%-10.2f'%(ME.T*cst.au2K)
                            val_str+=   '%-10.3E'%(ME.p*cst.au2torr)
                            for i in range(eig_num):
                                val_str += '%15.5E'%(ME.L_p[i]/cst.au2s)
                            f.write(val_str+'\n')
                            val_str =   ''
                        f.write('Ei denotes i-th least eigenvalue \n')
                        f.write('\n')
            f.close()


    def printEvec(self, gvr):
        if self.__file_evec:
            ME_stack    =   gvr.ME_stack
            Nwell       =   len( gvr.well_stack )

            f           =   self.file_evec
            f.write('#\n# Eigenvector ouput for TUMME \n#')
            f.write('\n\n')

            if gvr.printevecnum is None:
                eig_num =   Nwell
            else:
                eig_num     =   int(gvr.printevecnum)

            f.write('\n\n******************* Eigenvector *******************\n\n')
            for ME in ME_stack:
                ener_step = ME.ener_step
                ener_ref  = ME.ener_ref
                title   =   '%-12s'%('T =')+'%8.2f'%(ME.T*cst.au2K) + '%10s'%(' K, p = ') + '%10.3E'%(ME.p*cst.au2torr)+' torr \n'
                title  +=   '%-12s'%('E(kcal/mol)')
                for i in range(eig_num):
                    title  +=   '%15s'%( 'P'+str(i) )
                f.write(title+'\n')

                val_str          =  ''
                I                =  0
                for k in range(Nwell):
                    val_str     +=  '%12s\n'%( gvr.well_stack[k].name )
                    for i in range( ME.well_sz[k] ):
                        val_str +=  '%12.4f'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                        for J in range(eig_num):
                            val_str += '%15.5E'%( ME.U[ I, J ] )
                        f.write( val_str+'\n' )
                        val_str  = ''
                        I        =  I + 1
                    f.write('\n')

            f.write('\n\n************* Eigenvector x Sqrt(rho E) *************\n\n')
            for ME in ME_stack:
                ener_step = ME.ener_step
                ener_ref  = ME.ener_ref
                title   =   '%-12s'%('T =')+'%8.2f'%(ME.T*cst.au2K) + '%10s'%(' K, p = ') + '%10.3E'%(ME.p*cst.au2torr)+' torr \n'
                title  +=   '%-12s'%('E(kcal/mol)')
                for i in range(eig_num):
                    title  +=   '%15s'%('P'+str(i))
                f.write(title+'\n')

                val_str          =  ''
                I                =  0
                for k in range(Nwell):
                    val_str     +=  '%12s\n'%( gvr.well_stack[k].name )
                    for i in range(ME.well_sz[k]):
                        val_str +=  '%12.4f'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                        for J in range(eig_num):
                            val_str += '%15.5E'%( ME.U[ I, J ]*ME.F[ I ] )
                        f.write( val_str+'\n' )
                        val_str  = ''
                        I        =  I + 1
                    f.write('\n')

            f.write('\n\n************ Thermal population (rho E) ************\n\n')
            for ME in ME_stack:
                ener_step = ME.ener_step
                ener_ref  = ME.ener_ref
                title   =   '%-12s'%('T =')+'%8.2f'%(ME.T*cst.au2K) + '%10s'%(' K, p = ') + '%10.3E'%(ME.p*cst.au2torr)+' torr \n'
                title  +=   '%-12s'%('E(kcal/mol)')
                for i in range(eig_num):
                    title  +=   '%15s'%('P'+str(i))
                f.write(title+'\n')

                val_str          =  ''
                I                =  0
                for k in range(Nwell):
                    val_str     +=  '%12s\n'%( gvr.well_stack[k].name )
                    for i in range( ME.well_sz[k] ):
                        val_str +=  '%12.4f'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                        for J in range(eig_num):
                            val_str += '%15.5E'%( ME.F[ I ]**2 )
                        f.write( val_str+'\n' )
                        val_str  = ''
                        I        =  I + 1
                    f.write('\n')
   
            f.close()


    def printkE(self, gvr):
        if self.__file_kE:
            Nwell       =   len( gvr.well_stack  )
            ph_name     =   [ x.name for x in gvr.phase_stack ]
            Nph         =   len( gvr.phase_stack )
            NT          =   len(gvr.temperature)
            Np          =   len(gvr.pressure)
            f           =   self.file_kE
            f.write('#\n# Microcanonical flux coefficient ouput for TUMME \n#')
            f.write('\n\n')
            f.write( '* Output for unimolecular reactions is the microcanonical flux coefficients with unit of s^(-1).\n')
            f.write( '* Output for bimolecular reaction is the general microcanonical flux coefficient with unit of cm^3*molecule^(-1)*s^(-1).\n')
            f.write( '* The general microcanonical flux coefficient equals the microcanonical flux coefficient multiplying the Boltzmann distribution.' )
            f.write('\n\n')

            for t in range(NT):
                ME        = gvr.ME_stack[ t*Np ]
                ener_step = ME.ener_step
                ener_ref  = ME.ener_ref
                kE_map    = ME.kE_map
                title0    = '%-12s'%('T =')+'%8.2f'%(ME.T*cst.au2K) + '%10s'%(' K') + '\n'
                for j in range(Nph):
                    title = title0 + '%12s'%( 'E(kcal/mol)' )
                    for  rxn in gvr.rxn_map[j]:
                        if rxn:
                            title += '%15s'%( ph_name[j] + '->'+rxn.Pname ) 
                    f.write(title+'\n')

                    val_str = ''
                    if  j < Nwell:
                        unit = 1 / cst.au2s 
                        for i in range(kE_map[j][0].size):
                            val_str += '%12.4f'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                            for kE in kE_map[j]:
                                val_str += '%15.5E'%( kE[i] * unit )
                            val_str += '\n'
                    else:
                        unit = (cst.au2m*cst.m2cm)**3 / cst.au2s 
                        for i in range( max( [ x.size for x in kE_map[j] ] ) ):
                            val_str += '%12.4f'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                            for kE in kE_map[j]:
                                if i < kE.size:
                                    val_str += '%15.5E'%( kE[i] * unit )
                                else:
                                    val_str += '%15s'%(' ')
                            val_str += '\n'
                    f.write( val_str+'\n' )
                    f.write('\n')
                f.write('\n\n')

            f.close()
    

    def printTimeEvolution(self, gvr):
        if self.__file_timeevol:
            #time evolution
            f       =   self.__file_timeevol
            Nwell   =   len(gvr.well_stack)
            f.write('#\n# Time evolution ouput for TUMME \n# the unit for energy is kcal/mol; the unit for time is second \n#')
            f.write('\n\n')
            
            for ME in gvr.ME_stack:
                ener_step = ME.ener_step
                ener_ref  = ME.ener_ref
                # characteristic time scale
                charac_t  = ME.charac_t
                # time step
                if gvr.time_step == None:
                    time_step = charac_t / 300
                else:
                    time_step = gvr.time_step 
                # numbers of step
                if gvr.time_num == None:
                    Ntime = 50
                else:
                    Ntime = int( gvr.time_num )

                title  = '%-12s'%('T =')+'%8.2f'%(ME.T*cst.au2K) + '%10s'%(' K, p = ') + '%10.3E'%(ME.p*cst.au2torr)+' torr \n'
                title += '%-20s'%( 'y(E)/time' )
                for t in range( Ntime ): 
                    title += '%15.5E'%( t*time_step*cst.au2s )
                f.write(title+'\n')

                I = 0
                for j in range(Nwell):
                    f.write(gvr.well_stack[j].name + '\n')
                    val_str    =  ''
                    for i in range(ME.well_sz[j]):
                        val_str +=  '%-20.5E'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                        for t in range( Ntime ): 
                            val_str +=  '%15.5E'%( ME.y[t,I] )
                        f.write( val_str+'\n' )
                        val_str  = ''
                        I        =  I + 1    
                if gvr.pseubimname:
                    val_str  = ''
                    for i in range(len(gvr.pseubimname)):
                        f.write(gvr.pseubimname[i] + '\n')
                        val_str += '%-20s'%'[n]'
                        for t in range( Ntime ): 
                            val_str +=  '%15.5E'%( ME.y[t,I] )
                        f.write( val_str+'\n' )
                        val_str  = ''
                        I        =  I + 1 
                f.write('\n')


    def printPartitionfunction(self, gvr):
        if self.__file_partifunc:
            rxn_map             =   gvr.rxn_map
            Temperature         =   gvr.temperature
            Nph                 =   len(gvr.phase_stack)
            species_stack       =   gvr.well_stack + [x.sp1 for x in gvr.bim_stack] + [x.sp2 for x in gvr.bim_stack] + gvr.ts_stack
            Nsp                 =   len(species_stack)
            Qsp                 =   np.empty( [ len(Temperature), Nsp, 6 ] )
            Qgts                =   [ [ [ [] for i in range(Nph)] for j in range(Nph) ] for t in range(len(Temperature)) ]
            f                   =   self.__file_partifunc
            f.write('#\n# Partition function ouput for TUMME \n#')
            f.write('\n')
            f.write( '* Qt is the translation partition function. The unit for is 1/cm^3.\n')
            f.write( '* Qe is the electronic partition function.\n')
            f.write( '* Qv is the vibration partition function with single-structure-harmonic-approximation based on the E0K of each species. \n')
            f.write( '* Qv_gl is the vibration partition function with single-structure-harmonic-approximation based on the E0K of ground species. \n')
            f.write( '* Qr is the rotational partition function with single-structure-rigid-rotor approximation.\n') 
            f.write( '* Qevr is the electronic-vibrational-rotational partition function.\n') 
            f.write( '*   If MSTor file is defined, Qevr is got from the integration of rho_mst, otherwise, it is Qe*Qv*Qr.\n' )
            f.write('\n\n')

            for t,T in enumerate(Temperature):
                gvr.cycle_param.set_T ( T )
                #partition function for all stationary species
                for i in range(Nsp):
                    species       =  species_stack[i]
                    species.setThermal( gvr.dE )
                    Qsp[t,i,0]    =  species.Qt / ( ( cst.au2m*cst.m2cm )**3 )
                    Qsp[t,i,1]    =  species.Qe
                    Qsp[t,i,2]    =  species.Qv
                    Qsp[t,i,3]    =  species.Qv * np.exp(-species.E0 / cst.kb / T)
                    Qsp[t,i,4]    =  species.Qr
                    Qsp[t,i,5]    =  species.Qevr
                #partition function for all general transition state
                for i in range(Nph):
                    for j in range(i):
                        if rxn_map[i][j]:
                            if len( rxn_map[i][j].gts_dict ) != 0:
                                for  s  in rxn_map[i][j].gts_dict:
                                    rxn_map[i][j].gts_dict[s].setThermal( gvr.dE )
                                    Qgts[t][i][j].append( rxn_map[i][j].gts_dict[s].Qevr )
            
            #print stationary species
            f.write('******************** Stationary Species Partition Function ******************** \n')
            for i in range(Nsp):
                f.write('Species : '+species_stack[i].name+'\n')
                f.write( '%-12s'%('T(K)/Q')+'%-15s'%('Qt')+'%-15s'%('Qe')+'%-15s'%('Qv')+'%-15s'%('Qv_gl')+'%-15s'%('Qr')+'%-15s'%('Qevr') )
                f.write('\n')
                val_str = ''
                for  t,T in enumerate(Temperature):
                    val_str += '%-12.2f'%(T*cst.au2K)+'%-15e'%Qsp[t,i,0]+'%-15e'%Qsp[t,i,1]+'%-15e'%Qsp[t,i,2]+'%-15e'%Qsp[t,i,3]+'%-15e'%Qsp[t,i,4]+'%-15e'%Qsp[t,i,5]+'\n'
                f.write(val_str)
                f.write('\n')

            #print general transition state
            f.write('\n******************** General Transition States Partition Function ********************\n')
            for i in range(Nph):
                for j in range(i):
                    if rxn_map[i][j]:
                        val_str  = ''
                        if len(rxn_map[i][j].gts_dict) != 0:
                            f.write( 'Reaction : ' + rxn_map[i][j].reactant.name+'-'+rxn_map[i][j].ts.name+'-'+rxn_map[i][j].product.name +'\n')
                            head_str = '%-15s'%('T(K)/Qevr(s)')
                            for x in rxn_map[i][j].gts_dict:
                                head_str +=  's=%-13.2f'%x    
                            f.write(head_str)
                            f.write('\n')
                            for  t,T in enumerate(Temperature):
                                val_str += '%-15.2f'%(T*cst.au2K)
                                for x in Qgts[t][i][j]:
                                    val_str += '%-15e'%x
                                val_str += '\n'
                            f.write( val_str )
                            f.write( '\n' )

    def printRelaxKernel(self, gvr):
        if self.__file_relaxkernel:
            #relaxational collisional possibility kernel
            f       =   self.__file_relaxkernel
            f.write('#\n# Relaxational collisional possibility kernel ouput for TUMME \n#')
            f.write("# P(E|E'), where E' is the initial energy and E is the final energy. \n")
            f.write('\n\n')
            NT      =   len(gvr.temperature)
            Np      =   len(gvr.pressure)

            for t in range(NT):
                ME          =   gvr.ME_stack[ t*Np ]
                ener_step   =   ME.ener_step
                ener_ref    =   ME.ener_ref
                title0      =  '%-12s'%('# T = ')+'%8.2f'%(ME.T*cst.au2K) + ' K. \n\n'
                f.write( title0 )
                
                for welli, name in enumerate(gvr.printkernelname):
                    kernel  = ME.well_kernel[welli]
                    well_sz = kernel.shape[0]
                    title  =  '%-12s'%('# Isomer ') + name + ' \n'
                    title +=  '%-20s'%( "# E<-E' (kcal/mol)" )
                    for i in range(well_sz): 
                        title +=  '%-15.5E'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                    f.write(title+'\n')
                    val_str    =  ''
                    for i in range(well_sz):
                        val_str +=  '%-20.5E'%( (ener_ref - i*ener_step)*cst.au2kcalmol )
                        for j in range(well_sz): 
                            val_str +=  '%-15.5E'%(kernel[i,j])
                        f.write( val_str+'\n' )
                        val_str  = ''
                    f.write('\n')



class serial_out_class(std_out_class):
    def __init__(self):
        std_out_class.__init__(self)

    def flushStdOut(self):
        self.flush_stdout()

    def error(self, _str):
        self.stdout( _str )
        self.flush_stdout()
        exit(1)



class mp_out_class(std_out_class):
    def __init__(self, _stdout_q , _mpv):
        std_out_class.__init__(self)
        self.__stdout_queue     =   _stdout_q  
        self.__mpv              =   _mpv

    def flushStdOut(self):
        rank    =   self.__mpv.rank
        barrier =   self.__mpv.barrier
        nproc   =   self.__mpv.nproc
        #push stdout of all processes into stdout_queue
        self.__stdout_queue.put ( (rank, self.stdout_buff) )
        #barrier
        barrier.wait()
        #print stdout_queue
        if rank == 0:
            content = [ self.__stdout_queue.get() for i in range(nproc) ]
            content.sort()
            content = [ x[1] for x in content ]
            self.flush_stdout(content)        

    def error(self, _str):
        self.stdout('Error occurs in the process of rank '+str(self.__mpv.rank) )
        self.stdout( _str )
        self.flushStdOut()
        exit(1)


class mpi_out_class(std_out_class):
    def __init__(self, comm ):
        std_out_class.__init__(self)
        self.__comm  =  comm


    def flushStdOut(self):
        #push stdout of all processes into stdout_queue
        content   =   self.__comm.gather(self.stdout_buff, root = 0 )
        #print stdout_queue
        if self.__comm.Get_rank() == 0:
            self.flush_stdout(content)       


    def error(self, _str):
        self.stdout('Error occurs in the process of rank '+str(self.__comm.Get_rank()) )
        self.stdout( _str )
        self.flushStdOut()
        exit(1)

