#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import numpy as np


class MPV ():
    def __init__(self, _rank, _nproc, _ME_q, _state_q, _barrier):
        self.__rank         =   _rank
        self.__nproc        =   _nproc
        self.__ME_queue     =   _ME_q
        self.__state_queue  =   _state_q
        self.__barrier      =   _barrier
        
    @property
    def rank       (self): return  self.__rank         
    @property
    def ME_queue   (self): return  self.__ME_queue     
    @property
    def state_queue(self): return  self.__state_queue  
    @property
    def barrier    (self): return  self.__barrier      
    @property
    def nproc      (self): return  self.__nproc 



def MPI_gather_ME_class( ME_stack, comm, gvr ):
    
    from mpi4py import MPI

    my_rank  =  comm.Get_rank()
    nproc    =  comm.Get_size()
    numME    =  len( ME_stack )
    Ntime    =  0
    y        =  None
    L_p      =  None
    U_p      =  None

    if my_rank == 0:
        import me_lib as me
        for source_rank in range( 1, nproc ):
            numME = comm.recv( source = source_rank, tag = 100* source_rank + 0 )
            for i in range(numME):
                #recv python type 
                T           =   comm.recv( source = source_rank, tag = 100* source_rank +  1 )
                p           =   comm.recv( source = source_rank, tag = 100* source_rank +  2 )
                ener_step   =   comm.recv( source = source_rank, tag = 100* source_rank +  3 )
                ener_ref    =   comm.recv( source = source_rank, tag = 100* source_rank +  4 )
                Z           =   comm.recv( source = source_rank, tag = 100* source_rank +  5 )
                size        =   comm.recv( source = source_rank, tag = 100* source_rank +  6 )
                well_sz     =   comm.recv( source = source_rank, tag = 100* source_rank +  7 )
                well_kernel =   comm.recv( source = source_rank, tag = 100* source_rank +  8 )
                Ncse        =   comm.recv( source = source_rank, tag = 100* source_rank +  9 )
                well_eig    =   comm.recv( source = source_rank, tag = 100* source_rank + 10 )
                merge_well  =   comm.recv( source = source_rank, tag = 100* source_rank + 11 )
                M           =   comm.recv( source = source_rank, tag = 100* source_rank + 12 )
                epsrs       =   comm.recv( source = source_rank, tag = 100* source_rank + 13 )
                kappa       =   comm.recv( source = source_rank, tag = 100* source_rank + 14 )
                epcs        =   comm.recv( source = source_rank, tag = 100* source_rank + 15 )
                kE_map      =   comm.recv( source = source_rank, tag = 100* source_rank + 16 )
                K           =   comm.recv( source = source_rank, tag = 100* source_rank + 17 )
                charac_t    =   comm.recv( source = source_rank, tag = 100* source_rank + 18 )
                #recv eigenvalue  1D array
                L           =   np.empty( size, dtype = 'float' )
                comm.Recv( [L, MPI.DOUBLE],             source = source_rank, tag = 100* source_rank + 19 )
                #recv eigenvector 2D matrix
                U           =   np.empty( [size, size], dtype = 'float', order='C' )
                comm.Recv( [U, MPI.DOUBLE],             source = source_rank, tag = 100* source_rank + 20 )
                #recv F 1D array
                F           =   np.empty( size, dtype = 'float' )
                comm.Recv( [F, MPI.DOUBLE],             source = source_rank, tag = 100* source_rank + 21 )
    
                if gvr.timeevolution_sw:
                    #set the size of concentration vector, y or y*
                    if gvr.pseubimname:
                        sz_p    = size + len(gvr.pseubimname)
                    else:
                        sz_p    = size
                    #recv y
                    Ntime   =  comm.recv(               source = source_rank, tag = 100* source_rank + 22 )
                    y       =  np.empty( [ Ntime, sz_p ], dtype = 'float', order='C' )
                    comm.Recv( [ y, MPI.DOUBLE ],       source = source_rank, tag = 100* source_rank + 23 )
                    #recv pseudo-first-order eigenpairs
                    if gvr.pseubimname:
                        L_p =  np.empty(  sz_p , dtype = 'float' )
                        U_p =  np.empty(  [sz_p, sz_p] , dtype = 'float', order='C' )
                        comm.Recv( [ L_p, MPI.DOUBLE ], source = source_rank, tag = 100* source_rank + 24 )
                        comm.Recv( [ U_p, MPI.DOUBLE ], source = source_rank, tag = 100* source_rank + 25 )
                #copy ME
                ME          =  me.MasterEquation()
                ME.set_ME(
                    _T          = T             ,
                    _p          = p             ,
                    _ener_step  = ener_step     ,
                    _ener_ref   = ener_ref      ,
                    _Z          = Z             ,
                    _size       = size          ,
                    _well_sz    = well_sz       ,
                    _well_kernel= well_kernel   ,
                    _Ncse       = Ncse          ,
                    _well_eig   = well_eig      ,
                    _merge_well = merge_well    ,
                    _M          = M             ,
                    _epsrs      = epsrs         ,
                    _kappa      = kappa         ,
                    _epcs       = epcs          ,
                    _kE_map     = kE_map        ,
                    _K          = K             ,
                    _charac_t   = charac_t      ,
                    _L          = L             ,
                    _U          = U             ,
                    _L_p        = L_p           ,
                    _U_p        = U_p           ,
                    _F          = F             ,
                    _y          = y             
                    )
                ME_stack.append( ME )
    else:
        comm.send( numME, dest = 0, tag = 100* my_rank + 0 )
        for i in range( numME ):
            #send python type 
            comm.send( ME_stack[i].T,                           dest = 0, tag = 100* my_rank +  1 )
            comm.send( ME_stack[i].p,                           dest = 0, tag = 100* my_rank +  2 )
            comm.send( ME_stack[i].ener_step,                   dest = 0, tag = 100* my_rank +  3 )
            comm.send( ME_stack[i].ener_ref,                    dest = 0, tag = 100* my_rank +  4 )
            comm.send( ME_stack[i].Z,                           dest = 0, tag = 100* my_rank +  5 )
            comm.send( ME_stack[i].size,                        dest = 0, tag = 100* my_rank +  6 )
            comm.send( ME_stack[i].well_sz,                     dest = 0, tag = 100* my_rank +  7 )
            comm.send( ME_stack[i].well_kernel,                 dest = 0, tag = 100* my_rank +  8 )
            comm.send( ME_stack[i].Ncse,                        dest = 0, tag = 100* my_rank +  9 )
            comm.send( ME_stack[i].well_eig,                    dest = 0, tag = 100* my_rank + 10 )
            comm.send( ME_stack[i].merge_well,                  dest = 0, tag = 100* my_rank + 11 )
            comm.send( ME_stack[i].M,                           dest = 0, tag = 100* my_rank + 12 )
            comm.send( ME_stack[i].epsrs,                       dest = 0, tag = 100* my_rank + 13 )
            comm.send( ME_stack[i].kappa,                       dest = 0, tag = 100* my_rank + 14 )
            comm.send( ME_stack[i].epcs,                        dest = 0, tag = 100* my_rank + 15 )
            comm.send( ME_stack[i].kE_map,                      dest = 0, tag = 100* my_rank + 16 )
            comm.send( ME_stack[i].K,                           dest = 0, tag = 100* my_rank + 17 )
            comm.send( ME_stack[i].charac_t,                    dest = 0, tag = 100* my_rank + 18 )
            #send eigenvalue  1D array
            comm.Send( [ME_stack[i].L, MPI.DOUBLE],             dest = 0, tag = 100* my_rank + 19 )
            #send eigenvector 2D matrix
            comm.Send( [ME_stack[i].U, MPI.DOUBLE],             dest = 0, tag = 100* my_rank + 20 )
            #send F 1D array
            comm.Send( [ME_stack[i].F, MPI.DOUBLE],             dest = 0, tag = 100* my_rank + 21 )
            
            if gvr.timeevolution_sw:
                #send y
                comm.send( ME_stack[i].y.shape[0],              dest = 0, tag = 100* my_rank + 22 )
                comm.Send( [ME_stack[i].y, MPI.DOUBLE],         dest = 0, tag = 100* my_rank + 23 )
                #send pseudo-first-order eigenpairs array or NoneType
                if gvr.pseubimname:
                    comm.Send( [ME_stack[i].L_p, MPI.DOUBLE],   dest = 0, tag = 100* my_rank + 24 )
                    comm.Send( [ME_stack[i].U_p, MPI.DOUBLE],   dest = 0, tag = 100* my_rank + 25 )

            
            
            



