#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import numpy        as np
import mep_lib
import species_lib  as sl
import global_lib   as gl
from   const_lib    import cst
import data_lib     as dl  
from scipy.special import gamma

class BarrierRxn_class(gl.CycleParam_class):

    def __init__( self, **kwargs):
        self.__kE           =   None
        self.__ts_sos       =   None
        self.__dk           =   None
        self.__rxn_type     =   None
        self.__pyr_file     =   None
        self.__reactant     =   None 
        self.__Rname        =   None 
        self.__ts           =   None 
        self.__TSname       =   None 
        self.__product      =   None 
        self.__Pname        =   None 
        self.__barrier      =   None 
        self.__rxn_ener     =   None 
        self.__mep          =   None 
        self.__variation    =   None 
        self.__tunnel       =   None 
        self.__PE           =   None 
        self.__gts_dict     =   {} 
        self.__avgdiameter  =   None
        self.__s_lower      =   None                
        self.__s_upper      =   None   
        self.__invlaplace   =   None
        self.__expression   =   None
        self.__A1           =   None
        self.__T1           =   None
        self.__n1           =   None
        self.__Ea1          =   None
        self.__A2           =   None
        self.__T2           =   None
        self.__n2           =   None
        self.__Ea2          =   None

        if '_reverse_rxn' in kwargs.keys():
            #set reaction fro reverse reaction ojb
            r_rxn               =   kwargs['_reverse_rxn']
            self.__reactant     =   r_rxn.product
            self.__Rname        =   r_rxn.Pname
            self.__ts           =   r_rxn.ts
            self.__TSname       =   r_rxn.TSname
            self.__product      =   r_rxn.reactant
            self.__Pname        =   r_rxn.Rname
            self.__mep          =   r_rxn.mep   
            self.__variation    =   r_rxn.variation
            self.__tunnel       =   r_rxn.tunnel
            self.__PE           =   r_rxn.PE
            self.__gts_dict     =   r_rxn.gts_dict
            self.__avgdiameter  =   r_rxn.avgdiameter
            self.__s_lower      =   r_rxn.s_lower
            self.__s_upper      =   r_rxn.s_upper
            self.__invlaplace   =   r_rxn.invlaplace
            self.__expression   =   r_rxn.expression
            self.__A1           =   r_rxn.A1        
            self.__T1           =   r_rxn.T1      
            self.__n1           =   r_rxn.n1         
            self.__Ea1          =   r_rxn.Ea1      
            self.__A2           =   r_rxn.A2        
            self.__T2           =   r_rxn.T2      
            self.__n2           =   r_rxn.n2         
            self.__Ea2          =   r_rxn.Ea2  

            if r_rxn.barrier  is not None:
                self.__barrier  =   r_rxn.barrier - r_rxn.rxn_ener
            if r_rxn.rxn_ener is not None:
                self.__rxn_ener =  -r_rxn.rxn_ener
        else:
            for key in kwargs.keys():
                if   key                ==      '_Rname':
                    self.__Rname        =       kwargs[key]
                elif key                ==      '_TSname':
                    self.__TSname       =       kwargs[key]
                elif key                ==      '_Pname':
                    self.__Pname        =       kwargs[key]
                elif key                ==      '_reactant':
                    self.__reactant     =       kwargs[key]
                elif key                ==      '_ts':
                    self.__ts           =       kwargs[key]
                elif key                ==      '_product':
                    self.__product      =       kwargs[key]
                elif key                ==      '_variation':
                    self.__variation    =       kwargs[key]
                elif key                ==      '_tunnel':
                    self.__tunnel       =       kwargs[key]
                elif key                ==      '_pyr_file':
                    self.__pyr_file     =       kwargs[key]
                elif key                ==      '_avgdiameter':
                    self.__avgdiameter  =       kwargs[key]
                elif key                ==      '_s_lower':
                    self.__s_lower      =       kwargs[key]
                elif key                ==      '_s_upper':
                    self.__s_upper      =       kwargs[key]
                elif key                ==      '_invlaplace':
                    self.__invlaplace   =       kwargs[key]
                elif key                ==      '_expression':
                    self.__expression   =       kwargs[key]
                elif key                ==      '_A1': 
                    self.__A1           =       kwargs[key]
                elif key                ==      '_T1':
                    self.__T1           =       kwargs[key]
                elif key                ==      '_n1':     
                    self.__n1           =      kwargs[key]
                elif key                ==      '_Ea1':           
                    self.__Ea1          =       kwargs[key]
                elif key                ==      '_A2': 
                    self.__A2           =       kwargs[key]
                elif key                ==      '_T2':
                    self.__T2           =       kwargs[key]
                elif key                ==      '_n2':     
                    self.__n2           =      kwargs[key]
                elif key                ==      '_Ea2':           
                    self.__Ea2          =       kwargs[key]        
                elif key                ==      'errorf':
                    errorf              =       kwargs[key]     
                elif key                ==      'gvr':
                    gvr                 =       kwargs[key]
                else:
                    print(' WRONG KEYWORDS FOR REACTION')
                    exit(1)

            self.__check( errorf )
            self.__setFromPolyrate( gvr, errorf )

    @property  
    def reactant (self):     return self.__reactant  
    @property  
    def ts       (self):     return self.__ts      
    @property  
    def product  (self):     return self.__product   
    @property  
    def Rname    (self):     return self.__Rname  
    @property  
    def TSname   (self):     return self.__TSname   
    @property  
    def Pname    (self):     return self.__Pname  
    @property  
    def rxn_type (self):     return self.__rxn_type  
    @property
    def barrier  (self):     return self.__barrier
    @property
    def rxn_ener (self):     return self.__rxn_ener
    @property  
    def kE       (self):     return self.__kE        
    @property  
    def PE       (self):     return self.__PE    
    @property
    def ts_sos   (self):     return self.__ts_sos  
    @property
    def dk       (self):     return self.__dk  
    @property
    def variation(self):     return self.__variation
    @property
    def tunnel   (self):     return self.__tunnel
    @property
    def mep      (self):     return self.__mep
    @property
    def gts_dict (self):     return self.__gts_dict
    @property
    def pyr_file (self):     return self.__pyr_file
    @property
    def avgdiameter(self):   return self.__avgdiameter
    @property
    def  s_lower   (self):   return self.__s_lower
    @property
    def  s_upper   (self):   return self.__s_upper
    @property 
    def  invlaplace(self):   return self.__invlaplace
    @property
    def  expression(self):   return self.__expression
    @property 
    def  A1        (self):   return self.__A1        
    @property 
    def  T1        (self):   return self.__T1       
    @property 
    def  n1        (self):   return self.__n1         
    @property 
    def  Ea1        (self):   return self.__Ea1        
    @property 
    def  A2        (self):   return self.__A2        
    @property 
    def  T2        (self):   return self.__T2       
    @property 
    def  n2        (self):   return self.__n2         
    @property 
    def  Ea2       (self):   return self.__Ea2        


    def __check(self, errorf):
        if   not self.__Rname:
            errorf( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+' reactant not set ...')
        elif  not self.__TSname:
            errorf( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+' transition stste not set ...')
        elif not self.__Pname:
            errorf( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+' product not set ...')
        if self.__Rname == self.__Pname:
            errorf( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+' reactant & product are the same species ...')


    def __setRxnType(self): 
        self.__rxn_type = ''
        if isinstance(self.__reactant, sl.Well_class):
            self.__rxn_type += "w"
        elif isinstance(self.__reactant, sl.Bimolecular_class):
            self.__rxn_type += 'b'

        self.__rxn_type += '2'

        if isinstance(self.__product, sl.Well_class):
            self.__rxn_type += "w"
        elif isinstance(self.__product, sl.Bimolecular_class):
            self.__rxn_type += 'b' 


    def __setFromPolyrate(self, gvr, errorf):
        if self.__pyr_file is not None:

            #read mep
            self.__mep          =   mep_lib.MEP(  self.__Rname, self.__TSname,  self.__Pname )
            self.__mep.readPYR ( self.__pyr_file, self.__s_lower, self.__s_upper, gvr, errorf )
            
            #read reactant
            if   self.__Rname in gvr.well_dict:
                self.__reactant = gvr.well_dict[self.__Rname]
            elif self.__Rname in gvr.bim_dict:
                self.__reactant = gvr.bim_dict[self.__Rname]
            else:
                self.__reactant =   self.__mep.reactant
                if isinstance(self.__reactant, sl.Well_class):
                    gvr.push_well_stack ( self.__reactant )
                    gvr.push_well_dict  ( self.__Rname, self.__reactant )
                else:
                    gvr.push_bim_stack ( self.__reactant )
                    gvr.push_bim_dict  ( self.__Rname, self.__reactant  )

            #read product
            if  self.__Pname in gvr.well_dict:
                self.__product = gvr.well_dict[self.__Pname]
            elif self.__Pname in gvr.bim_dict:
                self.__product = gvr.bim_dict[self.__Pname]
            else:
                self.__product  =   self.__mep.product
                if isinstance(self.__product, sl.Well_class):
                    gvr.push_well_stack( self.__product )
                    gvr.push_well_dict ( self.__Pname, self.__product )
                else:
                    gvr.push_bim_stack ( self.__product )
                    gvr.push_bim_dict  ( self.__Pname, self.__product )

            #read TS
            if self.__TSname in gvr.ts_dict:
                self.__ts       =   gvr.ts_dict[self.__TSname]
            else:
                self.__ts       =   self.__mep.ts
                gvr.push_ts_stack ( self.__ts )
                gvr.push_ts_dict  ( self.__TSname, self.__ts )

            #read barrier
            self.__barrier  =   self.__mep.barrier
            #reaf rxn_ener
            self.__rxn_ener =   self.__mep.rxn_ener
            #read gts_dict
            self.__gts_dict =   self.__mep.gts_dict
            #set rxn type
            self.__setRxnType()


    def __setPE( self, Emax, dE, errorf ):
        if self.__tunnel == 'lct':
            PE      =       self.__mep.lctPE
            PE[:,0] =       PE[:, 0] - PE[0, 0]
            VAG_I   =       int ( PE[ -1 , 0 ] // dE )
            PE_spr  =       dl.SplineFit( PE[:, 0] , PE[:, 1], 'lg')
            PE_tmp  =       np.ones( int( Emax / dE) )
            PE_tmp[ :VAG_I ]            =   PE_spr.val( np.arange( VAG_I )*dE )
            PE_tmp[ VAG_I : 2*VAG_I ]   =   1- PE_tmp[ :VAG_I ][::-1]
            self.__PE   =       PE_tmp
        elif self.__tunnel == 'omt':
            PE      =       self.__mep.omtPE
            PE[:,0] =       PE[:, 0] - PE[0, 0]
            VAG_I   =       int ( PE[ -1 , 0 ] // dE )
            PE_spr  =       dl.SplineFit( PE[:, 0] , PE[:, 1], 'lg')
            PE_tmp  =       np.ones( int( Emax / dE) )
            PE_tmp[ :VAG_I ]            =   PE_spr.val( np.arange( VAG_I )*dE )
            PE_tmp[ VAG_I : 2*VAG_I ]   =   1- PE_tmp[ :VAG_I ][::-1]
            self.__PE   =       PE_tmp
        elif self.__tunnel == 'sct':
            PE      =       self.__mep.sctPE
            PE[:,0] =       PE[:, 0] - PE[0, 0]
            VAG_I   =       int ( PE[ -1 , 0 ] // dE )
            PE_spr  =       dl.SplineFit( PE[:, 0] , PE[:, 1], 'lg')
            PE_tmp  =       np.ones( int( Emax / dE) )
            PE_tmp[ :VAG_I ]            =   PE_spr.val( np.arange( VAG_I )*dE )
            PE_tmp[ VAG_I : 2*VAG_I ]   =   1- PE_tmp[ :VAG_I ][::-1]
            self.__PE   =       PE_tmp
        elif self.__tunnel == 'zct':
            PE      =       self.__mep.zctPE
            PE[:,0] =       PE[:, 0] - PE[0, 0]
            VAG_I   =       int ( PE[ -1 , 0 ] // dE )
            PE_spr  =       dl.SplineFit( PE[:, 0] , PE[:, 1], 'lg')
            PE_tmp  =       np.ones( int( Emax / dE) )
            PE_tmp[ :VAG_I ]            =   PE_spr.val( np.arange( VAG_I )*dE )
            PE_tmp[ VAG_I : 2*VAG_I ]   =   1- PE_tmp[ :VAG_I ][::-1]
            self.__PE   =       PE_tmp
        elif self.__tunnel == 'Eckart':
            # reference : H. S. Johnston, J. Heicklen J. Phy. Chem. 1962 Vol. 66
            # but slightly change eq (15) because the relative energy E here is different. 
            V1      =       self.__ts.E0 - self.__reactant.E0
            V2      =       self.__ts.E0 - self.__product.E0
            nu      =       ( - self.__ts.image_freq )
            # relative to the maximum of E0 of reactant's and product's
            E       =       np.arange(int( Emax / dE)) * dE   
            alpha1  =       2*np.pi*V1 / nu
            alpha2  =       2*np.pi*V2 / nu
            kesi    =       E / V1
            a       =       2*np.sqrt( alpha1 * kesi ) / ( 1/np.sqrt(alpha1) + 1/np.sqrt(alpha2) )
            b       =       2*np.sqrt( ( alpha1 * kesi + np.fabs( alpha1- alpha2) ) ) / ( 1/np.sqrt(alpha1) + 1/np.sqrt(alpha2) )
            d       =       2*np.sqrt( np.fabs(alpha1*alpha2 - 4*np.pi*np.pi/16 ))
            self.__PE    =  1 - ( np.cosh(a-b) + np.cosh (d) )/( np.cosh(a+b) + np.cosh(d) )
            #check NaN (happens when a, b, d are too large)
            if not np.isfinite(self.__PE).all():
                errorf("    Wrong, NaN countered in Eckart tunneling calculation for the rxn "+self.Rname+"->"+self.Pname+", \
                usually your input is somehow wrong or do not use Eckart tuenneling here.")



    def __setTSSoS( self, dE ):
        ener_ref        =   self.ener_ref
        ener_step       =   self.ener_step
        E0_R            =   self.__reactant.E0
        mE0_t           =   max([ self.__reactant.E0_t, self.__product.E0_t ])
        mE0             =   max([ self.__reactant.E0, self.__product.E0 ])
        size            =   int( ( ener_ref - mE0_t ) // ener_step )
        E0_ts           =   self.__ts.E0
        E0_gts          =   0
        ts_dos_idx      =   lambda x: int( round( (ener_ref - E0_ts      - x*ener_step )/ dE) )
        gts_dos_idx     =   lambda x: int( round( (ener_ref - E0_gts     - x*ener_step )/ dE) )
        tnl_idx         =   lambda x: int( round( (ener_ref - mE0        - x*ener_step )/ dE) )
        R_dos_idx       =   lambda x: int( round( (ener_ref - E0_R       - x*ener_step )/ dE) )
        self.__ts_sos   =   np.zeros(size)
        #
        # based on the inverse-Laplace-transform to the HPL
        #
        if self.__invlaplace:
            if self.__expression == 'uniexp':
                if isinstance(self.__reactant, sl.Well_class):
                    expo1  = self.__n1
                    C1     = self.__A1 * (1/self.__T1)**self.__n1 * cst.h / gamma(expo1)
                    #set unimolecular dos
                    R_dos  = self.__reactant.dos
                else:
                    bim    = self.__reactant 
                    expo1  = self.__n1 + 3/2
                    C1     = (2*np.pi*bim.mu)**(3/2) * self.__A1 * (1/self.__T1)**self.__n1 / cst.h**2 / gamma(expo1)  
                    #set bimolecular joint dos
                    R_dos  = np.empty( int( round( ( ener_ref - bim.E0 ) / dE ) ) )
                    for i in range( R_dos.size ):
                        if   isinstance(bim.sp1, sl.FragmentAtom_class) and len(bim.sp1.ele_level) == 1:
                            R_dos[i]       =  bim.sp2.dos[ i ] * bim.sp1.ele_level[0][0] *dE
                        elif isinstance(bim.sp2, sl.FragmentAtom_class) and len(bim.sp2.ele_level) == 1:
                            R_dos[i]       =  bim.sp1.dos[ i ] * bim.sp2.ele_level[0][0] *dE
                        else:
                            R_dos[i]       =  np.sum( bim.sp1.dos[ : i+1 ] * bim.sp2.dos[ : i+1][::-1])*dE    

                for i in range(size):
                    j = R_dos_idx(i)
                    ener1 = np.maximum(  np.arange(j) * dE - self.__Ea1 , 0 )
                    ener1 = ener1**(expo1-1)
                    self.__ts_sos[ i ] = C1 * np.sum( R_dos[ :j ] * ener1[::-1] ) * dE

            elif self.__expression == 'biexp':
                if isinstance(self.__reactant, sl.Well_class):
                    expo1  = self.__n1
                    C1     = self.__A1 * (1/self.__T1)**self.__n1 * cst.h / gamma(expo1)
                    expo2  = self.__n2
                    C2     = self.__A2 * (1/self.__T2)**self.__n2 * cst.h / gamma(expo2)
                    #set unimolecular dos
                    R_dos  = self.__reactant.dos
                else:
                    expo1  = self.__n1 + 3/2
                    C1     = (2*np.pi*self.__reactant.mu)**(3/2) * self.__A1 * (1/self.__T1)**self.__n1 / cst.h**2 / gamma(expo1)         
                    expo2  = self.__n2 + 3/2
                    C2     = (2*np.pi*self.__reactant.mu)**(3/2) * self.__A2 * (1/self.__T2)**self.__n2 / cst.h**2 / gamma(expo2)    
                    #set bimolecular joint dos
                    R_dos  = np.empty( int( round( ( ener_ref - bim.E0 ) / dE ) ) )
                    for i in range( R_dos.size ):
                        if   isinstance(bim.sp1, sl.FragmentAtom_class) and len(bim.sp1.ele_level) == 1:
                            R_dos[i]       =  bim.sp2.dos[ i ] * bim.sp1.ele_level[0][0] *dE
                        elif isinstance(bim.sp2, sl.FragmentAtom_class) and len(bim.sp2.ele_level) == 1:
                            R_dos[i]       =  bim.sp1.dos[ i ] * bim.sp2.ele_level[0][0] *dE
                        else:
                            R_dos[i]       =  np.sum( bim.sp1.dos[ : i+1 ] * bim.sp2.dos[ : i+1][::-1])*dE       

                for i in range(size):
                    j = R_dos_idx(i)
                    ener1 = np.maximum( np.arange(j) * dE - self.__Ea1 , 0 )
                    ener1 = ener1**(expo1-1)
                    self.__ts_sos[ i ] = C1 * np.sum( R_dos[ :j ] * ener1[::-1] ) * dE  

                    ener2 = np.maximum( np.arange(j) * dE - self.__Ea2 , 0 )
                    ener2 = ener2**(expo2-1)
                    self.__ts_sos[ i ] += C2 * np.sum( R_dos[ :j ] * ener2[::-1] ) * dE
            return
        #
        # based on the analytical expression (RRKM theory framework)
        #
        if self.__variation == 'tst':    
            if self.__tunnel == None:
                for i in range(size):
                    j = ts_dos_idx(i) 
                    if j  >  0 : 
                        self.__ts_sos[ i ]  =  self.__ts.sos[ j ]
                    else:
                        break
            else:
                for i in range(size):
                    j =  tnl_idx ( i )
                    self.__ts_sos [ i ] = np.dot( self.__PE[ : j+1 ],  self.__ts.dos[ : j+1 ][::-1] ) * dE

        elif self.__variation == 'cvt':
            gts_G   =   []
            for  s  in self.__gts_dict:
                gts =   self.__gts_dict[s]
                gts.setThermal( dE )
                gts_G.append( - cst.kb * self.T * np.log( gts.Q_real ) )
            gts     =  list( self.__gts_dict.values() ) [np.argmax(gts_G)]
            E0_gts  =   gts.E0
            if self.__tunnel == None:
                for i in range(size):
                    j_ts  = ts_dos_idx(i)
                    j_gts = gts_dos_idx(i) 
                    if j_gts >  0 : 
                        if  j_ts > 0 : 
                            GAMMA =  self.__ts.sos [ j_ts ] / self.__ts.sos_ho[ j_ts ]
                        else:
                            GAMMA = 1
                        self.__ts_sos[ i ]  =  gts.sos[ j_gts ] * GAMMA
                    else:
                        break
            else:
                try:
                    for i in range(size):
                        j                   =   tnl_idx ( i )
                        GAMMA               =   np.dot( self.__PE[ : j+1 ],  self.__ts.dos[ : j+1 ][::-1] ) / np.dot( self.__PE[ : j+1 ], self.__ts.dos_ho[ : j+1 ][::-1] )
                        self.__ts_sos [ i ] =   np.dot( self.__PE[ : j+1 ],  gts.dos[ : j+1 ][::-1] )* dE * GAMMA
                except IndexError or ValueError:
                    print(' Error: index exceed, please increase the max energy for density of states and tansmission possibility. ')
        elif self.__variation == 'muvt':
            if self.__tunnel == None:
                VAG     =       np.amax([ x.E0 for x in list(self.__gts_dict.values())])
                for i in range(size):
                    if ( ener_ref - i*ener_step ) > VAG :
                        gts_sos     =   []
                        for  s  in self.__gts_dict:
                            E0_gts  =   self.__gts_dict[s].E0
                            j_ts    =   ts_dos_idx(i)
                            j_gts   =   gts_dos_idx(i) 
                            if  j_gts >  0 : 
                                if j_ts > 0 :
                                    GAMMA =  self.__ts.sos [ j_ts ] /self.__ts.sos_ho[ j_ts ]
                                else:
                                    GAMMA =  1 
                                gts_sos.append( self.__gts_dict[s].sos[ j_gts ] * GAMMA )
                        self.__ts_sos [ i ] = np.amin( gts_sos )
            else:
                try:
                    s_list = list(self.__gts_dict.keys())
                    for i in range(size):
                        gts_sos     =   []
                        for  s  in s_list:
                            E0_ts   =   self.__gts_dict[s].E0
                            j       =   tnl_idx(i) 
                            if j >  0 : 
                                GAMMA = np.dot( self.__PE[ : j+1 ],  self.__ts.dos[ : j+1 ][::-1] ) / np.dot( self.__PE[ : j+1 ], self.__ts.dos_ho[ : j+1 ][::-1] )
                                gts_sos.append(   np.dot( self.__PE[ : j+1 ],  self.__gts_dict[s].dos[ : j+1 ][::-1] )* dE * GAMMA )
                        s_idx = np.argmin( gts_sos )
                        if s_idx == 0  or s_idx == len(gts_sos) - 1:
                            self.__ts_sos [ i ] = np.amin( gts_sos )
                        else:
                            A = np.zeros((3,3)); b = np.zeros((3,1)) 
                            A[0,0]   = (s_list[s_idx-1])**2;  A[0,1] = s_list[s_idx-1];  A[0,2] = 1; b[0,0] = gts_sos[s_idx-1] 
                            A[1,0]   = (s_list[s_idx])**2;    A[1,1] = s_list[s_idx];    A[1,2] = 1; b[1,0] = gts_sos[s_idx] 
                            A[2,0]   = (s_list[s_idx+1])**2;  A[2,1] = s_list[s_idx+1];  A[2,2] = 1; b[2,0] = gts_sos[s_idx+1] 
                            abc      = np.linalg.solve(A, b)
                            s_muvt   = -abc[1] / abc[0] / 2
                            self.__ts_sos [ i ] = abc[0] * s_muvt**2 + abc[1] * s_muvt +abc[2]
                except IndexError or ValueError:
                    print(' Error: index exceed, please increase the max energy for density of states and tansmission possibility. ')
                    

    def setReaction( self, gvr, errorf ):   
        ener_step       =   self.ener_step
        ener_ref        =   self.ener_ref
        Emax            =   gvr.Emax
        dE              =   gvr.dE
        
        ph_name         =   [ x.name for x in gvr.phase_stack ]
        R_i             =   ph_name.index( self.__reactant.name )
        P_j             =   ph_name.index( self.__product.name  )
        
        #set PE, check if done in reverse rxn
        if isinstance(gvr.rxn_map[P_j][R_i].PE, np.ndarray):
            self.__PE     =  gvr.rxn_map[P_j][R_i].PE
        else:
            self.__setPE ( Emax, dE, errorf )

        #set ts_sos, check if done in reverse rxn
        if  isinstance(gvr.rxn_map[P_j][R_i].ts_sos, np.ndarray):
            self.__ts_sos = gvr.rxn_map[P_j][R_i].ts_sos
        else:
            self.__setTSSoS( dE )
        
        #set micro rate coefficient
        size            =   self.__ts_sos.size
        if self.__rxn_type == 'b2w':
            #set dk for bim to well rxn
            ts_sos                          =   np.zeros( self.__product.size )
            ts_sos[ :size ]                 =   self.__ts_sos  

            ener_factor                     =   np.exp( -( ener_ref - np.arange(self.__product.size)*ener_step ) / cst.kb /self.T )
            self.__dk                       =   ts_sos *  ener_factor * ener_step / self.__reactant.Q_real / cst.h 
        else:
            #set kE for well to well and well to bim rxn
            ts_sos                          =   np.zeros( self.__reactant.size )
            ts_sos[ :size ]                 =   self.__ts_sos  

            self.__kE                       =   ts_sos / self.__reactant.dos_t /cst.h


    def check(self, gvr, printf, errorf):
        #check species
        if self.__reactant          is None:
            if self.__Rname in gvr.phase_dict:
                self.__reactant = gvr.phase_dict[ self.__Rname ]
            else:
                errorf ( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+', reactant '+self.__Rname + ' has not been defined ! ' )
        if self.__ts                is None:
            if self.__TSname in gvr.ts_dict:
                self.__ts       = gvr.ts_dict   [ self.__TSname ]
            else:
                errorf ( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+', transition state '+self.__TSname + ' has not been defined ! ')
        if self.__product           is None:
            if self.__Pname in gvr.phase_dict:
                self.__product  = gvr.phase_dict[ self.__Pname ]
            else:
                errorf ( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+', product '+self.__Pname + ' has not been defined ! ')
        #check energies
        if self.__rxn_ener          is None: 
            if self.__product.E0    is None:
                errorf ( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+', product '+self.__product.name + ' energy has not been defined' )
            if self.__reactant.E0   is None:
                errorf ( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', reactant '+self.__reactant.name + ' energy has not been defined' )
            self.__rxn_ener =   self.__product.E0  - self.__reactant.E0
        if self.__barrier           is None:
            if self.__reactant.E0   is None:
                errorf ( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', reactant '+self.__reactant.name + ' energy has not been defined' )
            if self.__ts.E0         is None:
                errorf ( ' Wrong for reaction '+self.__Rname+'-'+self.__TSname+ '-'+self.__Pname+', transition state '+self.__ts.name + ' energy has not been defined' )
            self.__barrier  =   self.__ts.E0       - self.__reactant.E0 
        #check rxn_type
        if self.__rxn_type          is None:
            self.__setRxnType()
            if self.__rxn_type != 'b2w' and self.__rxn_type != 'w2b' and self.__rxn_type != 'w2w':
                errorf ( ' Wrong reaction type for reaction '+self.__reactant.name+'-'+self.__TSname+'-'+self.__product.name+', only bim-well or well-bim or well-well is acceptable.' )
        #check tunneling
        if self.__tunnel != 'lct' and self.__tunnel != 'omt' and self.__tunnel != 'sct' and self.__tunnel != 'zct' and self.__tunnel != 'Eckart' and self.__tunnel != None:
            errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', tunneling value \"'+str(self.__tunnel)+'\" is not acceptable.')
        if self.__tunnel == 'lct' and len(self.__mep.lctPE) == 0:
            printf('!! For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', lct tunneling is set,\n!! but found none transmission possibility, no tunneling considered here!')
            self.__tunnel = None
        if self.__tunnel == 'omt' and len(self.__mep.omtPE) == 0:
            printf('!! For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', mu-omt tunneling is set,\n!! but found none transmission possibility, no tunneling considered here!')
            self.__tunnel = None
        if self.__tunnel == 'sct' and len(self.__mep.sctPE) == 0:
            printf('!! For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', sct tunneling is set,\n!! but found none transmission possibility, no tunneling considered here!')
            self.__tunnel = None
        if self.__tunnel == 'zct' and len(self.__mep.zctPE) == 0:
            printf('!! For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', zct tunneling is set,\n!! but found none transmission possibility, no tunneling considered here!')
            self.__tunnel = None
        if self.__tunnel == 'Eckart' and self.__ts.image_freq is None:
            errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', imaginary frequency value \"'+str(self.__ts.image_freq)+'\" is not defined.')
        if self.__tunnel == 'Eckart' and self.__ts.image_freq > 0 :
            errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', imaginary frequency value \"'+str(self.__ts.image_freq)+'\" is not negtive.')
        #check variation
        if self.__variation != 'cvt' and self.__variation != 'muvt' and self.__variation != 'tst':
            errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', variation value \"'+str(self.__variation)+'\" is not acceptable.')
        if len(self.__gts_dict) == 0 and self.__variation == 'cvt':
            printf('!! For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', cvt variation is set,\n!! but found none general transition state, no variation considered here!')
            self.__variation = 'tst'
        if len(self.__gts_dict) == 0 and self.__variation == 'muvt':
            printf('!! For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', muvt variation is set,\n!! but found none general transition state, no variation considered here!')
            self.__variation = 'tst'
        #check inverse-Laplace-transform
        if self.__invlaplace:
            if  self.__expression is None:
                errorf(" Wrong for expression-type, should be defined as 'uniexp' or  'biexp'. ")
            elif self.__expression == 'uniexp':
                if self.__A1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but A1 is not defined.')
                if self.__A1 <= 0:
                    errorf(' Wrong for expression A1*(T/T1)^n1*exp(-Ea1/kB/T), A1 should be positive. ')
                if self.__T1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but T1 is not defined.')
                if self.__n1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but n1 is not defined.')
                if self.__Ea1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but Ea1 is not defined.')
                if self.__n1 <= 0:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', wrong for factor n1 in A1*(T/T1)^n1*exp(-Ea1/kB/T), n1 should be positive. ')
            elif self.__expression == 'biexp':
                if self.__A1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but A1 is not defined.')
                if self.__A1 <= 0:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', wrong for expression A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), A1 should be positive. ')
                if self.__T1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but T1 is not defined.')
                if self.__n1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but n1 is not defined.')
                if self.__Ea1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but Ea1 is not defined.')
                if self.__n1 <= 0:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', wrong for factor n1 in A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), n1 should be positive. ')
                if self.__A2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but A2 is not defined.')
                if self.__A2 <= 0:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', wrong for expression A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), A2 should be positive. ')
                if self.__T2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but T1 is not defined.')
                if self.__n2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but n1 is not defined.')
                if self.__Ea2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but Ea1 is not defined.')
                if self.__n2 <= 0:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', wrong for factor n2 in A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), n2 should be positive. ')
            else:
                errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', wrong for expression-type '+str(self.__expression)+", should exactly be 'uniexp' or 'biexp'. " )


    def printProperty( self, printf ):
        printf('[ REACTION INFO ]')
        printf('%-30s'%('  Reactant:')          + self.__Rname   )
        printf('%-30s'%('  TS:')                + self.__TSname  )
        printf('%-30s'%('  Product:')           + self.__Pname   )
        printf('%-30s'%('  Polyrate file:')     + str(self.__pyr_file) )
        printf('%-30s'%('  Variation:')         + str(self.__variation) )
        printf('%-30s'%('  Tunneling:')         + str(self.__tunnel) )
        printf('%-30s'%('  Lower limit of MEP:')+ str(self.__s_lower))
        printf('%-30s'%('  Upper limit of MEP:')+ str(self.__s_upper))
        if self.__invlaplace:
            if self.__expression == 'uniexp':
                printf('  The k(E) will be calculated from the inverse-Laplace-transform (ILT) to the high-pressure limit rate constant:')
                printf('      k^HPL(T) = A1 * (T/T1)^n1 * exp(-Ea1/kB/T)  (single-exponential function) ')
                printf('  where, ')
                printf('%-30s'%('   A1 [a.u.]:')      +  '%.5E'%(self.__A1))
                printf('%-30s'%('   T1[a.u.]:')   +  '%.5E'%(self.__T1))
                printf('%-30s'%('   n1:')       +  '%.5E'%(self.__n1))
                printf('%-30s'%('   Ea1[a.u.]:')      +  '%.5E'%(self.__Ea1))
            elif self.__expression == 'biexp':
                printf('  The k(E) will be calculated from the inverse-Laplace-transform (ILT) to the high-pressure limit rate constant:')
                printf('      k^HPL(T) = A1 * (T/T1)^n1 * exp(-Ea1/kB/T) + A2 * (T/T2)^n2 * exp(-Ea2/kB/T)  (biexponential function)')
                printf('  where, ')
                printf('%-30s'%('   A1 [a.u.]:')      +  '%.5E'%(self.__A1))
                printf('%-30s'%('   T1[a.u.]:')   +  '%.5E'%(self.__T1))
                printf('%-30s'%('   n1:')       +  '%.5E'%(self.__n1))
                printf('%-30s'%('   Ea1[a.u.]:')      +  '%.5E'%(self.__Ea1))
                printf('%-30s'%('   A2 [a.u.]:')      +  '%.5E'%(self.__A2))
                printf('%-30s'%('   T2[a.u.]:')   +  '%.5E'%(self.__T2))
                printf('%-30s'%('   n2:')       +  '%.5E'%(self.__n2))
                printf('%-30s'%('   Ea2[a.u.]:')      +  '%.5E'%(self.__Ea2))
        else:
            printf('  The k(E) will be calculated from the analytical expression. ')
        printf('[ END OF INFO ]\n')


    def clear(self):
        self.__kE           =   None
        self.__ts_sos       =   None
        self.__dk           =   None




class BarrierlessRxn_class(gl.CycleParam_class):
    def __init__( self, **kwargs):
        self.__kE           =   None
        self.__ts_sos       =   None
        self.__dk           =   None
        self.__rxn_type     =   None
        self.__reactant     =   None 
        self.__Rname        =   None 
        self.__product      =   None 
        self.__Pname        =   None 
        self.__rxn_ener     =   None 
        self.__avgdiameter  =   None 
        self.__rxn_type     =   None
        self.__invlaplace   =   None
        self.__expression   =   None
        self.__A1           =   None
        self.__T1           =   None
        self.__n1           =   None
        self.__Ea1          =   None
        self.__A2           =   None
        self.__T2           =   None
        self.__n2           =   None
        self.__Ea2          =   None

        if '_reverse_rxn' in kwargs.keys():
            #set reaction from reverse reaction ojb
            r_rxn               =   kwargs['_reverse_rxn']
            self.__reactant     =   r_rxn.product
            self.__Rname        =   r_rxn.Pname
            self.__product      =   r_rxn.reactant
            self.__Pname        =   r_rxn.Rname
            self.__avgdiameter  =   r_rxn.avgdiameter
            self.__invlaplace   =   r_rxn.invlaplace
            self.__expression   =   r_rxn.expression
            self.__A1           =   r_rxn.A1        
            self.__T1           =   r_rxn.T1       
            self.__n1           =   r_rxn.n1         
            self.__Ea1          =   r_rxn.Ea1
            self.__A2           =   r_rxn.A2        
            self.__T2           =   r_rxn.T2       
            self.__n2           =   r_rxn.n2         
            self.__Ea2          =   r_rxn.Ea2     

            if r_rxn.rxn_ener is not None:
                self.__rxn_ener =  -r_rxn.rxn_ener
        else:
            for key in kwargs.keys():
                if   key                ==      '_Rname':
                    self.__Rname        =       kwargs[key]
                elif key                ==      '_Pname':
                    self.__Pname        =       kwargs[key]
                elif key                ==      '_reactant':
                    self.__reactant     =       kwargs[key]
                elif key                ==      '_product':
                    self.__product      =       kwargs[key]
                elif key                ==      '_avgdiameter':
                    self.__avgdiameter  =       kwargs[key]
                elif key                ==      '_rxn_ener':
                    self.__rxn_ener     =       kwargs[key]
                elif key                ==      '_invlaplace':
                    self.__invlaplace   =       kwargs[key]
                elif key                ==      '_expression':
                    self.__expression   =       kwargs[key]
                elif key                ==      '_A1': 
                    self.__A1           =       kwargs[key]
                elif key                ==      '_T1':
                    self.__T1           =       kwargs[key]
                elif key                ==      '_n1':     
                    self.__n1           =       kwargs[key]
                elif key                ==      '_Ea1':         
                    self.__Ea1          =       kwargs[key]
                elif key                ==      '_A2': 
                    self.__A2           =       kwargs[key]
                elif key                ==      '_T2':
                    self.__T2           =       kwargs[key]
                elif key                ==      '_n2':     
                    self.__n2           =       kwargs[key]
                elif key                ==      '_Ea2':         
                    self.__Ea2          =       kwargs[key]   
                elif key                ==      'errorf':
                    errorf              =       kwargs[key]     
                elif key                ==      'gvr':
                    gvr                 =       kwargs[key]               
                else:
                    print(" Wrong key in reaction classs:", key)
                    exit(1)
            
            self.__check( errorf )

    @property  
    def reactant (self):     return self.__reactant     
    @property  
    def product  (self):     return self.__product   
    @property  
    def Rname    (self):     return self.__Rname  
    @property  
    def Pname    (self):     return self.__Pname  
    @property  
    def rxn_type (self):     return self.__rxn_type  
    @property
    def barrier  (self):     return 0
    @property
    def rxn_ener (self):     return self.__rxn_ener
    @property  
    def kE       (self):     return self.__kE         
    @property
    def dk       (self):     return self.__dk  
    @property
    def ts_sos   (self):     return self.__ts_sos
    @property
    def avgdiameter(self):   return self.__avgdiameter
    @property
    def variation(self):     return None
    @property
    def tunnel   (self):     return None
    @property
    def gts_dict (self):     return {}
    @property
    def pyr_file (self):     return None
    @property 
    def invlaplace(self):    return self.__invlaplace
    @property
    def expression(self):    return self.__expression 
    @property 
    def A1        (self):    return self.__A1        
    @property 
    def T1        (self):    return self.__T1      
    @property 
    def n1        (self):    return self.__n1         
    @property 
    def Ea1       (self):    return self.__Ea1     
    @property 
    def A2        (self):    return self.__A2        
    @property 
    def T2        (self):    return self.__T2      
    @property 
    def n2        (self):    return self.__n2         
    @property 
    def Ea2       (self):    return self.__Ea2    

    def __check (self, errorf):
        #check
        if   not self.__Rname:
            errorf( ' Wrong for reaction '+self.__Rname+ '-->'+self.__Pname+' reactant not set ...')
        elif not self.__Pname:
            errorf( ' Wrong for reaction '+self.__Rname+ '-->'+self.__Pname+' product not set ...')
        if self.__Rname == self.__Pname:
            errorf( ' Wrong for reaction '+self.__Rname+ '-->'+self.__Pname+' reactant & product are the same species ...')


    def __setRxnType(self): 
        self.__rxn_type = ''
        if isinstance(self.__reactant, sl.Well_class):
            self.__rxn_type += "w"
        elif isinstance(self.__reactant, sl.Bimolecular_class):
            self.__rxn_type += 'b'

        self.__rxn_type += '2'

        if isinstance(self.__product, sl.Well_class):
            self.__rxn_type += "w"
        elif isinstance(self.__product, sl.Bimolecular_class):
            self.__rxn_type += 'b' 


    def __setTSSoS(self, dE):
        ener_ref        =   self.ener_ref
        ener_step       =   self.ener_step
        mE0_t           =   max([ self.__reactant.E0_t, self.__product.E0_t ])
        E0_R            =   self.__reactant.E0
        if self.__rxn_type == 'b2w':
            Ebim        =   self.__reactant.E0
        else:
            Ebim        =   self.__product.E0
        tnl_idx         =   lambda x: int( round( (ener_ref - Ebim - x*ener_step )/ dE) )
        R_dos_idx       =   lambda x: int( round( (ener_ref - E0_R       - x*ener_step )/ dE) )
        size            =   int( ( ener_ref - mE0_t ) // ener_step )
        self.__ts_sos   =   np.zeros(size)

        #
        # based on the inverse-Laplace-transform to the HPL
        #
        if self.__invlaplace:
            if self.__expression == 'uniexp':
                if isinstance(self.__reactant, sl.Well_class):
                    expo1  = self.__n1
                    C1     = self.__A1 * (1/self.__T1)**self.__n1 * cst.h / gamma(expo1)
                    #set unimolecular dos
                    R_dos  = self.__reactant.dos
                else:
                    bim    = self.__reactant 
                    expo1  = self.__n1 + 3/2
                    C1     = (2*np.pi*bim.mu)**(3/2) * self.__A1 * (1/self.__T1)**self.__n1 / cst.h**2 / gamma(expo1)  
                    #set bimolecular joint dos
                    R_dos  = np.empty( int( round( ( ener_ref - bim.E0 ) / dE ) ) )
                    for i in range( R_dos.size ):
                        if   isinstance(bim.sp1, sl.FragmentAtom_class) and len(bim.sp1.ele_level) == 1:
                            R_dos[i]       =  bim.sp2.dos[ i ] * bim.sp1.ele_level[0][0] *dE
                        elif isinstance(bim.sp2, sl.FragmentAtom_class) and len(bim.sp2.ele_level) == 1:
                            R_dos[i]       =  bim.sp1.dos[ i ] * bim.sp2.ele_level[0][0] *dE
                        else:
                            R_dos[i]       =  np.sum( bim.sp1.dos[ : i+1 ] * bim.sp2.dos[ : i+1][::-1])*dE    

                for i in range(size):
                    j = R_dos_idx(i)
                    ener1 = np.maximum(  np.arange(j) * dE - self.__Ea1 , 0 )
                    ener1 = ener1**(expo1-1)
                    self.__ts_sos[ i ] = C1 * np.sum( R_dos[ :j ] * ener1[::-1] ) * dE

            elif self.__expression == 'biexp':
                if isinstance(self.__reactant, sl.Well_class):
                    expo1  = self.__n1
                    C1     = self.__A1 * (1/self.__T1)**self.__n1 * cst.h / gamma(expo1)
                    expo2  = self.__n2
                    C2     = self.__A2 * (1/self.__T2)**self.__n2 * cst.h / gamma(expo2)
                    #set unimolecular dos
                    R_dos  = self.__reactant.dos
                else:
                    expo1  = self.__n1 + 3/2
                    C1     = (2*np.pi*self.__reactant.mu)**(3/2) * self.__A1 * (1/self.__T1)**self.__n1 / cst.h**2 / gamma(expo1)         
                    expo2  = self.__n2 + 3/2
                    C2     = (2*np.pi*self.__reactant.mu)**(3/2) * self.__A2 * (1/self.__T2)**self.__n2 / cst.h**2 / gamma(expo2)    
                    #set bimolecular joint dos
                    R_dos  = np.empty( int( round( ( ener_ref - bim.E0 ) / dE ) ) )
                    for i in range( R_dos.size ):
                        if   isinstance(bim.sp1, sl.FragmentAtom_class) and len(bim.sp1.ele_level) == 1:
                            R_dos[i]       =  bim.sp2.dos[ i ] * bim.sp1.ele_level[0][0]
                        elif isinstance(bim.sp2, sl.FragmentAtom_class) and len(bim.sp2.ele_level) == 1:
                            R_dos[i]       =  bim.sp1.dos[ i ] * bim.sp2.ele_level[0][0]
                        else:
                            R_dos[i]       =  np.sum( bim.sp1.dos[ : i+1 ] * bim.sp2.dos[ : i+1][::-1])*dE       

                for i in range(size):
                    j = R_dos_idx(i)
                    ener1 = np.maximum( np.arange(j) * dE - self.__Ea1 , 0 )
                    ener1 = ener1**(expo1-1)
                    self.__ts_sos[ i ] = C1 * np.sum( R_dos[ :j ] * ener1[::-1] ) * dE  

                    ener2 = np.maximum( np.arange(j) * dE - self.__Ea2 , 0 )
                    ener2 = ener2**(expo2-1)
                    self.__ts_sos[ i ] += C2 * np.sum( R_dos[ :j ] * ener2[::-1] ) * dE
            return
        #
        # based on the analytical expression (hard-sphere collision model)
        #
        if isinstance(  self.__reactant, sl.Bimolecular_class ):
            bim             = self.__reactant  
        else:
            bim             = self.__product
        #set joint dos
        joint_dos           =   np.empty( int( round( ( ener_ref - bim.E0 ) / dE ) ) )
        for i in range( joint_dos.size ):
            if   isinstance(bim.sp1, sl.FragmentAtom_class) and len(bim.sp1.ele_level) == 1:
                joint_dos[i]       =  bim.sp2.dos[ i ] * bim.sp1.ele_level[0][0]
            elif isinstance(bim.sp2, sl.FragmentAtom_class) and len(bim.sp2.ele_level) == 1:
                joint_dos[i]       =  bim.sp1.dos[ i ] * bim.sp2.ele_level[0][0]
            else:
                joint_dos[i]       =  np.sum( bim.sp1.dos[ : i+1 ] * bim.sp2.dos[ : i+1][::-1])*dE 
        #constant
        C                          =  8 * np.pi**2 * self.__avgdiameter **2 * bim.mu / cst.h**2
        #sos of ts            
        for i in range(size):
            j                      =  tnl_idx ( i )
            self.__ts_sos[i]       =  C * np.sum( joint_dos [ : j] * np.arange(j-1, -1, -1) * dE *dE )
    

    def setReaction(self,gvr, errorf):   
        ener_step       =   self.ener_step
        ener_ref        =   self.ener_ref
        dE              =   gvr.dE
        
        ph_name         =   [ x.name for x in gvr.phase_stack ]
        R_i             =   ph_name.index( self.__reactant.name )
        P_j             =   ph_name.index( self.__product.name  )

        #set ts_sos, check if done in reverse rxn
        if  isinstance(gvr.rxn_map[P_j][R_i].ts_sos, np.ndarray):
            self.__ts_sos = gvr.rxn_map[P_j][R_i].ts_sos
        else:
            self.__setTSSoS(dE)
        
        #set micro rate coefficient
        size            =   self.__ts_sos.size
        if self.__rxn_type == 'b2w':
            #set dk for bim to well rxn
            ts_sos                          =   np.zeros( self.__product.size )
            ts_sos[ :size ]                 =   self.__ts_sos  
            ener_factor                     =   np.exp( -( ener_ref - np.arange(self.__product.size)*ener_step ) / cst.kb /self.T )
            self.__dk                       =   ts_sos *  ener_factor * ener_step / self.__reactant.Q_real / cst.h
        else:
            #set kE for well to well and well to bim rxn
            ts_sos                          =   np.zeros( self.__reactant.size )
            ts_sos[ :size ]                 =   self.__ts_sos  
            self.__kE                       =   ts_sos / self.__reactant.dos_t /cst.h


    def check(self, gvr, printf, errorf):  
        #check species
        if self.__reactant              is None:
            if self.__Rname in gvr.phase_dict:
                self.__reactant = gvr.phase_dict[ self.__Rname ]
            else:
                errorf ( ' Wrong for reaction '+self.__Rname+ '-->'+self.__Pname+', reactant '+self.__Rname + ' has not been defined ! ' )
        if self.__product               is None:
            if self.__Pname in gvr.phase_dict:
                self.__product  = gvr.phase_dict[ self.__Pname ]
            else:
                errorf ( ' Wrong for reaction '+self.__Rname+ '-->'+self.__Pname+', product '+self.__Pname + ' has not been defined ! ')
        #check energies
        if self.__rxn_ener              is None: 
            if self.__product.E0        is None:
                errorf ( ' Wrong for reaction '+self.__Rname+ '-->'+self.__Pname+', product '+self.__product.name + ' energy has not been defined' )
            
            if self.__reactant.E0       is None:
                errorf ( ' Wrong for reaction '+self.__Rname+ '-->'+self.__Pname+', reactant '+self.__reactant.name + ' energy has not been defined' )

            self.__rxn_ener =   self.__product.E0  - self.__reactant.E0
        #check rxn_type
        if self.__rxn_type              is None:
            self.__setRxnType()
            if self.__rxn_type != 'b2w' and self.__rxn_type != 'w2b':
                errorf ( ' Wrong reaction type for barrierless reaction '+self.__reactant.name+'-'+self.__product.name+', only bim-well or well-bim is acceptable.' )
        #check inverse-Laplace-transform
        if self.__invlaplace:
            if  self.__expression is None:
                errorf(" Wrong for expression-type, should be defined as 'uniexp' or  'biexp'. ")
            elif self.__expression == 'uniexp':
                if self.__A1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but A1 is not defined.')
                if self.__A1 <= 0:
                    errorf(' Wrong for expression A1*(T/T1)^n1*exp(-Ea1/kB/T), A1 should be positive. ')
                if self.__T1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but T1 is not defined.')
                if self.__n1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but n1 is not defined.')
                if self.__Ea1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but Ea1 is not defined.')
                if self.__n1 <= 0:
                    errorf(' Wrong for factor n1 in A1*(T/T1)^n1*exp(-Ea1/kB/T), n1 should be positive. ')
            elif self.__expression == 'biexp':
                if self.__A1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but A1 is not defined.')
                if self.__A1 <= 0:
                    errorf(' Wrong for expression A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), A1 should be positive. ')
                if self.__T1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but T1 is not defined.')
                if self.__n1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but n1 is not defined.')
                if self.__Ea1 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but Ea1 is not defined.')
                if self.__n1 <= 0:
                    errorf(' Wrong for factor n1 in A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), n1 should be positive. ')
                if self.__A2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but A2 is not defined.')
                if self.__A2 <= 0:
                    errorf(' Wrong for expression A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), A2 should be positive. ')
                if self.__T2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but T1 is not defined.')
                if self.__n2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but n1 is not defined.')
                if self.__Ea2 is None:
                    errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', the inverse-Laplace-transform is used but Ea1 is not defined.')
                if self.__n2 <= 0:
                    errorf(' Wrong for factor n2 in A1*(T/T1)^n1*exp(-Ea1/kB/T)+A2*(T/T2)^n2*exp(-Ea2/kB/T), n2 should be positive. ')
            else:
                errorf(' For reaction '+self.__Rname+'-'+self.__TSname+'-'+self.__Pname+', wrong for expression-type '+str(self.__expression)+", should exactly be 'uniexp' or 'biexp'. " )


    def printProperty( self, printf ):
        printf('[ REACTION INFO ]')
        printf('%-30s'%('  Barrierless Reaction ... ') )
        printf('%-30s'%('  Reactant:')          + self.__Rname   )
        printf('%-30s'%('  Product:')           + self.__Pname   )
        if self.__invlaplace:
            if self.__expression == 'uniexp':
                printf('  The k(E) will be calculated from the inverse-Laplace-transform (ILT) to the high-pressure limit rate constant:')
                printf('      k^HPL(T) = A1 * (T/T1)^n1 * exp(-Ea1/kB/T)  (single-exponential function) ')
                printf('  where, ')
                printf('%-30s'%('   A1 [a.u.]:')      +  '%.5E'%(self.__A1))
                printf('%-30s'%('   T1[a.u.]:')   +  '%.5E'%(self.__T1))
                printf('%-30s'%('   n1:')       +  '%.5E'%(self.__n1))
                printf('%-30s'%('   Ea1[a.u.]:')      +  '%.5E'%(self.__Ea1))
            elif self.__expression == 'biexp':
                printf('  The k(E) will be calculated from the inverse-Laplace-transform (ILT) to the high-pressure limit rate constant:')
                printf('      k^HPL(T) = A1 * (T/T1)^n1 * exp(-Ea1/kB/T) + A2 * (T/T2)^n2 * exp(-Ea2/kB/T)  (biexponential function)')
                printf('  where, ')
                printf('%-30s'%('   A1 [a.u.]:')      +  '%.5E'%(self.__A1))
                printf('%-30s'%('   T1[a.u.]:')   +  '%.5E'%(self.__T1))
                printf('%-30s'%('   n1:')       +  '%.5E'%(self.__n1))
                printf('%-30s'%('   Ea1[a.u.]:')      +  '%.5E'%(self.__Ea1))
                printf('%-30s'%('   A2 [a.u.]:')      +  '%.5E'%(self.__A2))
                printf('%-30s'%('   T2[a.u.]:')   +  '%.5E'%(self.__T2))
                printf('%-30s'%('   n2:')       +  '%.5E'%(self.__n2))
                printf('%-30s'%('   Ea2[a.u.]:')      +  '%.5E'%(self.__Ea2))
        else:
            printf('%-30s'%('  Reaction energy:')   + '%.5E'%(self.__rxn_ener) )
            printf('%-30s'%('  Averaged diameter:') + '%.5E'%(self.__avgdiameter) )
            printf('  The k(E) will be calculated from the inverse-Laplace-transform (ILT) to the high-pressure limit (hard-sphere) rate constant:')
            printf('      k^HPL(beta) = pi*(d_gamma+d_M)^2 * sqrt(8/pi/m_reduced/beta) / 4 ')
            
        printf('[ END OF INFO ]\n')


    def clear(self):
        self.__kE           =   None
        self.__ts_sos       =   None
        self.__dk           =   None


