#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#

import numpy            as np
import global_lib       as gl
import molecule_lib     as ml
import const_lib        as cl
import output_lib       as io
import collision_lib    as coll


class Well_class(ml.Molecule_class):
    def __init__(
        self,
        _name,
        _symbol,
        _geom,
        _freq,
        _freqscale,
        _ener_type,
        _energy,
        _ele_level,              
        _rotsigma,
        _opticalnum,
        _epsilon,
        _diameter,
        _bmax,
        _collid,
        _mst_file,
        _g09_file,
        _source,
        errorf   =   print   
):
        #check
        if _name is None:
            errorf ( ' Wrong name for Well. You forgot to define it. ' )
        if _freqscale <= 0:
            errorf ( ' Wrong scaling factor for frequencies, it should be positive. ' )
        if _g09_file is None :
            if _symbol is None:
                errorf ('  Wrong geoamtries for ' + _name + '. You forgot to define it.')
            if  _freq is None:
                errorf ('  Wrong frequencies for ' + _name + '. You forgot to define it.')
            for x in _freq:
                if x <= 0:
                    errorf ('  Wrong frequencies for ' + _name + '. All should be positive.')
            if len(_symbol) != len(_geom):
                errorf ('  Number of symbols' + _name + ' does not equal to that of geometry coordinates ' )
            if _rotsigma <= 0:
                errorf ('  Rotsigma of '+ _name +'  should be positive.' )
            if _opticalnum != 1 and _opticalnum != 2:
                errorf ('  Optical-isomer number of '+ _name +'  should 1 or 2.') 
            if _collid is None:
                errorf ('  Wrong collision for ' + _name + '. You should define it firstly in Parameter.')
            if len(_symbol) <= 1:
                errorf (' Wrong for number of atoms ' + _name+'. It should be greater than 1.')

        self.__kernel             =   []
        self.__E0_t               =   0
        self.__dos_t              =   []
        self.__size               =   1
        self.__bolztmann          =   []
        self.__sqrt_bolztmann     =   []
        self.__Q_static           =   1
        self.__Q_real             =   1
        self.__epsilon            =   _epsilon
        self.__diameter           =   _diameter
        self.__bmax               =   _bmax

        #init father class
        ml.Molecule_class.__init__(
            self,
            _name,
            _geom,
            _symbol,
            _freq,
            _freqscale,
            _ener_type,
            _energy,
            _ele_level,
            _rotsigma,
            _opticalnum,
            _mst_file,
            _g09_file, 
            None,
            _source,
            errorf )

        # check frequency
        if self.islinear == True   and len(_freq) != 3 * len(_geom) - 5:
            errorf ('  Degrees of freedom of isomer '+ _name +'  is not correct. It should be 3N-5 for linear molecule.' )
        if self.islinear == False  and len(_freq) != 3 * len(_geom) - 6:
            errorf ('  Degrees of freedom of isomer '+ _name +'  is not correct. It should be 3N-6 for non-linear molecule.' )


    @property
    def kernel         (self): return self.__kernel
    @property
    def E0_t           (self): return  self.__E0_t           
    @property
    def dos_t          (self): return  self.__dos_t         
    @property
    def size           (self): return  self.__size          
    @property
    def bolztmann      (self): return  self.__bolztmann     
    @property
    def sqrt_bolztmann (self): return  self.__sqrt_bolztmann
    @property
    def Q_static       (self): return  self.__Q_static      
    @property
    def Q_real         (self): return  self.__Q_real             
    @property
    def epsilon        (self): return  self.__epsilon
    @property
    def diameter       (self): return  self.__diameter
    @property
    def bmax           (self): return  self.__bmax
    @property
    def collid         (self): return  self.__collid

    def set_epsilon( self, x ):     self.__epsilon  = x
    def set_diameter( self, x ):    self.__diameter = x
    def set_bmax( self, x ):        self.__bmax     = x

    def set_collid( self, _collid, errorf ):
        if _collid.name == 'LJ':
            #set LJ collid
            if self.__diameter is None:
                if _collid.diameterA is None:
                    errorf(' Wrong diameter for ' + self.name + '. You forgot to define it.')
                elif _collid.diameterA <= 0 :
                    errorf(' Wrong diameter for ' + self.name + '. Should be positive.')
                else:
                    self.__diameter = _collid.diameterA
            elif self.__diameter <= 0:
                errorf(' Wrong diameter for ' + self.name + '. Should be positive.')
            if  self.__epsilon is None:
                if _collid.epsilonA is None:
                    errorf(' Wrong epsilon for ' + self.name + '. You forgot to define it.')
                elif _collid.epsilonA <= 0 :
                    errorf(' Wrong diameter for ' + self.name + '. Should be positive.')
                else:
                    self.__epsilon = _collid.epsilonA
            elif self.__epsilon <= 0:
                errorf(' Wrong epsilon for ' + self.name + '. Should be positive.')

            self.__collid = coll.LJ_collision_class( _collid.epsilonM, self.__epsilon, _collid.massM, self.mass, _collid.diameterM, self.__diameter )
        else:
            #set HS collid
            if  self.__bmax is None:
                if _collid.bmax is None:
                    errorf(' Wrong bmax for ' + self.__name + ' in HSCollison. You forgot to define it.')
                else:
                    self.__bmax = _collid.bmax
            elif self.__bmax <= 0 :
                errorf(' BMAX for '+ self.name +' in HSCollision should be positive.')
            
            self.__collid = coll.HS_collision_class( self.__bmax, _collid.massM, self.mass )



    def setKernel( self, gvr ):
        ener_step               =   self.ener_step
        ener_ref                =   self.ener_ref
        T                       =   self.T
        dE                      =   gvr.dE
        size                    =   int( ( ener_ref - self.E0) // ener_step )
        kernelT                 =   np.empty( [size,size] )
        Normal                  =   np.zeros(size)
        I0                      =   0
        E0                      =   self.E0
        E0_t                    =   self.E0 + (ener_ref - self.E0) % ener_step
        dos                     =   self.dos
        idx                     =   lambda x: int( ( x-self.E0)/dE + 0.5 )
        
        # consider the temperature-depedency of <DeltaE_d>  
        T_factor = ( T / gvr.Tc )**(gvr.exponent)

        #primary relaxation matrix
        #Notice: the kernelT[0]=1*ener_step
        # energy-dependent <DeltaE_d> = (T/Tc) *[slope* E + EdConst], where E is the energy relative to the E0
        slope   = gvr.slope
        EdConst = gvr.EdConst
        for j in range(size):             
            for i in range(j+1):       
                #  E'>=E
                e_j          =   (j + 1) *   ener_step   +   E0_t
                e_i          =   (i + 1) *   ener_step   +   E0_t
                kernelT[j,i]    =   np.exp( -( e_j - e_i ) / T_factor / ( slope * ( e_j - E0 ) + EdConst ) )
            for i in range(j+1,size):
                e_j          =   (j + 1) *   ener_step   +   E0_t
                e_i          =   (i + 1) *   ener_step   +   E0_t
                kernelT[j,i] =   np.exp(-(e_i - e_j) / T_factor / ( slope * ( e_i - E0 ) + EdConst ) ) * dos[idx(e_i)] / dos[idx(e_j)] * np.exp(-(e_i - e_j) / cl.cst.kb /T)


        #normalization constant
        itmpt=0
        while True:
            I0 = itmpt
            for i in reversed(range( I0, size )):
                sum1 = np.sum(kernelT[ i, I0:i+1 ])
                sum2 = np.sum(kernelT[ i, i+1:size ] * Normal[ i+1:size ])
                if sum2 >   1.0:
                    itmpt  =  i + 1 
                    break  
                else:
                    Normal[i]    =   (1.0 - sum2) / sum1
            if I0 == itmpt:
                break

        #resize the P(E|E') matrix
        if  I0 > 0 :
            kernelT_t       =   kernelT[ I0:size, I0:size ]
        else:
            kernelT_t       =   kernelT

        #resize normal constant
        Normal_t =   Normal [ I0 :size ]
        size     =   size    -   I0
        
        #finnal relaxtion kernel
        for i in range(size):
            kernelT_t[i,  0:i+1 ] = kernelT_t[i,  0:i+1 ]    * Normal_t[i]
            kernelT_t[i,i+1:size] = kernelT_t[i,i+1:size]    * Normal_t[i+1:size]

        self.__E0_t    =    E0_t + ( I0 +1 )*ener_step
        self.__kernel  =    kernelT_t.T[::-1,::-1]      
        self.__size    =    size


    def setThermal(self, dE):
        ener_ref                =   self.ener_ref
        ener_step               =   self.ener_step
        T                       =   self.T
        idx                     =   lambda x: int( round( ( ener_ref - self.E0 - x*ener_step ) / dE ) )
        
        # trancated dos
        self.__dos_t            =   np.empty(self.size)
        for i in range(self.size):
            self.__dos_t [i]    =   self.dos[idx(i)]

        self.setPartitionfunc( dE ) 
        self.__Q_real           =   self.Qevr * np.exp(-self.E0 / cl.cst.kb / T)
        self.__bolztmann        =   self.dos_t * np.exp( -( ener_ref- np.arange(self.size)*ener_step) / cl.cst.kb / T)
        self.__sqrt_bolztmann   =   np.sqrt( self.bolztmann )
        self.__Q_static         =   np.sum( self.bolztmann )


class TS_class(ml.Molecule_class):

    def __init__(
        self,
        _name,
        _symbol,
        _geom,
        _freq,
        _freqscale,
        _ener_type,
        _energy,
        _ele_level ,
        _rotsigma  ,
        _opticalnum,
        _mst_file  ,
        _g09_file  ,
        _image_freq,
        _source,
        errorf = print
):
        #check
        if _name is None:
            errorf ( ' Wrong name for TS. You forgot to define it. ' )
        if _freqscale <= 0:
            errorf ( ' Wrong scaling factor for frequencies, it should be positive. ' )
        if _g09_file is None:
            if _symbol is None:
                errorf ('  Wrong geoamtries for' + _name + '. You forgot to define it.')
            if _freq is None:
                errorf ('  Wrong frequencies for ' + _name + '. You forgot to define it.')
            for x in _freq:
                if x <= 0:
                    errorf ('  Wrong frequencies for ' + _name + '. All should be positive.')
            if len(_symbol) != len(_geom):
                errorf ( ' Number of symbols' + _name + ' does not equal to that of geometry coordinates ' )
            if _rotsigma <= 0:
                errorf ( ' Rotsigma of '+ _name +'  should be positive.' )
            if _opticalnum != 1 and _opticalnum != 2:
                errorf ('  Optical-isomer number of '+ _name +'  should 1 or 2.') 
            if _image_freq is not None and _image_freq > 0:
                errorf ( ' Image frequency of  '+ _name +'  should be negtive.' )
            if len(_symbol) <= 1:
                errorf (' Wrong for number of atoms ' + _name+'. It should be greater than 1.')

        #init father class
        ml.Molecule_class.__init__(
            self,
            _name,
            _geom,
            _symbol,
            _freq,
            _freqscale,
            _ener_type,
            _energy,
            _ele_level,
            _rotsigma,
            _opticalnum,
            _mst_file,
            _g09_file,
            _image_freq,
            _source,
            errorf
)
        # check frequencies
        if self.islinear == True  and len(self.freq) != 3 * self.Natom - 6:
            errorf ( ' Degrees of freedom of transition state '+ _name +'  is not correct. It should be 3N-6 for the linear.')
        if self.islinear == False and len(self.freq) != 3 * self.Natom - 7:
            errorf ( ' Degrees of freedom of transition state '+ _name +'  is not correct. It should be 3N-7 for the non-linear.')


    @property
    def Q_real(self):  return self.__Q_real


    def setThermal(self, dE):
        self.setPartitionfunc( dE )
        self.__Q_real         =   self.Qevr * np.exp(-self.E0 / cl.cst.kb / self.T)



class FragmentAtom_class(ml.Atom_class):
    def __init__(
        self,
        _name,
        _energy,
        _ele_level,
        _symbol,
        _source,
        errorf    = print
):
        #check
        if _name is None:
            errorf ( ' Wrong name for atomic fragment in bimolecular. You forgot to define it. ' )
        if _symbol is None:
            errorf ( '  Wrong symbol for ' + _name + '. You forgot to define it. ' )
        if _energy is None:
            errorf ( '  Wrong energy for ' + _name + '. You forgot to define it. ') 

        ml.Atom_class.__init__( self, _name, _energy, _ele_level, _symbol, _source )
        self.__Q_real = 1
    
    def setThermal(self, dE):
        self.setPartitionfunc()
        self.__Q_real         =   self.Qevr *self.Qt

    @property
    def Q_real  (self):     return  self.__Q_real



class FragmentMol_class(ml.Molecule_class):
    def __init__(
        self,
        _name,
        _symbol,
        _geom,
        _freq,
        _freqscale,
        _ener_type,
        _energy,
        _ele_level,
        _rotsigma,
        _opticalnum,
        _mst_file,
        _g09_file,
        _source,
        errorf = print
):
        #check
        if _name is None:
            errorf ( ' Wrong name for molecular fragment in bimolecular. You forgot to define it. ' )
        if _freqscale <= 0:
            errorf ( ' Wrong scaling factor for frequencies, it should be positive. ' )
        if _g09_file is None:
            if _symbol is None:
                errorf ('  Wrong geoamtries for ' + _name + '. You forgot to define it.')
            if _freq is None:
                errorf ('  Wrong frequencies for ' + _name + '. You forgot to define it.')
            for x in _freq:
                if x <= 0:
                    errorf ('  Wrong frequencies for ' + _name + '. All should be positive.')
            if len(_symbol) != len(_geom):
                errorf ('  Number of symbols' + _name + ' does not equal to that of geometry coordinates ' )
            if _rotsigma <= 0:
                errorf ('  Rotsigma of '+ _name +'  should be positive.' )
            if _opticalnum != 1 and _opticalnum != 2:
                errorf ('  Optical-isomer number of '+ _name +'  should 1 or 2.') 
            if len(_symbol) <= 1:
                errorf (' Wrong for number of atoms ' + _name+'. It should be greater than 1.')

        #init
        ml.Molecule_class.__init__(
            self,
            _name,
            _geom,
            _symbol,
            _freq,
            _freqscale,
            _ener_type,
            _energy,
            _ele_level,
            _rotsigma,
            _opticalnum,
            _mst_file,
            _g09_file,
            None,
            _source,
            errorf
)
        # check frequency
        if self.islinear == True   and len(_freq) != 3 * len(_geom) - 5:
            errorf ('  Degrees of freedom of bimolecular-fragment '+ _name +'  is not correct. It should be 3N-5 for linear molecule.' )
        if self.islinear == False  and len(_freq) != 3 * len(_geom) - 6:
            errorf ('  Degrees of freedom of bimolecular-fragment '+ _name +'  is not correct. It should be 3N-6 for non-linear molecule.' )

        self.__Q_real = 1

    def setThermal(self, dE):
        self.setPartitionfunc( dE )
        self.__Q_real         =   self.Qevr *self.Qt

    @property
    def Q_real  (self):     return  self.__Q_real
    


class Bimolecular_class:
    def __init__(
        self,
        _name,
        _ener_type,
        _energy, 
        _sp1,
        _sp2,
        errorf = print
):
        if _name is None:
            errorf ( ' Wrong name for bimolecular. You forgot to define it. ' )
        
        self.__name           =   _name
        self.__sp1            =   _sp1
        self.__sp2            =   _sp2
        self.__E0             =   self.__setE0(_ener_type,_energy)
        self.__E0_t           =   self.__E0
        self.__mu             =   self.__setReduceMass()
        self.__Q_real         =   1
        self.__opticalnum     =   self.__sp1.opticalnum * self.__sp2.opticalnum
    
    @property
    def name        (self):     return  self.__name
    @property
    def sp1         (self):     return  self.__sp1
    @property
    def sp2         (self):     return  self.__sp2
    @property
    def E0          (self):     return  self.__E0
    @property
    def E0_t        (self):     return  self.__E0_t
    @property
    def mu          (self):     return  self.__mu
    @property
    def Q_real      (self):     return  self.__Q_real
    @property
    def opticalnum  (self):     return self.__opticalnum


    def __setE0(self,_ener_type,_energy):
        if _ener_type   ==  "E0K":
            E0 =   _energy
        elif _ener_type ==  "Eele":
            E0 =   _energy + self.__sp1.ZPVE +self.__sp2.ZPVE
        else:
            E0 =   self.__sp1.E0 +self.__sp2.E0
        return E0


    def __setReduceMass(self):        
        return self.__sp1.mass*self.__sp2.mass  /   (self.__sp1.mass + self.__sp2.mass)

    def set_E0(self,x):
        self.__E0 = x

    def setThermal(self, dE):
        self.__sp1.setThermal( dE )
        self.__sp2.setThermal( dE )
        self.__E0_t     = self.__E0
        T               = self.__sp1.T
        Qt_rel          = np.power( 2 * np.pi * self.__mu * cl.cst.kb * T / cl.cst.h**2, 3/2 )
        ener_factor     = np.exp(-self.E0/cl.cst.kb/T)
        self.__Q_real   = Qt_rel * self.__sp1.Qevr * self.__sp2.Qevr *ener_factor 

