#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import sys
import global_lib   as gl
import parallel_lib as pl
import output_lib   as io
import tumme_readin
import tumme_pre
import tumme_solver
import multiprocessing as mp



def serial_main(f, gvr, out):
    # Print header
    out.stdout(" \nSerial TUMME Starts ... \n")
    
    # Read Everything Input
    tumme_readin.std_readInput( f, gvr, out )

    # Pre-Setting for States and Reaction-Map
    tumme_pre.serial_preSetting( gvr, out )

    # Solve
    tumme_solver.serial_solver(gvr, out)

    # Flush stdout
    out.flushStdOut()

    # Print rate
    out.printRate(gvr)

    # Print eval
    out.printEval(gvr)

    # Print evec
    out.printEvec(gvr)

    # Print kE
    out.printkE(gvr)

    # Print time evolution
    out.printTimeEvolution(gvr)

    # Print partition function
    out.printPartitionfunction(gvr)

    # Print relaxation possibility kernel
    out.printRelaxKernel(gvr)


def mp_main(f, rank, nproc, ME_q, state_q, stdout_q, barrier, gvr):
    # Set mpv
    mpv     =   pl.MPV( rank, nproc, ME_q, state_q, barrier)
    
    # Set out
    out     =   io.mp_out_class( stdout_q, mpv )

    # Print header
    if rank == 0:
        out.stdout(" \nMP TUMME Starts ... \n")

    # Set stdout file
    out.set_stdout_file ( f.split('.')[0] )
    out.set_rate_file( f.split('.')[0] )

    # Read Everything Input
    tumme_readin.std_readInput( f, gvr, out )

    # Pre-Setting for States and Reaction-Map
    tumme_pre.mp_preSetting( mpv, gvr, out )

    # Solve
    tumme_solver.mp_solver( mpv, gvr, out)
 
    # Flush stdout
    out.flushStdOut()

    if mpv.rank == 0:

        # Print rate
        out.printRate(gvr)

        # Print eval
        out.printEval(gvr)

        # Print evec
        out.printEvec(gvr)

        # Print kE
        out.printkE(gvr)

        # Print time evolution
        out.printTimeEvolution(gvr)

        # Print partition function
        out.printPartitionfunction(gvr)

        # Print relaxation possibility kernel
        out.printRelaxKernel(gvr)


def mpi_main(f, comm, gvr, out):
    # Print header
    if comm.Get_rank() == 0:
        out.stdout(" \nMPI TUMME Starts ... \n")

    # Read Everything Input
    tumme_readin.std_readInput( f, gvr, out )

    # Pre-Setting for States and Reaction-Map
    tumme_pre.mpi_preSetting( comm, gvr, out )

    # Solve
    tumme_solver.mpi_solver(comm, gvr, out)

    # Flush stdout
    out.flushStdOut()
    
    if comm.Get_rank() == 0:

        # Print rate
        out.printRate(gvr)

        # Print eval
        out.printEval(gvr)

        # Print evec
        out.printEvec(gvr)

        # Print kE
        out.printkE(gvr)

        # Print time evolution
        out.printTimeEvolution(gvr)

        # Print partition function
        out.printPartitionfunction(gvr)

        # Print relaxation possibility kernel
        out.printRelaxKernel(gvr)



#main process entrance
if __name__ == '__main__':
    # analyse command
    try:
        std_in_file =  sys.argv[1]
    except:
        print(' Wrong input form\nIt should be like \n\tpython /home/tumme/src/tumme_main.py test.in\n\tor\n\tmpirun -np 4 python /home/tumme/src/tumme/src/tumme_main.py test.in')

    # set global variavls and root path for tumme
    if sys.argv[0].find('/src/tumme_main.py') == -1:
        root_path   =  '../'
    else:
        root_path   =  (sys.argv[0])[ : sys.argv[0].find('/src/tumme_main.py') ]

    # set global variable
    gvr             =   gl.GlobalVar_class()
    gvr             =   gl.GlobalVar_class()
    parallel, nproc =   tumme_readin.readParallel( std_in_file, print )
    Ncpu            =   mp.cpu_count()
    gvr.set_root_path( root_path)

    if parallel == 'mp':
        #set nproc
        nproc   =   min( mp.cpu_count(), nproc )  
        #set omp_num_thread
        if gvr.ompnumthread is None:
            gvr.set_ompnumthread( max( int( Ncpu / nproc ), 1 ) )
        #set queue
        ME_q    =   mp.Queue()
        state_q =   mp.Queue()
        stdout_q=   mp.Queue()
        #set barrier
        barrier =   mp.Barrier(nproc)
        #set processes
        process =   []
        for rank in range(1, nproc):
            process.append( mp.Process(target = mp_main, args = (std_in_file, rank, nproc, ME_q, state_q, stdout_q, barrier, gvr)))

        #children execute serial main()
        for rank in range(1, nproc):
            process[rank-1].start()
    
        #father execute serial main()
        mp_main ( std_in_file, 0, nproc, ME_q, state_q, stdout_q, barrier, gvr)

    elif parallel == 'mpi':
        try:
            from mpi4py import MPI 
        except ImportError:
            print( " No mpi4py module ... " )
            exit(1)
        #set comm
        comm = MPI.COMM_WORLD
        #set out
        out  = io.mpi_out_class( comm )
        out.set_stdout_file (  std_in_file.split('.')[0] )
        out.set_rate_file( std_in_file.split('.')[0] )
        #set omp_num_thread
        if gvr.ompnumthread is None:
            gvr.set_ompnumthread( max( int( Ncpu / comm.Get_size() ), 1 ) )
        #execute main()
        mpi_main(std_in_file, comm, gvr, out )
        
    elif parallel == 'serial':
      #set omp_num_thread
        if gvr.ompnumthread is None:
            gvr.set_ompnumthread( int( Ncpu ) )
            
        #set out
        out     =   io.serial_out_class()
        out.set_stdout_file (  std_in_file.split('.')[0] )
        out.set_rate_file( std_in_file.split('.')[0] )

        #execute serial main()        
        serial_main(std_in_file,gvr,out)
        
    else:
        print("Wrong of parallel ... ")


    print("Hello world")


