#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#


import  numpy       as      np 
from    math        import  ceil
import species_lib  as      sl
import reaction_lib as      rl

def std_preSetting(gvr, out):
    
    # set phase stack and dictionary 
    gvr.set_phase_stack ( gvr.well_stack + gvr.bim_stack )
    gvr.set_phase_dict  ( { **gvr.well_dict ,  **gvr.bim_dict}  )

    # check global paramter
    gvr.check( out.error )
    
    # check if mass conserved
    MASS = round(gvr.well_stack[0].mass,2)
    for well in gvr.well_stack:
        if MASS != round(well.mass,2):
          out.error(gvr.well_stack[0].name+'\'s mass is not equal to '+well.name+'\'s mass')
    for ts in gvr.ts_stack:
       if MASS != round(ts.mass,2):
          out.error(gvr.well_stack[0].name+'\'s mass is not equal to '+ts.name+'\'s mass')
    for bim in gvr.bim_stack:
       if MASS != round(bim.sp1.mass+bim.sp2.mass,2):
          out.error(gvr.well_stack[0].name+'\'s mass is not equal to '+bim.name+'\'s mass')

    # check rxn 
    for rxn in gvr.rxn_stack:
        rxn.check( gvr, out.stdout, out.error )

    # check and set collision
    for well in gvr.well_stack:
        well.set_collid( gvr.collid, out.error )

    # Reset Phase's E0 by Solve AE0 * xE0 = bE0
    Nph     =   len( gvr.phase_stack )
    ph_name =   [ x.name for x in gvr.phase_stack ]
    ts_name =   [ x.name for x in gvr.ts_stack    ]
    N       =   Nph + len(gvr.ts_stack)
    AE0     =   np.zeros( (N, N) )
    bE0     =   np.zeros( N )
    #gnd    
    gnd_j                   =   ph_name.index(gvr.gndSpecies)
    AE0 [0, gnd_j]          =    1
    #assign AE0 and bE0
    i = 1
    for rxn in gvr.rxn_stack:
        if isinstance(rxn, rl.BarrierlessRxn_class):
            R_j                 =   ph_name.index ( rxn.Rname )
            P_j                 =   ph_name.index ( rxn.Pname ) 
            AE0[i, R_j ]        =   -1
            AE0[i, P_j]         =    1
            if np.linalg.matrix_rank( AE0 ) == i + 1:
                bE0[i]          =   rxn.rxn_ener
                i               =   i   +   1
                if i == N:
                    break
            else:
                AE0[i, R_j ]    =   0
                AE0[i, P_j]     =   0
        else:
            R_j                 =   ph_name.index ( rxn.Rname )
            TS_j                =   Nph +   ts_name.index ( rxn.TSname)   
            AE0[i, R_j ]        =   -1
            AE0[i, TS_j]        =    1
            if np.linalg.matrix_rank( AE0 ) == i + 1:
                bE0[i]              =   rxn.barrier
                i                   =   i   +   1
                if i == N:
                    break
            else:
                AE0[i, R_j ]        =   0
                AE0[i, TS_j]        =   0
            
    #solve xE0
    xE0 = np.linalg.solve(AE0, bE0)
    #reassign E0
    for i in range(N):
        if i < Nph:
            gvr.phase_stack[i].set_E0( xE0[i] )
        else:
            gvr.ts_stack[i-Nph].set_E0( xE0[i] )

    # set rxn map
    gvr.set_rxn_map( [ [ None for i in range(Nph)] for i in range(Nph)] )
    for rxn in gvr.rxn_stack:
        R_i = ph_name.index ( rxn.reactant.name )
        P_j = ph_name.index ( rxn.product.name )
        gvr.rxn_map [R_i][P_j] = rxn
           
    # set gts E0
    for i in range( len(gvr.rxn_map) ):
        for j in range( i, len(gvr.rxn_map) ):
            rxn = gvr.rxn_map[i][j]
            if rxn:
                if rxn.gts_dict:
                    s0_E0 = rxn.gts_dict[0].E0
                    ts_E0 = rxn.ts.E0
                    for s in rxn.gts_dict:
                        rxn.gts_dict[s].set_E0( rxn.gts_dict[s].E0 - s0_E0 + ts_E0  )

    # set density of state
    spec_stack = gvr.well_stack + gvr.ts_stack

    if gvr.Emax <= gvr.EEOT * max(gvr.temperature)  + max([ sp.E0 for sp in gvr.ts_stack+gvr.phase_stack ]) :
        out.error(" The max-energy for density of state is less than the max-energy of master equation. Increase EMAX or reduce EEOT. ")

    for i in range( len(gvr.rxn_map) ):
        for j in range( i, len(gvr.rxn_map) ):
            if gvr.rxn_map[i][j]:
                if gvr.rxn_map[i][j].variation:
                    spec_stack  =   spec_stack + [ x for x in gvr.rxn_map[i][j].gts_dict.values()]
    for bim in gvr.bim_stack:
        spec_stack.append(bim.sp1)
        spec_stack.append(bim.sp2)
    for spec in spec_stack: 
        spec.setState(gvr.dE, gvr.Emax, out.error)

    #set HO state for TS
    for ts in gvr.ts_stack:
        ts.setState( gvr.dE, gvr.Emax, out.error, True )


def serial_preSetting(gvr, out):
    # std_presetting
    std_preSetting( gvr, out )
    
    #print global paramters
    out.printGlobalParameter( gvr )
    #print species
    out.printIsomer( gvr.well_stack )
    out.printTransState( gvr.ts_stack )
    out.printBimolecule( gvr.bim_stack )
    #print out rxn stack    
    out.printRxnStack( gvr.rxn_stack )
    #print out rxn map
    out.printRxnMap( gvr )


def mp_preSetting( mpv, gvr, out):
    # std_presetting
    std_preSetting( gvr, out )

    #print out rxn map
    if mpv.rank == 0:
        #print global paramters
        out.printGlobalParameter( gvr )
        #print species
        out.printIsomer( gvr.well_stack )
        out.printTransState( gvr.ts_stack )
        out.printBimolecule( gvr.bim_stack )
        #print out rxn stack    
        out.printRxnStack( gvr.rxn_stack )
        #print out rxn map
        out.printRxnMap( gvr )


def mpi_preSetting(comm, gvr, out):
    # std_presetting
    std_preSetting( gvr, out )

    #set rank
    rank        =   comm.Get_rank()

    #print out rxn map
    if rank == 0:
        #print global paramters
        out.printGlobalParameter( gvr )
        #print species
        out.printIsomer( gvr.well_stack )
        out.printTransState( gvr.ts_stack )
        out.printBimolecule( gvr.bim_stack )
        #print out rxn stack    
        out.printRxnStack( gvr.rxn_stack )
        #print out rxn map
        out.printRxnMap( gvr )
    

