#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#


import  numpy            as         np
from    const_lib        import     cst
import  global_lib       as         gl
import  collision_lib      
import  species_lib       
import  reaction_lib         
 


def __check(_str, _name, out):
    if _str.strip() == '':
        pass
    else:
        string = ' Wrong input \" '+_str+' \"\n'
    
        if   _name == 'input':
            string +=  gl.Key_class.printInputKeywords()
        elif _name == 'parameter':
            string +=  gl.Key_class.printParameterKeywords()
        elif _name == 'ljcollision':
            string +=  gl.Key_class.printLJCollisionKeywords()
        elif _name == 'hscollision':
            string +=  gl.Key_class.printHSCollisionKeywords()   
        elif _name == 'edown':
            string +=  gl.Key_class.printEdownKeywords()
        elif _name == 'energy':
            string +=  gl.Key_class.printEnergyKeywords()
        elif _name == 'pseubimolecular':
            string +=  gl.Key_class.printPseuBimolecularKeywords()
        elif _name == 'print':
            string +=  gl.Key_class.printPrintsKeywords()
        elif _name == 'reaction':
            string +=  gl.Key_class.printReactionKeywords()
        elif _name == 'barrierrxn':
            string +=  gl.Key_class.printBarrierRxnKeywords()
        elif _name == 'barrierlessrxn':
            string +=  gl.Key_class.printBarrierlessRxnKeywords()
        elif _name == 'invlaplace':
            string +=  gl.Key_class.printInvLaplaceKeywords()
        elif _name == 'species':
            string +=  gl.Key_class.printSpeciesKeywords()
        elif _name == 'unimolecule':
            string +=  gl.Key_class.printUnimolecularKeywords()
        elif _name == 'bimolecule':
            string +=  gl.Key_class.printBimolecularKeywords()
        elif _name == 'atom':
            string +=  gl.Key_class.printAtomKeywords()
        elif _name == 'timeevolution':
            string +=  gl.Key_class.printTimeEvolutionKeywords()
        elif _name == 'eigenvalue':
            string +=  gl.Key_class.printEigenvalueKeywords()
        elif _name == 'eigenvector':
            string +=  gl.Key_class.printEigenvectorKeywords()
        elif _name == 'relaxkernel':
            string +=  gl.Key_class.printRelaxKernelKeywords()

        out.error( string )


def __getUnit( out, _str):
    tmp = _str.strip().split('[')
    if len(tmp) == 1:
        unit = 1
    else:
        string = _str.strip().split('[')[1].split(']')[0]
        if string in cst.unit_dict:
            unit = cst.unit_dict[ string ]
        else:
            tmp  = ' Wrong unit input \"'+ string +'\"\n\nThe available unit are :\n'
            for i in cst.unit_dict:
                tmp += i + '\n'
            out.error( tmp )

    return unit


def  __getVar( out,  _str,  _type = cst.FLOAT ):
    #check
    vallist = _str.split('#')[0].strip().split()
    if len(vallist) < 2:
        out.error(_str+' has no value ...')
    elif len(vallist) >2:
        out.error(_str+' has more than one values ...')
    val = vallist[1]

    #read values
    if   _type == cst.STR:
        re = val
    elif _type == cst.FLOAT:
        try:
            re = float(val) * __getUnit( out, _str)
        except:
            out.error( ' Wrong float input \"' + val +' \"' )
    elif _type == cst.BOOL:
        if     val.lower() == "on"  or val.lower() == "yes":
            re = True
        elif   val.lower() == "off" or val.lower() == "no":
            re = False
        else:
            out.error(' Wrong bool input \"' + val +' \" \n\nThe available are:\non\nyes\noff\nno\nCaps is acceptable')
    return re


def  __getArr( out,  f, _unit,  _type ,  _end  ):
    
    if _type == cst.FLOAT :
        list_tmp = []
        line_str = f.readline()
        line     = line_str.split('#')[0]
        while _end not in line.lower() and line_str != '':
            cdata  = line.strip().split()
            if  cdata:
                for x in cdata:
                    try:
                        list_tmp.append( float(x) * _unit )
                    except:
                        out.error( ' Wrong float input \"' + x +' \"' )
            line_str    = f.readline()
            line        = line_str.split('#')[0]
        if list_tmp:
            return np.array(list_tmp)
        else:
            out.error(' Wrong '+(_end - 'end_')+' input.\n\nNo data ...' )
    elif _type == cst.STR :
        list_tmp    = []
        line_str    = f.readline()
        line        = line_str.split('#')[0]
        while _end not in line.lower() and line_str != '':
            cdata   = line.strip().split()
            if  cdata:
                for x in cdata:
                    list_tmp.append(x)
            line_str = f.readline()
            line = line_str.split('#')[0]
        if list_tmp:
            return list_tmp
        else:
            out.error(' Wrong '+(_end - 'end_')+' input.\n\nNo data ...' )
    elif _type == cst.ELEV : 
        list_tmp    = []
        line_str    = f.readline()
        line        = line_str.split('#')[0]
        no_ground   = True

        while _end not in line.lower() and line_str != '':
            cdata   = line.strip().split()
            if cdata:
                if len(cdata) == 2:   
                    try:
                        tmp = [ float(cdata[0]), float(cdata[1]) * _unit ]
                    except:
                        out.error ( ' Wrong float input \"' + line +' \"' )
                    if tmp[0] <= 0:
                        out.error ( " Wrong degeneracy of electronic level (it should be positive) : " + line )
                    if tmp[1] < 0:
                        out.error ( " Wrong energy of electronic level ( it should be positive, energies relative to 0 K ): " + line )
                    list_tmp.append( tmp )
                    if tmp[1] == 0:
                        no_ground = False
                else:
                    out.error( ' Wrong eletronic level input \"' + line +' \"\n\nEach line should only contain two floats, the first for energy, the second for degeneracy.' )
            line_str = f.readline()
            line     = line_str.split('#')[0]
            if no_ground:
                out.error( ' Wrong eletronic level input, there should be a line denoting the degeneracy for ground electronic state.' )
        if list_tmp:
            return list_tmp
        else:
            out.error(' Wrong '+(_end - 'end_')+' input.\n\nNo data ...' )  

    elif _type == cst.GEOM:
        symbol_tmp  = []
        geom_tmp    = []
        line_str    = f.readline()
        line        = line_str.split('#')[0]
        while _end not in line.lower() and line_str != '':
            geom_str    =   line.strip().split()
            if geom_str:
                if len(geom_str) == 4:
                    if geom_str[0] in cst.mass_dict:
                        symbol_tmp.append(geom_str[0])
                    else:
                        string = ' Wrong for symbol \"'+geom_str[0]+'\"\n\n The available atomic symbol are:\n'
                        for i in cst.mass_dict:
                            string += i + ' \n'
                        out.error( string )
                    try:
                        geom_tmp.append([ float(x)*_unit for x in geom_str[1:]])
                    except:
                        out.error ( ' Wrong float input \"' + line +' \"' )
                else:
                    out.error( ' Wrong Geom input \"' + line +' \"\n\nEach line should only contain four terms, the first for symbol, others for coordinates.' )
            line_str    = f.readline()
            line        = line_str.split('#')[0]
        if symbol_tmp:
            return np.array(symbol_tmp), np.array(geom_tmp)
        else:
            out.error(' Wrong '+(_end - 'end_')+' input.\n\nNo data ...' )     
        


def  __readParameter(f, key, default, gvr, out):
    #read Parameters
    EndKey      =   lambda x: "end_" + x
    line_str    = f.readline()
    line        = line_str.split('#')[0]
    while EndKey(key.param) not in line.lower() and line_str != '':
        #read LJ collision
        if key.LJcollid in line.lower():
            epM_tmp   = None; epA_tmp   = None
            masM_tmp  = None; masA_tmp  = None
            diamM_tmp = None; diamA_tmp = None
            
            line_str  = f.readline()
            line      = line_str.split('#')[0]
            while EndKey(key.LJcollid) not in line.lower() and line_str != '':
                if  key.epsilonM        in line.lower():
                    epM_tmp     =  __getVar( out, line)
                elif key.epsilonA       in line.lower():
                    epA_tmp     =  __getVar( out, line)
                elif  key.diameterM     in line.lower():
                    diamM_tmp   =  __getVar( out, line)
                elif key.diameterA      in line.lower():
                    diamA_tmp   =  __getVar( out, line)
                elif key.massM          in line.lower():
                    masM_tmp    =  __getVar( out, line)
                elif key.massA          in line.lower():
                    masA_tmp    =  __getVar( out, line)
                else:
                    __check(line, 'ljcollision', out)
                line_str = f.readline()
                line     = line_str.split('#')[0]
            gvr.set_collid ( collision_lib.LJ_collision_class(epM_tmp, epA_tmp, masM_tmp, masA_tmp, diamM_tmp, diamA_tmp) )
        #read HS collision
        elif key.HScollid in line.lower():
            bmax_tmp = None
            masM_tmp = None
            masA_tmp = None
            line_str = f.readline()
            line     = line_str.split('#')[0]
            while EndKey(key.HScollid) not in line.lower() and line_str != '':
                if  key.bmax      in line.lower():
                    bmax_tmp     =  __getVar( out, line)
                elif key.massM    in line.lower():
                    masM_tmp    =  __getVar( out, line)
                elif key.massA    in line.lower():
                    masA_tmp    =  __getVar( out, line)
                else:
                    __check(line, 'hscollision', out)
                line_str = f.readline()
                line     = line_str.split('#')[0]
            gvr.set_collid ( collision_lib.HS_collision_class(bmax_tmp,masM_tmp,masA_tmp) )
        #read Edown
        elif key.Edown in line.lower():
            line_str = f.readline()
            line     = line_str.split('#')[0]
            while EndKey(key.Edown) not in line.lower() and line_str != '':
                if   key.slope in line.lower():
                    gvr.set_slope ( __getVar( out, line) )
                elif key.EdConst in line.lower():
                    gvr.set_EdConst ( __getVar( out, line) ) 
                elif key.Tc in line.lower():
                    gvr.set_Tc ( __getVar( out, line) )     
                elif key.exponent in line.lower():
                    gvr.set_exponent ( __getVar( out, line) )       
                else:
                    __check(line, 'edown' , out)
                line_str = f.readline()
                line     = line_str.split('#')[0]
        #read temperture
        elif key.temperature in line.lower():
            gvr.set_temperature (  __getArr( out, f, __getUnit( out, line), cst.FLOAT, EndKey(key.temperature) ) )
        #read pressure
        elif key.pressure in line.lower():
            gvr.set_pressure  (  __getArr( out, f, __getUnit( out, line), cst.FLOAT, EndKey(key.pressure) ) )
        #read energy
        elif key.energy in line.lower():
            line_str = f.readline()
            line     = line_str.split('#')[0]
            while EndKey(key.energy) not in line.lower() and line_str != '':
                if   key.ESOT in line.lower():
                    gvr.set_ESOT ( __getVar( out, line) )
                elif key.EEOT in line.lower():
                    gvr.set_EEOT ( __getVar( out, line) )
                elif   key.dE in line.lower():
                    gvr.set_dE   ( __getVar( out, line) )
                elif key.Emax in line.lower():
                    gvr.set_Emax ( __getVar( out, line) )
                else:
                    __check(line, 'energy' , out)
                line_str = f.readline()
                line     = line_str.split('#')[0]
        #read pseudo-first-order bimolecular pairs
        elif key.pseubimolecular in line.lower():
            line_str = f.readline()
            line     = line_str.split('#')[0]
            while EndKey(key.pseubimolecular) not in line.lower() and line_str != '':
                if key.pseubimname  in line.lower():
                    gvr.set_pseubimname( __getArr( out, f, None, cst.STR, EndKey(key.pseubimname)) )
                elif key.excessconc in line.lower():
                    gvr.set_excessconc( __getArr( out, f, __getUnit( out, line), cst.FLOAT, EndKey(key.excessconc) ) )
                else:
                    __check(line, 'pseubimolecular', out)
                line_str = f.readline()
                line     = line_str.split('#')[0]
        #read ground energy species
        elif key.gndSpecies      in line.lower():
            gvr.set_gndSpecies       (  __getVar( out, line, cst.STR) )
        #read merge threshold 
        elif key.merge_threshold  in line.lower():
            gvr.set_merge_threshold   ( __getVar( out, line) )
        #read precision 
        elif key.precision       in line.lower():
            gvr.set_precision        ( __getVar( out, line, cst.STR) )
        elif key.ompnumthread    in line.lower():
            gvr.set_ompnumthread     ( __getVar( out, line) )
        #read print file
        elif key.prints           in line.lower():
            line_str = f.readline()
            line     = line_str.split('#')[0]
            while EndKey(key.prints) not in line.lower() and line_str != '':
                #eigenvalues
                if key.evals in line.lower():
                    line_str = f.readline()
                    line     = line_str.split('#')[0]
                    while EndKey(key.evals) not in line.lower() and line_str != '':
                        if   key.evalfile in line.lower():
                            out.set_evals_file ( __getVar( out, line,cst.STR) )
                        elif key.evalnum  in line.lower():
                            gvr.set_evalnum    ( __getVar( out, line,cst.FLOAT) )
                        else:
                            __check(line, 'eigenvalue', out) 
                        line_str = f.readline()
                        line     = line_str.split('#')[0]
                #eigenvectors
                elif key.evecs in line.lower():
                    line_str = f.readline()
                    line     = line_str.split('#')[0]
                    while EndKey(key.evecs) not in line.lower() and line_str != '' :
                        if   key.evecfile in line.lower():
                            out.set_evecs_file ( __getVar( out, line,cst.STR) )
                        elif key.evecnum  in line.lower():
                            gvr.set_evecnum    ( __getVar( out, line,cst.FLOAT) )
                        else:
                            __check(line, 'eigenvector', out) 
                        line_str = f.readline()
                        line     = line_str.split('#')[0]
                #microcanonical rate flux coeeficients
                elif key.kE in line.lower():
                    out.set_kE_file         (  __getVar( out, line,cst.STR) )
                #partition function
                elif key.partifunc in line.lower():
                    out.set_partifunc_file  (  __getVar( out, line,cst.STR) )
                #time evolution
                elif key.timeEvolution in line.lower():
                    gvr.turn_on_time_evolution()
                    line_str = f.readline()
                    line     = line_str.split('#')[0]
                    while EndKey(key.timeEvolution) not in line.lower() and line_str != '':
                        if   key.time_file in line.lower():
                            out.set_timeevol_file ( __getVar( out, line,cst.STR) )
                        elif key.time_step in line.lower():
                            gvr.set_time_step ( __getVar( out, line,cst.FLOAT) )
                        elif key.time_num  in line.lower():
                            gvr.set_time_num  ( __getVar( out, line,cst.FLOAT) )
                        elif key.initspecies in line.lower():
                            gvr.set_initspecies(  __getArr( out, f, None, cst.STR, EndKey(key.initspecies)) ) 
                        else:
                            __check(line, 'timeevolution', out) 
                        line_str = f.readline()
                        line     = line_str.split('#')[0]
                #relaxation collisional possibility kernel
                elif key.relaxkernel in line.lower():
                    line_str = f.readline()
                    line     = line_str.split('#')[0]
                    while EndKey(key.relaxkernel) not in line.lower() and line_str != '':
                        if   key.relaxfile in line.lower():
                            out.set_relaxkernel_file ( __getVar( out, line, cst.STR) )
                        elif key.printkernelname in line.lower():
                            gvr.set_printkernelname (  __getArr( out, f, __getUnit( out, line), cst.STR, EndKey(key.printkernelname) )  )
                        else:
                            __check(line, 'relaxkernel', out) 
                        line_str = f.readline()
                        line     = line_str.split('#')[0]
                else:
                    __check(line, 'print', out)
                line_str = f.readline()
                line     = line_str.split('#')[0]
        else:
            __check(line, 'parameter', out)
        line_str = f.readline()
        line     = line_str.split('#')[0]
        


def  __readSpecies(f, key, default, gvr, out):
#read species
    EndKey   =   lambda x: "end_"+x
    line_str =   f.readline()
    line     =   line_str.split('#')[0]
    while EndKey(key.species) not in line.lower() and line_str != '':
        name_tmp        =   default.name       ;  sp1_name_tmp        =   default.name       ;  sp2_name_tmp        =   default.name                        
        symbol_tmp      =   default.symbol     ;  sp1_symbol_tmp      =   default.symbol     ;  sp2_symbol_tmp      =   default.symbol                      
        geom_tmp        =   default.geom       ;  sp1_geom_tmp        =   default.geom       ;  sp2_geom_tmp        =   default.geom                        
        freq_tmp        =   default.freq       ;  sp1_freq_tmp        =   default.freq       ;  sp2_freq_tmp        =   default.freq                        
        e_type_tmp      =   default.e_type     ;  sp1_e_type_tmp      =   default.e_type     ;  sp2_e_type_tmp      =   default.e_type                      
        e_tmp           =   default.e          ;  sp1_e_tmp           =   default.e          ;  sp2_e_tmp           =   default.e                      
        ele_level_tmp   =   default.ele_level  ;  sp1_ele_level_tmp   =   default.ele_level  ;  sp2_ele_level_tmp   =   default.ele_level      
        rotsigma_tmp    =   default.rotsigma   ;  sp1_rotsigma_tmp    =   default.rotsigma   ;  sp2_rotsigma_tmp    =   default.rotsigma        
        mst_file_tmp    =   default.mst_file   ;  sp1_mst_file_tmp    =   default.mst_file   ;  sp2_mst_file_tmp    =   default.mst_file
        g09_file_tmp    =   default.g09_file   ;  sp1_g09_file_tmp    =   default.g09_file   ;  sp2_g09_file_tmp    =   default.g09_file
        opticalnum_tmp  =   default.opticalnum ;  sp1_opticalnum_tmp  =   default.opticalnum ;  sp2_opticalnum_tmp  =   default.opticalnum    
        freqscale_tmp   =   default.freqscale  ;  sp1_freqscale_tmp   =   default.freqscale  ;  sp2_freqscale_tmp   =   default.freqscale
        image_freq_tmp  =   None               ;  sp1_tmp             =   None               ;  sp2_tmp             =   None
        source_tmp      =   default.source     ;  sp1_source_tmp      =   default.source     ;  sp2_source_tmp      =   default.source
        diameter_tmp    =   default.diameter   ;  epsilon_tmp         =   default.epsilon    ;  bmax_tmp            =   default.bmax
        
        #read well
        if key.well in line.lower():
            line_str    = f.readline()
            line        = line_str.split('#')[0]
            while EndKey(key.well) not in line.lower() and line_str != '':
                if   key.geom       in line.lower():
                    symbol_tmp, geom_tmp =  __getArr( out,  f, __getUnit( out, line), cst.GEOM, EndKey(key.geom) )
                elif key.freq       in line.lower():
                    freq_tmp             =  __getArr( out,  f, __getUnit( out, line), cst.FLOAT, EndKey(key.freq) )
                elif key.E0K        in line.lower():
                    e_tmp                =  __getVar( out, line)
                    e_type_tmp           = "E0K"
                elif key.Eele       in line.lower():
                    e_tmp                =  __getVar( out, line)
                    e_type_tmp           = "Eele"
                elif key.ele_level  in line.lower():
                    ele_level_tmp        =  __getArr( out,  f, __getUnit( out, line), cst.ELEV, EndKey(key.ele_level))
                elif key.rotsigma   in line.lower():
                    rotsigma_tmp         =  __getVar( out, line, cst.FLOAT )
                elif key.mst_file   in line.lower():
                    mst_file_tmp         =  __getVar( out, line, cst.STR )
                elif key.name       in line.lower():
                    name_tmp             =  __getVar( out, line, cst.STR )
                elif key.g09_file   in line.lower():
                    g09_file_tmp         =  __getVar( out, line, cst.STR )
                elif key.diameter   in line.lower():
                    diameter_tmp         =  __getVar( out, line, cst.FLOAT )
                elif key.epsilon    in line.lower():
                    epsilon_tmp          =  __getVar( out, line, cst.FLOAT )
                elif key.bmax       in line.lower():
                    bmax_tmp             =  __getVar( out, line, cst.FLOAT )    
                elif key.opticalnum in line.lower():
                    opticalnum_tmp       =  __getVar( out, line, cst.FLOAT )
                elif key.freqscale  in line.lower():
                    freqscale_tmp        =  __getVar( out, line, cst.FLOAT )
                else:
                    __check(line, 'unimolecule', out)
                line_str    = f.readline()
                line        = line_str.split('#')[0]
            
            if name_tmp in gvr.well_dict:
                if isinstance(gvr.well_dict[name_tmp], species_lib.Well_class):    
                    #update some info of well and ignore other proerpties
                    gvr.well_dict[name_tmp].set_rotsigma(   rotsigma_tmp )
                    gvr.well_dict[name_tmp].set_mst_file(   mst_file_tmp )
                    gvr.well_dict[name_tmp].set_opticalnum( opticalnum_tmp )
                    gvr.well_dict[name_tmp].set_epsilon( epsilon_tmp )
                    gvr.well_dict[name_tmp].set_diameter( diameter_tmp )
                    gvr.well_dict[name_tmp].set_bmax( bmax_tmp )
                else:
                    out.error(' Error! The species-types for '+name_tmp+' are not consistent.')
            else:
                # create well object and push to well_stack
                well_tmp = species_lib.Well_class(  name_tmp,       symbol_tmp,         geom_tmp,           freq_tmp,     
                                                    freqscale_tmp,  e_type_tmp,         e_tmp,              ele_level_tmp,
                                                    rotsigma_tmp,   opticalnum_tmp,     epsilon_tmp,        diameter_tmp,       
                                                    bmax_tmp,       gvr.collid,         mst_file_tmp,       g09_file_tmp,      
                                                    source_tmp,     out.error )
                gvr.push_well_stack ( well_tmp )
                # push to well_dict
                gvr.push_well_dict  ( name_tmp, well_tmp)

        #read TS
        elif key.tranState in line.lower():
            line_str    = f.readline()
            line        = line_str.split('#')[0]
            while EndKey(key.tranState) not in line.lower() and line_str != '':
                if   key.geom       in line.lower():
                    symbol_tmp, geom_tmp =  __getArr( out,  f, __getUnit( out, line), cst.GEOM, EndKey(key.geom) )
                elif key.freq       in line.lower():
                    freq_tmp             =  __getArr( out,  f, __getUnit( out, line), cst.FLOAT, EndKey(key.freq) )
                elif key.E0K        in line.lower():
                    e_tmp                =  __getVar( out, line)
                    e_type_tmp           = "E0K"
                elif key.Eele       in line.lower():
                    e_tmp                =  __getVar( out, line)
                    e_type_tmp           = "Eele"
                elif key.ele_level  in line.lower():
                    ele_level_tmp        =  __getArr( out,  f, __getUnit( out, line), cst.ELEV,EndKey(key.ele_level))
                elif key.rotsigma   in line.lower():
                    rotsigma_tmp         =  __getVar( out, line)
                elif key.mst_file   in line.lower():
                    mst_file_tmp         =  __getVar( out, line, cst.STR )
                elif key.name       in line.lower():
                    name_tmp             =  __getVar( out, line, cst.STR )
                elif key.image_freq in line.lower():
                    image_freq_tmp       =  __getVar( out, line, cst.FLOAT )
                elif key.g09_file   in line.lower():
                    g09_file_tmp         =  __getVar( out, line, cst.STR )
                elif key.opticalnum in line.lower():
                    opticalnum_tmp       =  __getVar( out, line, cst.FLOAT )
                elif key.freqscale  in line.lower():
                    freqscale_tmp        =  __getVar( out, line, cst.FLOAT )
                else:
                    __check(line, 'unimolecule', out)
                line_str    = f.readline()
                line        = line_str.split('#')[0]
            
            if name_tmp in gvr.ts_dict:
                if isinstance( gvr.ts_dict[name_tmp], species_lib.TS_class):
                    #update some info of transition state
                    gvr.ts_dict[name_tmp].set_rotsigma( rotsigma_tmp )
                    gvr.ts_dict[name_tmp].set_mst_file( mst_file_tmp )
                    gvr.ts_dict[name_tmp].set_opticalnum( opticalnum_tmp )
                else:
                    out.error(' Error! The species-types for '+name_tmp+' are not consistent.')
            else:
                # create TS object and push to well_stact 
                ts_tmp = species_lib.TS_class(  name_tmp,           symbol_tmp,         geom_tmp,           freq_tmp, 
                                                freqscale_tmp,      e_type_tmp,         e_tmp,              ele_level_tmp,
                                                rotsigma_tmp,       opticalnum_tmp,     mst_file_tmp,       g09_file_tmp,       
                                                image_freq_tmp,     source_tmp,         out.error )
                gvr.push_ts_stack( ts_tmp )
                # push to well_dict
                gvr.push_ts_dict( name_tmp, ts_tmp) 

        #read bimolecular
        elif key.bim in line.lower():
            line_str    = f.readline()
            line        = line_str.split('#')[0]
            while EndKey(key.bim) not in line.lower() and line_str != '':
                if  key.name    in line.lower():
                    name_tmp    =  __getVar( out, line, cst.STR)
                elif  key.E0K   in line.lower():
                    e_tmp       =  __getVar( out, line)
                    e_type_tmp      = "E0K"
                elif key.Eele   in line.lower():
                    e_tmp       =  __getVar( out, line)
                    e_type_tmp  = "Eele"
                #sp1
                elif key.sp1_mol in line.lower():
                    line_str    = f.readline()
                    line        = line_str.split('#')[0]
                    while EndKey(key.sp1_mol) not in line.lower() and line_str != '':
                        if   key.geom     in line.lower():
                            sp1_symbol_tmp, sp1_geom_tmp =  __getArr( out,  f, __getUnit( out, line), cst.GEOM, EndKey(key.geom) )
                        elif key.freq     in line.lower():
                            sp1_freq_tmp                 =  __getArr( out,  f, __getUnit( out, line), cst.FLOAT, EndKey(key.freq) )
                        elif key.E0K      in line.lower():
                            sp1_e_tmp                =  __getVar( out, line)
                            sp1_e_type_tmp           = "E0K"
                        elif key.Eele     in line.lower():
                            sp1_e_tmp                =  __getVar( out, line)
                            sp1_e_type_tmp           = "Eele"
                        elif key.ele_level in line.lower():
                            sp1_ele_level_tmp        =  __getArr( out,  f, __getUnit( out, line), cst.ELEV, EndKey(key.ele_level))
                        elif key.rotsigma in line.lower():
                            sp1_rotsigma_tmp         =  __getVar( out, line)
                        elif key.mst_file in line.lower():
                            sp1_mst_file_tmp         =  __getVar( out, line, cst.STR  )
                        elif key.name     in line.lower():
                            sp1_name_tmp             =  __getVar( out, line, cst.STR  )
                        elif key.g09_file in line.lower():
                            sp1_g09_file_tmp         =  __getVar( out, line, cst.STR  )
                        elif key.opticalnum in line.lower():
                            sp1_opticalnum_tmp       =  __getVar( out, line, cst.FLOAT)
                        elif key.freqscale in line.lower():
                            sp1_freqscale_tmp        =  __getVar( out, line, cst.FLOAT )
                        else:
                            __check(line, 'unimolecule', out)
                        line_str    = f.readline()
                        line        = line_str.split('#')[0]
                    
                    if name_tmp in gvr.bim_dict:
                        if isinstance(gvr.bim_dict[name_tmp].sp1, species_lib.FragmentMol_class):
                            #update some info of fragment of bimolecule
                            gvr.bim_dict[name_tmp].sp1.set_rotsigma( sp1_rotsigma_tmp )
                            gvr.bim_dict[name_tmp].sp1.set_mst_file( sp1_mst_file_tmp )
                            gvr.bim_dict[name_tmp].sp1.set_opticalnum( sp1_opticalnum_tmp )
                        else:
                            out.error(' Error! The species-types for the sp1 of '+name_tmp+' are not consistent.')
                    else:
                        sp1_tmp = species_lib.FragmentMol_class(    sp1_name_tmp,           sp1_symbol_tmp,         sp1_geom_tmp,           sp1_freq_tmp, 
                                                                    sp1_freqscale_tmp,      sp1_e_type_tmp,         sp1_e_tmp,              sp1_ele_level_tmp, 
                                                                    sp1_rotsigma_tmp,       sp1_opticalnum_tmp,     sp1_mst_file_tmp,       sp1_g09_file_tmp,       
                                                                    sp1_source_tmp,         out.error )
                elif key.sp1_atom in line.lower():
                    line_str    = f.readline()
                    line        = line_str.split('#')[0]
                    while EndKey(key.sp1_atom) not in line.lower() and line_str != '':
                        if key.name        in line.lower():
                            sp1_name_tmp               =  __getVar( out, line, cst.STR )
                        elif key.symbol    in line.lower():
                            sp1_symbol_tmp             =  __getVar( out, line, cst.STR)
                        elif key.E0K       in line.lower():
                            sp1_e_tmp                  =  __getVar( out, line)
                        elif key.Eele      in line.lower():
                            sp1_e_tmp                  =  __getVar( out, line)
                        elif key.ele_level in line.lower():
                            sp1_ele_level_tmp          =  __getArr( out,  f, __getUnit( out, line), cst.ELEV, EndKey(key.ele_level))
                        else:
                           __check(line, 'atom', out)
                        line_str    = f.readline()
                        line        = line_str.split('#')[0]
                    
                    if name_tmp not in gvr.bim_dict:
                        sp1_tmp = species_lib.FragmentAtom_class( sp1_name_tmp, sp1_e_tmp, sp1_ele_level_tmp, sp1_symbol_tmp, sp1_source_tmp, out.error )
                
                #sp2
                elif key.sp2_mol in line.lower():
                    line_str    = f.readline()
                    line        = line_str.split('#')[0]
                    while EndKey(key.sp2_mol) not in line.lower() and line_str != '':
                        if   key.geom     in line.lower():
                            sp2_symbol_tmp, sp2_geom_tmp =  __getArr( out,  f, __getUnit( out, line), cst.GEOM, EndKey(key.geom) )
                        elif key.freq     in line.lower():
                            sp2_freq_tmp                 =  __getArr( out,  f, __getUnit( out, line), cst.FLOAT, EndKey(key.freq) )
                        elif key.E0K      in line.lower():
                            sp2_e_tmp                =  __getVar( out, line)
                            sp2_e_type_tmp           = "E0K"
                        elif key.Eele     in line.lower():
                            sp2_e_tmp                =  __getVar( out, line)
                            sp2_e_type_tmp           = "Eele"
                        elif key.ele_level in line.lower():
                            sp2_ele_level_tmp        =  __getArr( out,  f, __getUnit( out, line), cst.ELEV, EndKey(key.ele_level))
                        elif key.rotsigma in line.lower():
                            sp2_rotsigma_tmp         =  __getVar( out, line)
                        elif key.mst_file in line.lower():
                            sp2_mst_file_tmp         =  __getVar( out, line, cst.STR   )
                        elif key.name     in line.lower():
                            sp2_name_tmp             =  __getVar( out, line, cst.STR   )
                        elif key.g09_file in line.lower():
                            sp2_g09_file_tmp         =  __getVar( out, line, cst.STR   )
                        elif key.opticalnum in line.lower():
                            sp2_opticalnum_tmp       =  __getVar( out, line, cst.FLOAT )
                        elif key.freqscale in line.lower():
                            sp2_freqscale_tmp        =  __getVar( out, line, cst.FLOAT )
                        else:
                           __check(line, 'unimolecule', out)
                        line_str    = f.readline()
                        line        = line_str.split('#')[0]
                    
                    if name_tmp in gvr.bim_dict:
                        if isinstance(gvr.bim_dict[name_tmp].sp2, species_lib.FragmentMol_class):
                            #update some info of fragment of bimolecule
                            gvr.bim_dict[name_tmp].sp2.set_rotsigma( sp2_rotsigma_tmp )
                            gvr.bim_dict[name_tmp].sp2.set_mst_file( sp2_mst_file_tmp )
                            gvr.bim_dict[name_tmp].sp2.set_opticalnum( sp2_opticalnum_tmp )
                        else:
                            out.error(' Error! The species-types for the sp2  of '+name_tmp+' are not consistent.')
                    else:
                        sp2_tmp = species_lib.FragmentMol_class(    sp2_name_tmp,           sp2_symbol_tmp,         sp2_geom_tmp,           sp2_freq_tmp, 
                                                                    sp2_freqscale_tmp,      sp2_e_type_tmp,         sp2_e_tmp,              sp2_ele_level_tmp,  
                                                                    sp2_rotsigma_tmp,       sp2_opticalnum_tmp,     sp2_mst_file_tmp,       sp2_g09_file_tmp,
                                                                    sp2_source_tmp,         out.error           )
                elif key.sp2_atom in line.lower() and line_str != '':
                    line_str    = f.readline()
                    line        = line_str.split('#')[0]
                    while EndKey(key.sp2_atom) not in line.lower() and line_str != '':
                        if key.name        in line.lower():
                            sp2_name_tmp               =  __getVar( out, line, cst.STR)
                        elif key.symbol    in line.lower():
                            sp2_symbol_tmp             =  __getVar( out, line, cst.STR)
                        elif key.E0K       in line.lower():
                            sp2_e_tmp                  =  __getVar( out, line)
                        elif key.Eele      in line.lower():
                            sp2_e_tmp                  =  __getVar( out, line)
                        elif key.ele_level in line.lower():
                            sp2_ele_level_tmp          =  __getArr( out,  f, __getUnit( out, line), cst.ELEV, EndKey(key.ele_level))
                        else:
                           __check(line, 'atom', out)
                        line_str    = f.readline()
                        line        = line_str.split('#')[0]
                    
                    if name_tmp not in gvr.bim_dict:
                        sp2_tmp = species_lib.FragmentAtom_class( sp2_name_tmp, sp2_e_tmp, sp2_ele_level_tmp, sp2_symbol_tmp, sp2_source_tmp, out.error)
                else:
                   __check(line, 'bimolecule', out)
                line_str    = f.readline()
                line        = line_str.split('#')[0]
                
            if name_tmp not in gvr.bim_dict:
                # create bimolecular object and push it into bim_stack
                bim_tmp = species_lib.Bimolecular_class( name_tmp, e_type_tmp, e_tmp, sp1_tmp, sp2_tmp, out.error )
                gvr.push_bim_stack    ( bim_tmp )
                # push it to bim_dict
                gvr.push_bim_dict    ( name_tmp, bim_tmp)
        else:
           __check(line, 'species', out)
        line_str    = f.readline()
        line        = line_str.split('#')[0]


def  __readReaction(f, key, default, gvr, out):
#read reactions
    EndKey      =   lambda x: "end_"+x
    line_str    =   f.readline()
    line        =   line_str.split('#')[0]
    while EndKey(key.reaction) not in line.lower() and line_str !='':
        # barrier rxn
        if key.elem_rxn in line.lower():
            info_tmp        =   None
            pyr_file_tmp    =   default.pyr_file
            variation_tmp   =   default.variation
            tunnel_tmp      =   default.tunnel
            avgdiameter_tmp =   default.diameter
            s_lower_tmp     =   default.s_lower
            s_upper_tmp     =   default.s_upper
            invlaplace_tmp  =   default.invlaplace
            expression_tmp  =   default.expression
            A1_tmp          =   default.A1
            T1_tmp          =   default.T1
            n1_tmp          =   default.n1
            Ea1_tmp         =   default.Ea1
            A2_tmp          =   default.A2
            T2_tmp          =   default.T2
            n2_tmp          =   default.n2
            Ea2_tmp         =   default.Ea2
            
            line_str        =   f.readline()
            line            =   line_str.split('#')[0]
            while EndKey(key.elem_rxn) not in line.lower() and line_str != '':
                if key.info in line.lower():
                    info_tmp     =  __getVar( out,  line, cst.STR )
                    if len(info_tmp.split('-')) != 3:
                        out.error ( ' Wrong info ' + info_tmp +' of barriered reaction. It should be Reactant_name-TS_name-P_name ' ) 
                    R_name_tmp   = info_tmp.split('-')[0]
                    TS_name_tmp  = info_tmp.split('-')[1]
                    P_name_tmp   = info_tmp.split('-')[2]
                elif key.pyr_file  in line.lower():
                    pyr_file_tmp  =  __getVar( out,  line, cst.STR )
                elif key.variation in line.lower():
                    variation_tmp =  __getVar( out,  line, cst.STR )
                elif key.tunnel    in line.lower():
                    tunnel_tmp    =  __getVar( out,  line, cst.STR )
                elif key.s_lower   in line.lower():
                    s_lower_tmp   =  __getVar( out,  line, cst.FLOAT)
                elif key.s_upper   in line.lower():
                    s_upper_tmp   =  __getVar( out,  line, cst.FLOAT)    
                elif key.invlaplace in line.lower():
                    invlaplace_tmp=  True
                    line_str      =   f.readline()
                    line          =   line_str.split('#')[0]
                    while EndKey(key.invlaplace) not in line.lower() and line_str != '':
                        if key.expression in line.lower():
                            expression_tmp=  __getVar( out, line, cst.STR ) 
                        elif key.A1    in line.lower():
                            A1_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.T1 in line.lower():
                            T1_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.n1    in line.lower():
                            n1_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.Ea1   in line.lower():
                            Ea1_tmp       =   __getVar( out, line, cst.FLOAT )
                        elif key.A2    in line.lower():
                            A2_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.T2 in line.lower():
                            T2_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.n2    in line.lower():
                            n2_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.Ea2   in line.lower():
                            Ea2_tmp       =   __getVar( out, line, cst.FLOAT )
                        else:
                            __check(line, 'invlaplace', out)
                        line_str    = f.readline()
                        line        = line_str.split('#')[0]
                else:
                   __check(line, 'barrierrxn', out)
                line_str    = f.readline()
                line        = line_str.split('#')[0]
            #create rxn object up to whether read info from pyr_file
            if pyr_file_tmp == default.pyr_file:
                # create rxn from input species      
                # forward rxn
                rxn_f_tmp = reaction_lib.BarrierRxn_class(  _Rname      =   R_name_tmp, 
                                                            _TSname     =   TS_name_tmp,
                                                            _Pname      =   P_name_tmp,
                                                            _pyr_file   =   pyr_file_tmp, 
                                                            _variation  =   variation_tmp, 
                                                            _tunnel     =   tunnel_tmp,
                                                            _avgdiameter=   avgdiameter_tmp,
                                                            _s_lower    =   s_lower_tmp,
                                                            _s_upper    =   s_upper_tmp,
                                                            _invlaplace =   invlaplace_tmp,
                                                            _expression =   expression_tmp,
                                                            _A1         =   A1_tmp,
                                                            _T1         =   T1_tmp,
                                                            _n1         =   n1_tmp,
                                                            _Ea1        =   Ea1_tmp,
                                                            _A2         =   A2_tmp,
                                                            _T2         =   T2_tmp,
                                                            _n2         =   n2_tmp,
                                                            _Ea2        =   Ea2_tmp,
                                                            gvr         =   gvr,
                                                            errorf      =   out.error ) 
                #check redundant definition before pushing elementary rxn into rxn_stack
                for rxn in gvr.rxn_stack:
                    if rxn_f_tmp.Rname == rxn.Rname and rxn_f_tmp.Pname == rxn.Pname:
                        out.error( " Error: the elemntary reaction "+rxn_f_tmp.Rname+" --> "+rxn_f_tmp.Pname+" has already been defined, redundant definition is not allowed." ) 
                gvr.push_rxn_stack(rxn_f_tmp)
                
                # reverse rxn
                rxn_r_tmp = reaction_lib.BarrierRxn_class( _reverse_rxn   =   rxn_f_tmp )
                #check redundant definition before pushing elementary rxn into rxn_stack
                for rxn in gvr.rxn_stack:
                    if rxn_r_tmp.Rname == rxn.Rname and rxn_r_tmp.Pname == rxn.Pname:
                        out.error( " Error: the elemntary reaction from "+rxn_r_tmp.Rname+" to "+rxn_r_tmp.Pname+" has already been defined, redundant definition is not allowed." ) 
                gvr.push_rxn_stack(rxn_r_tmp)
            #create rxn from polyrate output file, inverse-Laplace-transform will be ingored here
            else:
                # forward rxn
                rxn_f_tmp = reaction_lib.BarrierRxn_class(  _Rname      =   R_name_tmp, 
                                                            _TSname     =   TS_name_tmp,
                                                            _Pname      =   P_name_tmp,
                                                            _pyr_file   =   pyr_file_tmp, 
                                                            _variation  =   variation_tmp, 
                                                            _tunnel     =   tunnel_tmp,
                                                            _avgdiameter=   avgdiameter_tmp,
                                                            _s_lower    =   s_lower_tmp,
                                                            _s_upper    =   s_upper_tmp,
                                                            gvr         =   gvr,
                                                            errorf      =   out.error ) 
                gvr.push_rxn_stack(rxn_f_tmp)
                # reverse rxn
                rxn_r_tmp = reaction_lib.BarrierRxn_class( _reverse_rxn   =   rxn_f_tmp )
                gvr.push_rxn_stack(rxn_r_tmp)
        # barrierless rxn
        elif key.barrierlessrxn in line.lower():
            info_tmp        =   None
            avgdiameter_tmp =   default.diameter
            rxn_ener_tmp    =   None
            invlaplace_tmp  =   default.invlaplace
            expression_tmp  =   default.expression
            A1_tmp          =   default.A1
            T1_tmp          =   default.T1
            n1_tmp          =   default.n1
            Ea1_tmp         =   default.Ea1
            A2_tmp          =   default.A2
            T2_tmp          =   default.T2
            n2_tmp          =   default.n2
            Ea2_tmp         =   default.Ea2
                
            line_str        = f.readline()
            line            = line_str.split('#')[0]
            while EndKey(key.barrierlessrxn) not in line.lower() and line_str != '':
                if key.info in line.lower():
                    info_tmp        =  __getVar( out,  line, cst.STR )
                    if len(info_tmp.split('-')) != 2:
                        out.error ( ' Wrong info ' + info_tmp +' of barriered reaction. It should be Reactant_name-P_name ' ) 
                    R_name_tmp      =  info_tmp.split('-')[0]
                    P_name_tmp      =  info_tmp.split('-')[1]
                elif key.avgdiameter in line.lower():
                    avgdiameter_tmp =  __getVar( out, line, cst.FLOAT )
                elif key.rxn_ener in line.lower():
                    rxn_ener_tmp    =  __getVar( out, line, cst.FLOAT )
                elif key.invlaplace in line.lower():
                    invlaplace_tmp=  True
                    line_str        =   f.readline()
                    line            =   line_str.split('#')[0]
                    while EndKey(key.invlaplace) not in line.lower() and line_str != '':
                        if key.expression in line.lower():
                            expression_tmp=  __getVar( out, line, cst.STR ) 
                        elif key.A1    in line.lower():
                            A1_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.T1 in line.lower():
                            T1_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.n1    in line.lower():
                            n1_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.Ea1   in line.lower():
                            Ea1_tmp       =   __getVar( out, line, cst.FLOAT )
                        elif key.A2    in line.lower():
                            A2_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.T2 in line.lower():
                            T2_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.n2    in line.lower():
                            n2_tmp        =   __getVar( out, line, cst.FLOAT )
                        elif key.Ea2   in line.lower():
                            Ea2_tmp       =   __getVar( out, line, cst.FLOAT )
                        else:
                            __check(line, 'invlaplace', out)
                        line_str    = f.readline()
                        line        = line_str.split('#')[0]
                else:
                   __check(line, 'barrierlessrxn', out)
                line_str    = f.readline()
                line        = line_str.split('#')[0]

            # create barrierless rxn from input  
            # forward rxn
            rxn_f_tmp = reaction_lib.BarrierlessRxn_class(  _Rname       =   R_name_tmp, 
                                                            _Pname       =   P_name_tmp,
                                                            _avgdiameter =   avgdiameter_tmp,
                                                            _rxn_ener    =   rxn_ener_tmp,
                                                            _invlaplace  =   invlaplace_tmp,
                                                            _expression  =   expression_tmp,
                                                            _A1          =   A1_tmp,
                                                            _T1          =   T1_tmp,
                                                            _n1          =   n1_tmp,
                                                            _Ea1         =   Ea1_tmp,
                                                            _A2          =   A2_tmp,
                                                            _T2          =   T2_tmp,
                                                            _n2          =   n2_tmp,
                                                            _Ea2         =   Ea2_tmp,
                                                            gvr          =   gvr,
                                                            errorf       =   out.error ) 
            #check redundant definition before push elementary rxn into rxn_stack
            for rxn in gvr.rxn_stack:
                if rxn_f_tmp.Rname == rxn.Rname and rxn_f_tmp.Pname == rxn.Pname:
                    out.error( " Error: the elemntary reaction "+rxn_f_tmp.Rname+" --> "+rxn_f_tmp.Pname+" has already been defined, redundant definition is not allowed." ) 
            gvr.push_rxn_stack(rxn_f_tmp)
            # reverse rxn
            rxn_r_tmp = reaction_lib.BarrierlessRxn_class( _reverse_rxn   =   rxn_f_tmp ) 
            #check redundant definition before push elementary rxn into rxn_stack
            for rxn in gvr.rxn_stack:
                if rxn_r_tmp.Rname == rxn.Rname and rxn_r_tmp.Pname == rxn.Pname:
                    out.error( " Error: the elemntary reaction from "+rxn_r_tmp.Rname+" to "+rxn_r_tmp.Pname+" has already been defined, redundant definition is not allowed." ) 
            gvr.push_rxn_stack(rxn_r_tmp)
        else:
           __check(line, 'reaction', out)
        line_str    = f.readline()
        line        = line_str.split('#')[0]


def std_readInput(stdin_file_str, gvr, out):
    try:
        f           =   open(stdin_file_str,mode='r')
    except IOError or PermissionError:
        out.error(' Can not open file '+ stdin_file_str )

    default     =   gl.Default_class()
    key         =   gl.Key_class() 
    
    line_str    = f.readline()
    line        = line_str.split('#')[0]
    while line_str != '':
        #read global variable
        if key.param in line.lower():
            __readParameter(f, key, default, gvr, out)

        #read reactions
        elif key.reaction in line.lower():
            __readReaction (f, key, default, gvr, out)     

        #read species variables
        elif key.species in line.lower():
            __readSpecies  (f, key, default, gvr, out) 
            
        # check validation of line string
        else:
           __check(line, 'input', out)

        line_str    = f.readline()
        line        = line_str.split('#')[0]


def readParallel( stdin_file_str, error ):
    try:
        f           =   open(stdin_file_str,mode='r')
    except IOError or PermissionError or FileNotFoundError:
        error(' Can not open file '+ stdin_file_str )
        exit(1)
    
    nproc       = 1
    parallel    = 'serial'
    cnt         = 0

    line        = f.readline()
    while line != '':
        if '#nproc' in line.lower():
            cdata = line.strip().split()
            if len(cdata) != 2:
                error(' Wrong in #nproc ...\n It should be like "#nproc  4" ')
                exit(1)
            try:
                nproc = int(cdata[1])
            except:
                error(' Wrong in the value of #nproc ' + cdata[1] + ', not a number.')
            cnt = cnt + 1

        elif '#parallel' in line.lower():
            cdata = line.strip().split()
            if len(cdata) != 2:
                error(' Wrong in #parallel ... It should be like "#parallel  mp" ')
                exit(1)
            parallel = cdata[1]
            if parallel != 'serial' and parallel != 'mp' and parallel != 'mpi':
                error(' Wrong in the value of #parallel ' + parallel + ', it should be "serial" or "mp" or "mpi". ' )
                exit(1)
            cnt = cnt + 1
        if cnt == 2 or  parallel=='mpi' :
            break
        line    = f.readline()
    f.close()

    return parallel, nproc
    
