#
#   TUMME 3.1  is licensed under the Apache License, Version 2.0 (the "License");
#   you may not use TUMME 3.1 except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   The license is also given in the LICENSE file.
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and limitations under the License.
#
# -------------------------------------------------------------------------------------------
# 
#   TUMME 3.1
#           
#   Rui Ming Zhang
#     
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#   
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#                  
#   Xuefei Xu
#
#   Center for Combustion Energy, Department of Energy and Power
#   Engineering, and Key Laboratory for Thermal Science and Power 
#   Engineering of Ministry of Education, Tsinghua University, Beijing
#
#   Donald G. Truhlar
# 
#   Department of Chemistry, Chemical Theory Center, and Minnesota 
#   Supercomputing Institute, University of Minnesota, Minneapolis, 
#   Minnesota
#  
#   The master equation code includes multi-structural variational 
#   transition state theory with small-curvature tunneling (MS-VTST/SCT)
#   for calculating the microcanonical flux coefficients needed as input
#   to the master equation. The TUMME program can do double, quadruple, 
#   and octuple precision, it can execute with MP or MPI parallel schemes, 
#   and it can interface to Gaussian, Polyrate, Gaussrate, and/or MSTor 
#   output files. Report bugs to: zrm17@tsinghua.org.cn
# 
#   Copyright 2020~2023
#   Donald G. Truhalar and Regents of the University of Minnesota
#   Xuefei Xu and the Tsinghua University 
#



import  me_lib          as      mel
import parallel_lib     as      pl
from    const_lib       import  cst
from    math            import  ceil


def std_solver(tp_list, gvr, out):
    #convert temperature-pressure list to temperature-pressure dictionary
    tp_dict     =   {}
    p_tmp       =   []
    T_tmp       =   tp_list[0][0]
    ME_queue    =   []

    for i in range( len(tp_list) ):
        if tp_list[i][0] == T_tmp:
            p_tmp.append( tp_list[i][1] )
        else:
            tp_dict.update( { T_tmp: p_tmp } )
            T_tmp   =   tp_list[i][0]
            p_tmp   =   []
            p_tmp.append( tp_list[i][1] )
    tp_dict.update( { T_tmp: p_tmp } )

    for T in tp_dict.keys():
        # set parameters for each cycle
        #
        gvr.cycle_param.set_ener_step( gvr.ESOT * T )

        gvr.cycle_param.set_ener_ref ( gvr.EEOT * T  + max([ sp.E0 for sp in gvr.ts_stack+gvr.phase_stack ]) )
        
        gvr.cycle_param.set_T ( T )

        #set relaxation kernels for wells
        #
        for well in gvr.well_stack:
            well.setKernel ( gvr )

        #set thermal for phases
        #
        for ph in gvr.phase_stack:
            ph.setThermal ( gvr.dE )

        #set kE for w2b&w2w
        #
        for rxn in gvr.rxn_stack:
            rxn.setReaction ( gvr, out.error )

        #for every pressure
        #
        for p in tp_dict[T]:    

            #set pressure
            #
            gvr.cycle_param.set_p ( p )

            #set collision
            #
            for well in gvr.well_stack:
                well.collid.setCollisionFrequency() 

            #print cycle params and kernels
            #
            out.printCycleparam( gvr ) 
            out.printKernelinfo( gvr.well_stack )

            #solve master equation
            #
            ME = mel.MasterEquation()

            ME.solveRateConstant( gvr, out )

            ME.solveTimeEvolution(gvr)

            ME.clear()

            ME_queue.append ( ME )

            #print ME info
            #
            out.printMEinfo(ME, [x.name for x in gvr.phase_stack] )

        #clear rxn
        #
        for rxn in gvr.rxn_stack:
            rxn.clear()

    return ME_queue




def serial_solver( gvr, out ):
    #set temperture-pressure list
    tp_list     =   []
    for t in gvr.temperature:
        for p in gvr.pressure:
            tp_list.append ( [t, p] )

    #standard solver
    ME_queue  =   std_solver( tp_list, gvr, out)
    
    #put ME into ME_stack
    gvr.set_ME_stack ( ME_queue )



def mp_solver( mpv, gvr, out ):
    rank        =   mpv.rank
    barrier     =   mpv.barrier
    nproc       =   mpv.nproc
    # set total tp_list
    tp_list     =   []   
    for t in gvr.temperature:
        for p in gvr.pressure:
            tp_list.append ( [t, p] )
    # set tp_list for every rank
    if len( tp_list ) < nproc:
        out.error(" The number of (T,p) pairs should be greater than nproc. ")
    n           =   int( ceil( len( tp_list ) / nproc ) )
    I           =   rank * n
    J           =   min( ( rank+1 )*n, len( tp_list ) )
    tp_list     =   tp_list[ I: J ]
    

    if tp_list != []:
        #standard solver
        ME_queue_tmp =  std_solver( tp_list, gvr, out)
        #put ME into queue
        mpv.ME_queue.put ( ( rank, ME_queue_tmp ) )
           
    #Synchronize
    barrier.wait()
    
    #set ME_stack
    if rank == 0:
        nproc    =  int(ceil( len(gvr.temperature)*len(gvr.pressure) / n ))
        ME_stack =  [mpv.ME_queue.get() for i in range( nproc ) ]
        ME_stack.sort()
        ME_stack =  [ x[1] for x in ME_stack ]

        for x in ME_stack:
            for xx in x:
                gvr.push_ME_stack(xx)



def mpi_solver( comm, gvr, out ):
    rank        =   comm.Get_rank()
    nproc       =   comm.Get_size()
    ME_queue_tmp=   []
    # set total tp_list
    tp_list     =   []   
    for t in gvr.temperature:
        for p in gvr.pressure:
            tp_list.append ( [t, p] )
    # set tp_list for every rank
    if  len( tp_list ) < nproc:
        out.error(" The number of (T,p) pairs should be greater than nproc. ")
    n           =   int( ceil( len( tp_list ) / nproc ) )
    I           =   rank * n
    J           =   min( ( rank+1 )*n, len( tp_list ) )
    tp_list     =   tp_list[ I: J ]
    
    if tp_list != []:
        #standard solver
        ME_queue_tmp = std_solver( tp_list, gvr, out)

    #Synchronize
    comm.barrier()
    
    # Gather
    pl.MPI_gather_ME_class(ME_queue_tmp, comm, gvr)
    if rank == 0:
        gvr.set_ME_stack (ME_queue_tmp)

